#! /bin/sh
# \
exec tclsh${TCL} "$0" ${1+"$@"}
#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------
namespace eval ::thatch {
	if {[package provide thatch] ne {}} { return }

	proc thatchBootstrap {{fn ""}} {
		if {$fn eq ""} { set fn [info script] }
		if {[catch {
			set f [open $fn r]
			fconfigure $f -eofchar \x1a
			read $f
			seek $f 1 current
			set d [read $f]
			close $f
		} err]} {
			return 0
		}
		if {[string length $d] == 0} { return 0 }
		uplevel #0 $d
		package require thatch
	}
	package ifneeded thatch 0.5 [namespace current]::thatchBootstrap
}

package require thatch
::thatch::thatchSetup [info script] -nofile -noglob

source /bin/thatchtool.tcl
#-------------------------------------------------------------------

# THATCH /thatch/thatch.tcl size 27626 mtime 1662916831
#
# Thatch - Tcl Attach
#
#
# Stuart Cassoff
#
# May 2009
# December 2009
# January 2010
# October 2010
# March 2012
# Summer 2012
# Spring 2018
# Version 0.1
#
# Spring 2018
# Version 0.2
#
# Summer 2018
# Version 0.3
#
# Summer 2022
# Version 0.4
#
# Summer 2022
# Version 0.5
#


[namespace eval ::thatch {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	thatch
		title	Thatch
		descr	"App Builder"
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.5
	}]
	dict set pkgconfig importfrom [namespace current]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2009,2018,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}; # pkgconfig


#-------------------------------------------------------------------
# Thatch preamble and bootstrap code
#-------------------------------------------------------------------


proc preamble {{wish {0}}} {
	return "#! /bin/sh\n# \\\nexec [expr {$wish ? "wi" : "tcl"}]sh\${TCL} \"\$0\" \${1+\"\$@\"}"
}


variable bootstrap [string range {
#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------
namespace eval ::thatch {
	if {[package provide thatch] ne {}} { return }

	proc thatchBootstrap {{fn ""}} {
		if {$fn eq ""} { set fn [info script] }
		if {[catch {
			set f [open $fn r]
			fconfigure $f -eofchar \x1a
			read $f
			seek $f 1 current
			set d [read $f]
			close $f
		} err]} {
			return 0
		}
		if {[string length $d] == 0} { return 0 }
		uplevel #0 $d
		package require thatch
	}
	package ifneeded thatch @VER@ [namespace current]::thatchBootstrap
}

package require thatch
::thatch::thatchSetup [info script] @THATCHOPTS@

source @MAIN@
#-------------------------------------------------------------------
} 1 end-1]


#-------------------------------------------------------------------
# Other
#-------------------------------------------------------------------

proc linterleave {l1 l2} {
	set l {}
	foreach e1 $l1 e2 $l2 { lappend l $e1 $e2 }
	return $l
}

proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

# In My Namespace
proc imn {what} {
	return [string trimright [uplevel 1 {namespace current}] :]::$what
}

proc aliasconst {name val args} {
	if {[llength $args] % 2 != 0} { error "must be %2!" }
	foreach {n v} [linsert $args 0 $name $val] {
		interp alias {} [imn $n] {} return -level 0 $v
	}
}

proc unknownOption {opt opts} {
	set msg "unknown option \"$opt\""
	if {[llength $opts] == 0} { return $msg }
	append msg ", must be: "
	append msg [lindex $opts 0]
	if {[llength $opts] == 2} {
		append msg " or [lindex $opts 1]"
	} elseif {[llength $opts] > 2} {
		append msg ", " [join [lrange $opts 1 end-1] ", "] " or " [lindex $opts end]
	}
	return $msg
}

#-------------------------------------------------------------------
# File
#-------------------------------------------------------------------

proc openItIn {fn {mode r} args} {
	if {$fn in {- stdin}} {
		set f stdin
	} else {
		if {[string index $fn 0] eq "@"} {
			set f [string range $fn 1 end]
		} else {
			set f [open $fn $mode]
		}
	}
	if {[llength $args] > 0} {
		chan configure $f {*}$args
	}
	return $f
}

proc openItOut {fn {mode w} args} {
	if {$fn in {- stdout}} {
		set f stdout
	} else {
		set f [open $fn $mode]
	}
	if {[llength $args] > 0} {
		chan configure $f {*}$args
	}
	return $f
}

proc closeIt {f} {
	if {$f in {stdin stdout}} {
		return {}
	}
	return [chan close $f]
}

proc inhaleIt {fn args} {
	return [read [set f [openItIn $fn r {*}$args]]][closeIt $f]
}

interp alias {} [imn openIt] {} [imn openItIn]

#-------------------------------------------------------------------
# Transforms
#-------------------------------------------------------------------

proc thatchWhatTransforms {{pat *}} {
	return [lsort -dictionary [lsearch -inline -all -glob [list copy base64 zip] $pat]]
}

if {[namespace ensemble exists binary] && "encode" in [dict keys [namespace ensemble configure binary -map]] && "decode" in [dict keys [namespace ensemble configure binary -map]]} {
	namespace eval ::base64 {
		proc encode {args} { binary encode base64 {*}$args }
		proc decode {args} { binary decode base64 {*}$args }
		package provide base64 2000
	}
}; # if binary encode/decode

namespace eval ::zip {
	proc encode {args} { zlib compress   {*}$args }
	proc decode {args} { zlib decompress {*}$args }
	package provide zip 2000
}

proc thatchTransform {data process args} {
	foreach alg $args {
		if {$alg eq "copy"} { continue }
		package require $alg
		set data [::${alg}::${process} $data]
	}
	return $data
}

#-------------------------------------------------------------------
# Checksums
#-------------------------------------------------------------------

proc thatchWhatChecksums {{pat *}} {
	return [lsort -dictionary [lsearch -inline -all -glob [list md5 sha256] $pat]]
}

proc thatchChecksum {data alg} {
	package require $alg
	set ns [string map [list md5 md5 sha256 sha2] $alg]
	return [::${ns}::${alg} -hex $data]
}

proc thatchMakeChecksums {data args} {
	set cksums {}
	foreach alg [lsort -nocase -dictionary $args] {
		set alg [string tolower $alg]
		lappend cksums $alg:[thatchChecksum $data $alg]
	}
	return [join $cksums ,]
}

proc thatchCheckChecksums {data args} {
	set errs {}
	foreach cksum [lsort -nocase -dictionary [split $args ,]] {
		lassign [split $cksum :] alg sum
		set alg [string tolower $alg]
		set csum [thatchChecksum $data $alg]
		if {$csum ne $sum} {
			lappend errs [list $cksum $csum]
		}
	}
	return $errs
}

#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#
#	i == file info
#	N == \n
#	Z == ^Z
#
#	ZN# THATCH iiiiiiiiiiiiiiiN
#	||__sig___||__file info__||
#	|          |file info line|
#	|_____file info chunk_____|
#
#
#	file info       fis
#	file info line  fil
#	file info chunk fic
#	file info dict  fid
#	file spec       fs
#
#-------------------------------------------------------------------


#
# File Spec (fs) fields
#
proc fsFields {} {
	return [list name size mtime base64 zip md5 sha256 remove rehome modbase]
}

proc fid->fis {fid} {
	set name [dict get $fid name]
	if {[dict exists $fid rehome]} {
		if {[dict get $fid rehome] ne ""} {
			set name [dict get $fid rehome]
		}
		dict unset fid name
		dict unset fid rehome
	}
	if {[dict exists $fid modbase] && [dict get $fid modbase] eq {}} {
		dict unset fid modbase
	}
	return "$name $fid"
}

proc fis->fid {fis} {
	set q [dict create name [lindex $fis 0] {*}[lrange $fis 1 end]]
}

proc fidFields {args} {
	return "name transforms checksums size stored-size mtime modbase"
}

proc newFid {args} {
	return [dict create ${*}[fidFields] {*}$args]
}

#
# File Spec (fs) -> File Info Dict (fid)
#
proc fs->fid {fs {dv {}}} {
	if {$dv ne ""} { upvar $dv data }
	lassign $fs {*}[fsFields]
	set transforms [list]
	set chanconf [dict create]

	# If zip & base64 do zip first
	if {$zip == 1} {
		lappend transforms zip
		dict set chanconf -translation binary
	}
	if {$base64 == 1} {
		lappend transforms base64
		dict set chanconf -translation binary
	}

	set fid [dict create name $name]

	if {[llength $transforms] > 0} { dict set fid transforms $transforms }

	if {$dv eq ""} {
		return $fid
	}

	set filename $name
	file stat $filename s
	set data [inhaleIt $filename {*}$chanconf]

	set algs [list]
	if {$md5 == 1}  { lappend algs md5 }
	if {$sha256 == 1} { lappend algs sha256 }

	set checksums [thatchMakeChecksums $data {*}$algs]
	set data [thatchTransform $data encode {*}$transforms]
	set s(stored-size) [string bytelength $data]

	if {$checksums ne ""} {
		dict set fid checksums $checksums
	}

	if {$size == 1} {
		dict set fid size $s(size)
		if {$s(size) != $s(stored-size)} {
			dict set fid stored-size $s(stored-size)
		}
	}

	if {$mtime == 1} {
		dict set fid mtime $s(mtime)
	}

	dict set fid rehome $rehome
	dict set fid modbase $modbase

	return $fid
}

#
# Writes a Thatch header
#
proc fid->fic {of fid} {
	puts -nonewline $of [^Z][sig][fid->fis $fid]\n
}

#
# Writes a File Info Chunk (fic): the Thatch header and the file
# from a File Spec (fs)
#
proc fs->fic {of fs} {
	set fid [fs->fid $fs data]
	fid->fic $of $fid
	puts -nonewline $of $data
}

#
# Arguments (ags) -> File Spec List (fsl)
#
proc ags->fsl {ags} {
	set fsl [list]
	#         name      var     val default
	set al [list {*}{
		{ -         doArgs    0  1 }
		{ allsize   allSize   1 -1 }
		{ allmtime  allMtime  1 -1 }
		{ allbase64 allBase64 1 -1 }
		{ allzip    allZip    1 -1 }
		{ allmd5    allMd5    1 -1 }
		{ allsha256 allSha256 1 -1 }
		{ noallsize   allSize   0 -1 }
		{ noallmtime  allMtime  0 -1 }
		{ noallbase64 allBase64 0 -1 }
		{ noallzip    allZip    0 -1 }
		{ noallmd5    allMd5    0 -1 }
		{ noallsha256 allSha256 0 -1 }
		{ size      size      1 -1 }
		{ mtime     mtime     1 -1 }
		{ base64    base64    1 -1 }
		{ zip       zip       1 -1 }
		{ md5       md5       1 -1 }
		{ sha256    sha256    1 -1 }
		{ nosize    size      0 -1 }
		{ nomtime   mtime     0 -1 }
		{ nobase64  base64    0 -1 }
		{ nozip     zip       0 -1 }
		{ nomd5     md5       0 -1 }
		{ nosha256  sha256    0 -1 }
		{ allremove allRemove 1  0 }
		{ remove    remove    1 -1 }
		{ file      argIsFs   1  0 }
		{ home      argIsHm   1  0 }
		{ rehome    argIsRHm  1  0 }
		{ modbase   argIsMb   1  0 }
		{ rename    argIsRen  1  0 }
	}]
	foreach a $al { set [lindex $a 1] [lindex $a 3] }

	set    homedir ""
	set  rehomedir ""
	set modbasedir ""
	set resetOpts 1
	set gopts {}
	# Arg options override FileSpec options
	foreach arg $ags {
		if {$resetOpts} {
			set resetOpts 0
			set rename ""
			foreach a $al {
				# Only reset specific options
				if {[lindex $a 0] ni [fsFields]} { continue }
				set [lindex $a 1] [lindex $a 3]
			}
		}
		set home    ""
		set rehome  ""
		set modbase ""
		if {$argIsHm} {
			set argIsHm 0
			set homedir $arg
			continue
		}
		if {$argIsRHm} {
			set argIsRHm 0
			set rehomedir $arg
			continue
		}
		if {$argIsMb} {
			set argIsMb 0
			set modbasedir $arg
			continue
		}
		if {$argIsRen} {
			set argIsRen 0
			set rename $arg
			continue
		}
		# Does arg look like an option or a FileSpec?
		if {$doArgs && !$argIsFs && [string index $arg 0] eq "-"} {
			# Valid options are assigned to vars and added to gopts
			if {[set i [lsearch -index 0 $al [string range $arg 1 end]]] != -1} {
				set [lindex $al $i 1] [lindex $al $i 2]
				if {[string range $arg 1 3] eq "all" || [string range $arg 1 5] eq "noall"} {
					lappend gopts [string range $arg 1 end]
				}
				continue
			}
		}
		set argIsFs 0
		# Check FileSpec for options
		set opts [lassign [split $arg %] name]
		foreach opt $opts {
			if {$argIsHm} {
				set argIsHm 0
				set home $opt
				continue
			}
			if {($opt ni $gopts || $opt eq "home") && [set i [lsearch -index 0 $al $opt]] != -1} {
				set [lindex $al $i 1] [lindex $al $i 2]
			}
		}

		# Empty name can be used to set global options
		if {$name eq ""} {
			set resetOpts 1
			continue
		}

		# Use defaults if options not supplied
		#foreach n {size mtime base64 zip md5 sha256 remove rehome modbase} {}
		foreach n {size mtime base64 zip md5 sha256 remove} {
			if {[set $n] == -1} { set $n [set all[string totitle $n]] }
		}

		if {$home eq ""} {
			set home [expr {[string index $name 0] eq "/" ? "" : $homedir}]
		}

		set file [file join $home $name]

		if {$rename ne ""} {
			set name $rename
		}
		if {$rehomedir ne ""} {
			set rehome [file join [expr {[string index $name 0] eq "/" ? "" : $rehomedir}] $name]
		}
		if {$modbasedir ne "" && [regexp -- {^([_[:alpha:]][:_[:alnum:]]*)-([[:digit:]].*)[.]tm$} [file tail $file] -> pn pv]} {
			set modbase $modbasedir
		}

		lappend fsl [list $file $size $mtime $base64 $zip $md5 $sha256 $remove $rehome $modbase]

		set resetOpts 1
	}

	return $fsl
}

proc thatchOpen {fn {dir {in}}} {
	if {$dir eq "out"} {
		set mode w
		set eofc ""
	} else {
		set mode r
		set eofc [^Z]
	}
	return [openIt[string totitle $dir] $fn $mode -eofchar $eofc -encoding utf-8 -translation lf]
}

#
# Create or modify a Thatch file
#
proc thatchOp {op ifn ofn args} {
	set if [openItIn $ifn]
	if {$op eq "create"} {
		chan configure $if -eofchar [^Z]
	} else {
		chan configure $if -eofchar ""
	}
	set of [thatchOpen $ofn out]
	chan copy $if $of
	closeIt $if
	foreach fs [ags->fsl $args] {
		fs->fic $of $fs
	}
	closeIt $of
	return 1
}

proc thatchCopyCallbackMain {t fn f fid args} {
	lassign $args fsl of
	puts -nonewline $of [read $f]
	return skip
}

proc thatchCopyCallbackFile {t fn f fid args} {
	lassign $args fsl of

	set dosize 0
	set dobase64 0
	set undobase64 0
	set dozip 0
	set undozip 0
	set ofid $fid
	if {[set i [lsearch -index 0 $fsl [dict get $fid name]]] != -1} {
		if {[lindex $fsl $i [lsearch [fsFields] remove]]} {
			return ""
		}
		foreach fld [list size mtime] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0 && [dict exists $fid $fld]} {
				set fid [dict remove $fid $fld]
			} elseif {$v == 1} {
				if {$fld eq "size"} {
					set dosize 1
				}
			}
		}
		foreach fld [list base64 zip] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0} {
				if {![dict exists $fid transforms]} { continue }
				set transforms [dict get $fid transforms]
				if {[set i [lsearch $transforms $fld]] != -1} {
					set transforms [lreplace $transforms $i $i]
				}
				if {$fld eq "base64"} {set undobase64 1}
				if {$fld eq "zip"}    {set undozip 1}
			} elseif {$v == 1} {
				if {[dict exists $fid transforms]} { continue }
				lappend transforms $fld
				if {$fld eq "base64"} {set dobase64 1}
				if {$fld eq "zip"}    {set dozip 1}
			}
			if {[llength $transforms] == 0} {
				set fid [dict remove $fid transforms]
			} else {
				dict set fid transforms $transforms
			}
		}
		set checksums [dgwd {} $fid checksums]
		set sums1 [split $checksums :,]
		foreach fld [list md5 sha256] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0} {
				if {[set i [lsearch $sums1 $fld]] != -1} {
					set sums1 [lreplace $sums1 $i $i+1]
				}
			} elseif {$v == 1} {
			}
		}
		if {[llength $sums1] == 0} {
			if {[dict exists $fid checksums]} {
				set fid [dict remove $fid checksums]
			}
		} else {
			#dict set fid checksums something
		}
	}

	set d [read $f]

	if {$undobase64} {
		set d [thatchTransform $d decode {*}[dgwd {} $ofid transforms]]
		if {[dict exists $fid stored-size]} {
			set fid [dict remove $fid stored-size]
		}
	}
	if {$undozip} {
		set d [thatchTransform $d decode {*}[dgwd {} $ofid transforms]]
		if {[dict exists $fid stored-size]} {
			set fid [dict remove $fid stored-size]
		}
	}
	if {$dosize} {
		dict set fid size [string length $d]
	}
	if {$dobase64} {
		set d [thatchTransform $d encode {*}[dict get $fid transforms]]
		dict set fid stored-size [string length $d]
	}
	if {$dozip} {
		set d [thatchTransform $d encode {*}[dict get $fid transforms]]
		dict set fid stored-size [string length $d]
	}
	fid->fic $of [dict remove $fid offset]
	puts -nonewline $of $d

	return skip
}

proc thatchCopy {ifn ofn args} {
	set of [thatchOpen $ofn out]
	thatchTraverse $ifn [list [imn thatchCopyCallbackFile] [imn thatchCopyCallbackMain]] [ags->fsl $args] $of
	closeIt $of
	return 1
}

proc thatchNameMatch {fid g} {
	return [string match $g [dict get $fid name]]
}

proc zthatchFullFileGlobCallback {t fn f fid args} {
	variable res
	if {$t == 2} {
		dict set res main [read $f]
		return skip
	}
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res thIndex [dict get $fid name] $fid
		}
	}
	return ""
}

proc zthatchFullFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFullFileGlobCallback]] {*}$args
	return $res
}

proc thatchFullFileGlobCallback {t fn f fid args} {
	variable res
	if {$t == 2} {
		set size [string length [read $f]]
		dict set res $fn [dict set fid size $size]
		return skip
	}
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res [dict get $fid name] $fid
		}
	}
	return ""
}

proc thatchFullFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFullFileGlobCallback]] {*}$args
	return $res
}

proc thatchFileGlobCallback {t fn f fid args} {
	if {$t == 2} { return "" }
	variable res
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res [dict get $fid name] $fid
		}
	}
	return ""
}

proc thatchFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFileGlobCallback]] {*}$args
	return $res
}

proc thatchTraverse {fn cb args} {
	# This is supposed to be a Thatch file so it should conform
	set f [thatchOpen $fn]

	# Setup callbacks
	lassign $cb cb mcb
	if {$mcb eq ""} {
		set mcb $cb
	}

	# Get the 'main'
	set fid [dict create name $fn offset 0]
	set r [$mcb 2 $fn $f $fid {*}$args]
	if {$r ne "skip"} {
		read $f
	}

	# Note its largesse
	set offset [tell $f]

	# Now let's see if there's anything else
	while {1} {
		# We have hit eof so check if it's a ^Z eof or a real eof.
		# Do this by turning off -eofchar detection and reading one char:
		# it should be a ^Z. Set -eofchar back to ^Z after.
		fconfigure $f -eofchar {}
		set cz [read $f 1]
		# Real eof? If so, $cz should also be "".
		if {[eof $f] || $cz eq ""} {
			# Normal case
			if {[eof $f] && $cz eq ""} {
				break
			}
			# todo, strange case
			break
		}
		# Gotta be control-zee
		if {$cz ne [^Z]} { binary scan $cz c c; error "Expected ^Z at $offset, got \"[format %#x $c]\"::[gets $f]" }
		fconfigure $f -eofchar [^Z]
		# After the ^Z should be the sig
		set something [read $f [string length [sig]]]
		if {$something ne [sig]} {
			if {[eof $f]} {
				if {$something eq ""} {
					# null block
					continue
				}
			} else {
				append something [read $f]
			}
			# anon block
			continue
		}
		# Now we should be at the start of the file info line
		set fis [gets $f]
		# Do some checks ...
		if {[eof $f]} { explode }

		# Recalc offset:
		#    ^Z           \n                                 fis                               \n
		incr offset; incr offset [string length [sig]]; incr offset [string length $fis]; incr offset

		set fid [fis->fid $fis]
		dict set fid offset $offset
		set r [$cb 1 $fn $f $fid {*}$args]
		if {$r eq "stop"} { break }
		set size  [dgwd -1 $fid size]
		set ssize [dgwd -1 $fid stored-size]
		if {$ssize > 0} { set size $ssize }
		if {$size == -1} {
			if {$r ne "skip"} {
				read $f
			}
			set offset [tell $f]
		} else {
			if {$r ne "skip"} {
				seek $f $size current
			}
			incr offset $size
		}
	}
	closeIt $f
}

proc thatchExtract {fn fid} {
	set f [thatchOpen $fn]

	set name [lindex [dict keys $fid] 0]
	seek $f [dict get $fid $name offset] start

	set size [dgwd -1 $fid $name size]
	set ssize [dgwd -1 $fid $name stored-size]

	set transforms [dgwd {} $fid $name transforms]
	set checksums [dgwd "" $fid $name checksums]

	if {$ssize < 1} {
		set ssize $size
	}

	if {$ssize > 0} {
		set data [read $f $ssize]
	} else {
		set data [read $f]
	}

	closeIt $f

	set data [thatchTransform $data decode {*}[lreverse $transforms]]

	set checksum_errs [thatchCheckChecksums $data {*}$checksums]
	if {[llength $checksum_errs] > 0} {
		error "checksum mismatches ($fid):\n![join $checksum_errs \n]!"
	}

	return $data
}

proc thatchExtractMain {fn} {
	set f [thatchOpen $fn]
	set data [read $f]
	closeIt $f
	return $data
}

proc thatchIndexGlob {fn args} {
	variable thIndex
	if {[llength $args] == 0} { set args [list *] }
	set res [dict create]
	foreach a $args {
		set res [dict merge $res [dict filter $thIndex key $a]]
	}
	return $res
}

proc thatchCreateIndex {fn} {
	thatchDestroyIndex
	variable thIndex [thatchFileGlob $fn]
	interp alias {} [imn thatchGlob] {} [imn thatchIndexGlob]
}

proc thatchDestroyIndex {} {
	variable thIndex
	unset -nocomplain thIndex
	interp alias {} [imn thatchGlob] {} [imn thatchFileGlob]
}

#-------------------------------------------------------------------
# Overrides
#-------------------------------------------------------------------

proc thatchSourceOverride {fn args} {
	if {[dict size [set fid [thatchGlob $fn {*}$args]]] > 0} {
		set  script [info script]
		info script [lindex [dict keys $fid]]
		set  retval [uplevel 1 [thatchExtract $fn $fid]]]
		info script $script
	} else {
		set retval [uplevel 1 [list ::tcl::tcl_source {*}$args]]
	}
	return $retval
}

proc thatchImageOverride {fn args} {
	if {
		   [llength $args] > 3
		&& [lindex $args 0] eq "create"
		&& [set ind [lsearch $args -file]] != -1
		&& [dict size [set fid [thatchGlob $fn [lindex $args $ind+1]]]] > 0
	} {
		lset args $ind -data
		lset args $ind+1 [thatchExtract $fn $fid]
	}
	return [uplevel 1 [list ::tcl::tcl_image {*}$args]]
}

proc thatchLoadOverride {fn args} {
	set lfi -1
	foreach lfn $args {
		incr lfi
		if {[string index $lfn 0] ne "-"} {
			break
		}
	}
	if {$lfi > -1 && [dict size [set fid [thatchGlob $fn $lfn]]] > 0} {
		set c [file tempfile t thatch]
		chan configure $c -translation binary
		puts -nonewline $c [thatchExtract $fn $fid]
		close $c
		lset args $lfi $t
		if {[llength $args] > ($lfi - 1)} {
			lappend args [string totitle [lindex [regexp -inline {^lib([a-zA-z_]+).*} [file tail $lfn]] end]]
		}
		set code [catch {uplevel 1 [list ::tcl::tcl_load {*}$args]} res]
		file delete -force $t
	} else {
		set code [catch {uplevel 1 [list ::tcl::tcl_load {*}$args]} res]
	}
	return -code $code $res
}

proc thatchFileOverride {fn args} {
	set ok 0
	if {[llength $args] == 2} {
		set fcmd [lindex $args 0]
		if {$fcmd in {exists isfile owned readable executable isdirectory writable atime mtime attributes size type}} {
			set ok 1
		}
	}
	if {[llength $args] == 3} {
		set fcmd [lindex $args 0]
		if {$fcmd in {lstat stat}} {
			set ok 1
		}
	}

	if {[lindex $args 0] eq "type" && [lindex $args 1] eq ""} { return "directory" }
	if {$ok && [dict size [set fid [thatchGlob $fn [lindex $args 1]]]] == 1} {
		switch -exact -- [lindex $args 0] {
		exists -
		isfile -
		owned -
		readable {
			return 1
		}
		executable -
		isdirectory -
		writable {
			return 0
		}
		type {
			return "file"
		}
		atime -
		mtime {
			return [dict get $fid mtime]
		}
		attributes {
			return 0o444
		}
		size {
			return $size
		}
		lstat -
		stat {
			upvar 1 [lindex $args 2] a
			unset -nocomplain a
			array set a [list atime [dict get $fid mtime] ctime [dict get $fid mtime] dev -1 gid -1 ino -1 mode 0o444 mtime [dict get $fid mtime] nlink 1 size [dict get $fid size] type "file" uid -1]
			return ""
		}
		default { explode! }
		}
	}
	return [uplevel 1 [list ::tcl::tcl_file {*}$args]]
}

proc thatchGlobOverride {fn args} {
#puts $fn:$args
#if {[catch {info level -1} e]} { set e :: }
#puts $e:$args
	variable thIndex
	array set g {-join 0 -tails 0 -types {} -directory "" -path "" patterns {}}
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -- $arg {
			-join - -tails { set g($arg) 1 }
			-dir - -directory - -path - -types {
				if {$arg eq "-dir"} { set arg "-directory" }
				set state setelem
				set elem $arg
			}
			-nocomplain {}
			-- { set state endopts }
			default { lappend g(patterns) $arg }
			}
		}
		endopts { lappend g(patterns) $arg }
		setelem { set state opt; set g($elem) $arg }
		}
	}
	if {$g(-join)} { set g(patterns) [file join {*}$g(patterns)] }
	set fid [dict create]
	foreach p $g(patterns) {
		set fid [dict merge $fid [thatchGlob $fn [file join $g(-directory) $g(-path) $p]]]
	}
	if {[dict size $fid] > 0} {
		return [dict keys $fid]
	}
	tailcall ::tcl::tcl_glob {*}$args
}

#-------------------------------------------------------------------
# Setup
#-------------------------------------------------------------------

proc overrideAProc {fn proc} {
	if {[llength [info commands ::$proc]]} {
		rename ::$proc ::tcl::tcl_$proc
		interp alias {} ::$proc {} [imn thatch[string totitle $proc]Override] $fn
	} else {
		
	}
}

proc thatchSetup {fn args} {
	set oo {}; foreach o [set op {file glob image load source index packages modules}] { lappend oo -no$o }
	foreach a $args {
		if {[lsearch -exact $oo $a] == -1} {
			error [unknownOption $a $oo]
		}
		set i [lsearch -exact $op [string range $a 3 end]]
		set op [lreplace $op $i $i]
	}
	set createIndex 0
	if {[set i [lsearch -exact $op index]] != -1} {
		set op [lreplace $op $i $i]
		set createIndex 1
	}
	set setupPackages 0
	if {[set i [lsearch -exact $op packages]] != -1} {
		set op [lreplace $op $i $i]
		set setupPackages 1
	}
	set setupModules 0
	if {[set i [lsearch -exact $op modules]] != -1} {
		set op [lreplace $op $i $i]
		set setupModules 1
	}
	foreach p $op { overrideAProc $fn $p }
	if {$createIndex} { thatchCreateIndex $fn }
	if {$setupPackages} {
		dict for {k v} [thatchGlob $fn *pkgIndex.tcl *tclIndex] {
			set ::dir [file dirname $k]
			uplevel #0 [thatchExtract $fn [dict create $k $v]]
		}
		unset -nocomplain ::dir
	}
	if {$setupModules} {
		dict for {k v} [thatchGlob $fn {*-[0-9]*.tm}] {
			if {![regexp -- {^([_[:alpha:]][:_[:alnum:]]*)-([[:digit:]].*)[.]tm$} [file tail $k] -> pn pv]} { continue }
			if {[dict exists $v modbase]} {
				set pn [string map {/ ::} [string trimleft [string range $k [string length [dict get $v modbase]] end-[string length [file tail $k]]] /]]$pn
			}
			package ifneeded $pn $pv "[::list package provide $pn $pv];[::list source -encoding utf-8 $k]"
		}
	}
}

#-------------------------------------------------------------------
# Package Setup
#-------------------------------------------------------------------

proc setup {} {
	package require Tcl 8.5
	aliasconst ^Z \x1a sig "\n# THATCH "
	interp alias {} [imn thatchCreate] {} [imn thatchOp] create
	interp alias {} [imn thatchAdd]    {} [imn thatchOp] add
	thatchDestroyIndex
	namespace export thatch*
	package provide thatch 0.5
}

namespace current}]::setup

# EOF

# THATCH /bin/thatchtool.tcl size 37730 mtime 1662921040
#! /bin/sh
# \
exec tclsh${TCL} "$0" ${1+"$@"}

#
# ThatchTool
#
# (formerly TkThatch)
#
#
# Stuart Cassoff
#
# December 2009, April 2010, Winter 2012, Summer 2012, Spring 2015, Winter 2018, Spring 2018
# Version 0.1
#
# Spring 2018
# Version 0.2
#
# Summer 2018
# Version 0.3
#
# Summer 2022
# Version 0.4
#
# Summer 2022
# Version 0.5
#

[namespace eval thatchtool {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	thatchtool
		title	ThatchTool
		descr	{Thatch Tool}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.5
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2009,2018 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


#---------------------------------------------------------------------
# Info
#---------------------------------------------------------------------

proc readme {} { return [string range {
<Control-q> will close windows or exit.

Home is optional.
Infile is optional.

To add files:
  Right-click in file name box and select a file.
  Erase mtime if mtime is not to be stored.
  Use checkbuttons to select whether to store size
  or enable base64 or zip transforms
  or md5 and sha256 chacksums,
  then click on the Keep button.

Thatchtool will load thatchtool project files (.tp)
from the command line.

"CSize" is Chunk Size.
Chunk Size is size of stored file + Thatch header inside the thatched file.

} 1 end-1]}

proc usage {} {
	return [format {%s%susage: %s [-?] [options] [filename] [...]} [cfg fullname] \n [string tolower [cfg name]]]
}

proc help {} {
	return [subst -nocommands [join {
		{-l  List (default)}
		{-c  Create}
		{-a  Add}
		{-r  Remove}
		{-e  Get}
		{-y  Copy}
		{-h  Home}
		{-H  Home + Thatch file home}
		{-o  fn Output file}
		{-m  fn Main}
		{-wp Wish preamble}
		{-np No preamble}
		{-pp Print preamble}
		{-pb Print bootstrap}
		{-tb Use Thatch bootstrap}
		{-g  Gui mode}
		{-v  Increase verbosity}
		{-?  This info}
	} \n]]
}

proc messages {} {
	return [join [cfg msgs] \n]
}

proc resetMessages {} {
	cfg msgs {}
}

proc log {msg {status 1}} {
	if {[catch {
		cfg+ msgs $msg
		if {[cfg guiCreated]} {
			if {$status} {
				setStatus $msg
			}
		} else {
			puts [cfg stdout] $msg
		}
	} err]} {
		#puts stderr $msg
	}
}

proc setStatus {msg} {}


#---------------------------------------------------------------------
# Misc
#---------------------------------------------------------------------

proc filterNv {data flds} {
	set res {}
	foreach l $data { if {[lindex $l 0] in $flds} { lappend res {*}$l } }
	return $res
}

proc mapCalc {ratio offset size width} {
	set res {}
	set sp [expr {int($offset * $ratio)}]
	set sx [expr {$sp % $width}]
	set sy [expr {$sp / $width}]
	set ep [expr {int(($offset + $size) * $ratio)}]
	set ex [expr {$ep % $width}]
	set ey [expr {$ep / $width}]

	if {$ey == $sy} {
		lappend res [list $sx $sy $ex $ey] {} {}
	} else {
		if {$sx == 0} {
			lappend res {}
		} else {
			lappend res [list $sx $sy [expr {$width + 1}] $sy]
			incr sy
		}
		if {$ey > $sy} {
			lappend res [list 0 $sy [expr {$width + 1}] $ey]
		} else {
			lappend res {}
		}
		if {$ex == $width} {
			lappend res {}
		} else {
			lappend res [list 0 $ey [expr {$ex + 1}] $ey]
		}
	}

	return $res
}

proc hexDump {d {w 8}} {
	set l [string length $d]
	set z {}
	for {set i 0; set ii [expr {$w - 1}]; set addr 0} {$i < $l} {incr i $w; incr ii $w; incr addr $w} {
		binary scan [string range $d $i $ii] @0H*@0a* hex ascii
		regsub -all -- {[^[:graph:] ]} $ascii {.} ascii
		regsub -all -- {..} $hex {& } hex
		lappend z [format "%04X  %-[expr {3 * $w}]s %-${w}s" $addr $hex $ascii]
	}
	return $z
}

proc moveFile {src dst} {
	if {1} {
	file rename -force $src $dst
	} else {
	file copy -force $src $dst
	file delete -force $src
	}
}


#---------------------------------------------------------------------
# Misc GUI
#---------------------------------------------------------------------

proc adjMainGeom {} {
	pjs!
	if {[dict size $pjs] == 0} {
		wm geometry . 300x100
	} elseif {[dict size $pjs] == 1} {
		wm geometry . {}
#		update idletasks
	}
	after idle [list [cfg NS]::centreToplevel .]
}

proc colTextWidth {tv txt} {
	return [expr {[font measure [ttk::style lookup [$tv cget -style] -font] $txt] + 30}]
}

proc adjTv {tv {cols {}}} {
	if {[llength $cols] == 0} {
		set cols [$tv columns]
	}
	foreach col $cols {
		set width [colTextWidth $tv [$tv heading $col -text]]
		foreach item [$tv children {}] {
			set txtwidth [colTextWidth $tv [$tv set $item $col]]
			if {$txtwidth > $width} {
				set width $txtwidth
			}
		}
		$tv column $col -width $width -minwidth $width
	}
}

proc autoscrollbar {w args} {
	set w [ttk::scrollbar $w {*}$args]
	::autoscroll::autoscroll $w
	return $w
}

proc changeWordWrap {} {
	set w [epmW]
	$w configure -wrap [expr {[$w cget -wrap] eq "none" ? "word" : "none"}]
}


#---------------------------------------------------------------------
# Dialog
#---------------------------------------------------------------------

proc oc {title message} {
	set choice [dialog [randWindow] $title $message "" 1 "&Ok" &Cancel"]
	if {$choice == -1} { set choice end }
	return [string index OC $choice]
}

proc yn {title message} {
	set choice [dialog [randWindow] $title $message "" 1 "&Yes" "&No"]
	if {$choice == -1} { set choice end }
	return [string index YN $choice]
}

proc okcancel {title message} {
	return [expr {[oc $title $message] eq "O"}]
}

proc yesno {title message} {
	return [expr {[yn $title $message] eq "Y"}]
}

proc dialog {window title text bitmap default args} {
	set ns [cfg NS]:
	destroyToplevel $window
	newToplevel $window
	groupToplevel $window [gui W]
	setToplevelTitle $window $title
	transientizeToplevel $window [gui W]
	set t [split $text \n]
	set weegs [dict create]
	eval [gub "
		[gubMultiGob [llength $t] L F]
		.{L} -text $t
		.{B} -command [lmap i [::gub::ExpandRange 0-[expr {[llength $args]-1}]] { list $ns:dialogClick $i}]
		;F [string repeat {B } [llength $args]]
		[gubButtonGob -buttonchar B -modifiers Alt {*}$args]
		[expr {$default > -1 ? "/B" : ""}]
	" $window weegs]
	foreach k {Control-q Control-Q Escape} { bind $window <$k> [list $ns:dialogClick -1] }
	onToplevelDelete $window [list $ns:dialogClick -1]
	centreToplevel $window [gui W]
#	wm attributes $window -topmost 1
	grab set $window
	if {$default > -1} { focus [dict get $weegs B$default] }
	gui dialogRetval -1
	vwait [gui& dialogRetval]
	grab release $window
	destroy $window
	return [gui dialogRetval]
}

proc dialogClick {what} {
	gui dialogRetval $what
}


#---------------------------------------------------------------------
# Messages
#---------------------------------------------------------------------

proc clearMessages {t} {
	$t delete 1.0 end
	resetMessages
}

proc viewMessages {} {
	set w .q
	destroy $w
	set w [newToplevel $w]
	groupToplevel $w [gui W]
	set weegs [dict create]
	eval [gub "
		;F t S
		;  C
		;F B B
		.B0 -command {[list after idle [list destroy $w]]}
		.B1 -command {[cfg NS]::clearMessages /t0/}
		:F0,t0
		-.,F0 ; |.,F0
		/t0 viewText
		/B0 okButton
		[gubButtonGob -buttonchar B &Ok &Clear]
		&B0 Control+q
		&B0 Escape
#		/. W
	" $w weegs]
	setToplevelTitle $w [string cat [getToplevelTitle [gui W]] " - Messages"]
	[dict get $weegs viewText] insert end [messages]
	focus [dict get $weegs okButton]
	centreToplevel $w [gui W]
}


#---------------------------------------------------------------------
# Config
#---------------------------------------------------------------------

proc configAdj {} {
	if {[cfg theme] eq ""} { cfg theme default }
	#set cfg(configFile) [file normalize $cfg(configFile)]
}

proc configLoad {{fn {?}}} {
	if {$fn eq {}} { set fn [cfg configFile] }
	if {![lassign [loadListFile $fn -filetypes [list [dict get [cfg filetypes] config]]] fn data]} { return 0 }
	dict for {k v} [filterNv $data [cfg configFileFields]] { cfg $k $v }
	cfg configFile $fn
	configAdj
	return 1
}

proc configSave {{fn {}}} {
	if {$fn eq {}} { set fn [cfg configFile] }
	foreach n [cfg configFileFields] { lappend l [list $n [cfg $n]] }
	if {![lassign [saveFile $fn [join $l \n]\n -filetypes [list [dict get [cfg filetypes] config]]] fn]} { return 0 }
	cfg configFile $fn
	configAdj
	return 1
}

proc autoloadConfig {} {
	configAdj
	#if {[info exists cfg(pConfigFile)]} { set cfg(configFile) $cfg(pConfigFile) }
	configAdj
	if {[cfg autoloadConfigFile]} { configLoad {} }
	configAdj
}

proc cfgThemeChanged {w} {
	ttk::style theme use [$w get]
}

proc configGo {} {
	::aitcs::aitcs new ecfg
	foreach f [cfg configBoolFields] { ecfg $f [cfg $f] }
	set wugs [createConfigGui]
	cfg->gui $wugs
	grab set [dict get $wugs W]
	focus [dict get $wugs B6]
}

proc configStop {wugs} {
	ttk::style theme use [cfg theme]
	grab release [dict get $wugs W]
	after idle [list destroy [dict get $wugs W]]
	ecfg---
}

proc configSaveStop {wugs} {
	foreach f [cfg configBoolFields] { cfg $f [ecfg $f] }
	gui->cfg $wugs
	configSave
	configStop $wugs
}

proc cfg->gui {wugs} {
	foreach p [cfg autoloadableProjects] {
		[dict get $wugs autoloadableProjects] insert {} end -values $p
	}
	[dict get $wugs themeSelector] current [lsearch [ttk::style theme names] [cfg theme]]
}

proc gui->cfg {wugs} {
	cfg autoloadableProjects {}
	foreach item [[dict get $wugs autoloadableProjects] children {}] {
		cfg+ autoloadableProjects [dict values [[dict get $wugs autoloadableProjects] set $item]]
	}
	cfg theme [[dict get $wugs themeSelector] get]
}

proc configAddProject {wugs} {
	set fn [tk_getOpenFile -filetypes [list [dict get [cfg filetypes] tp]] -parent [dict get $wugs W]]
	if {$fn eq ""} { return }
	[dict get $wugs autoloadableProjects] insert {} end -values [list $fn on]
}

proc configRemoveProject {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	[dict get $wugs autoloadableProjects] delete $item
}

proc configMoveProjectUp {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	set index [[dict get $wugs autoloadableProjects] index $item]
	if {$index == 0} { return }
	[dict get $wugs autoloadableProjects] move $item {} [incr index -1]
}

proc configMoveProjectDown {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	set index [[dict get $wugs autoloadableProjects] index $item]
	if {$index == [llength [[dict get $wugs autoloadableProjects] children {}]]} { return }
	[dict get $wugs autoloadableProjects] move $item {} [incr index]
}

proc configToggleProjectAutoload {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	[dict get $wugs autoloadableProjects] set $item autoload [expr {[[dict get $wugs autoloadableProjects] set $item autoload] ? "off" : "on"}]
}

proc createConfigGui {} {
	set w [newToplevel]
	groupToplevel $w [gui W]
	setToplevelTitle $w Configure
	set alpcols {project autoload}
	set weegs [dict create]
	eval [gub "
		;M Vp
		;  FpB B B B B
		;M LaKaLtXt
		;M KsKmKbKzKdKh
		;F B B
		.{M} -text {Autoloadable Projects} Options {Add Files Defaults}
		.Vp -columns {$alpcols} -show headings
		.{La,Lt} -text {Autoload Projects} Theme
		.Xt -state readonly -values {[ttk::style theme names]}
		.{K<smbzdh>} -text Size Mtime Base64 Zip Md5 Sha256
		.{K<asmbzdh>} -variable [lmap f [cfg configBoolFields] { ecfg& $f }]
		:M,V
		:L,K -sticky w
		:X -sticky ew -padx {4 0}
		-.,M0 ; |.,M0
		|M1 1
		[gubButtonGob -buttonchar B -modifiers Alt \
			&Add &Remove {Move &Up} {Move &Down} {&Toggle Autoload} &OK &Cancel \
		]
		/Vp autoloadableProjects
		/Xt themeSelector
		/B
		/. W
	" $w weegs]
	set tv [dict get $weegs autoloadableProjects]
	foreach col $alpcols {
		$tv heading $col -text [string totitle $col]
	}
	$tv column autoload -anchor center -stretch 0
	adjTv $tv $alpcols
	foreach b {0 1 2 3 4 5 6} cmd {configAddProject configRemoveProject configMoveProjectUp configMoveProjectDown configToggleProjectAutoload configSaveStop configStop} {
		[dict get $weegs B$b] configure -command [list [cfg NS]::$cmd $weegs]
	}
	bind [dict get $weegs themeSelector] <<ComboboxSelected>> [list [cfg NS]::cfgThemeChanged [dict get $weegs themeSelector]]
	foreach k {Control-q Control-Q Escape} { bind $w <$k> [list [cfg NS]::configStop $weegs] }
	onToplevelDelete $w [list [cfg NS]::configStop $weegs]
	centreToplevel $w [gui W]
	return $weegs
}


#---------------------------------------------------------------------
# Gui
#---------------------------------------------------------------------

proc createGui {{W {.}}} {
	pjs!;gui!
	set ns [cfg NS]:
	gui w [string trimright [gui W $W] .]
	gubWidgetMap [dict merge [gubWidgetMap] [dict create C [dict merge [dict get [gubWidgetMap] S] [dict create command $ns:autoscrollbar]]]]
	gui M [createMainMenu [gui w]]
	eval [gub -a "N ; : ; -. ; |. ; /N0 mainBook; &. Control+q $ns:fileExit" [gui w]]
	onToplevelDelete [gui W] $ns:fileExit
	catch {ttk::style theme use [cfg theme]}
	#ttk::style configure Treeview -rowheight [expr {[font metrics [ttk::style lookup Treeview -font] -linespace] + 2}]
	wm title [gui W] [cfg name]\ [cfg ver]
	[gui W] configure -menu [gui M]
	gui CVM [createChangeViewMenu [gui w]]
	epmCreateMenu [gui w]
	adjMainGeom
	#foreach i {3 4} { $gui(mainMenu,File) entryconfigure $i -state disabled }
}

proc createMainMenu {w} {
	set M $w.mMain
	set mob "
		.&File
		. &Configure	-command configGo
#		. &Messages	-command viewMessages
		. &Program Info	-command createProgramInfoGui
		. E&xit
		.&Project
		. &Open
		. &Save
		. Save &As	-command {[cfg NS]::projectSave ?}
		. &Close
		. &New
		.&Thatch
		. &Open		-command openThatch
	"
	eval [dict get [meb -menuname $M -namespace [cfg NS] $mob] tcl src]
	return $M
}

proc createChangeViewMenu {w} {
	set M $w.mChangeView
	set mob "
		.&Word Wrap -variable ::ww -command changeWordWrap
	"
	eval [dict get [meb -menuname $M -namespace [cfg NS] $mob] tcl src]
	return $M
}

proc createProgramInfoGui {} {
	set w .pinfo
	destroyToplevel $w
	set w [newToplevel $w]
	groupToplevel $w [gui W]
	setToplevelTitle $w "[cfg fullname] - Info"
	onToplevelDelete $w [list set ::ww $::ww]
	set weegs [dict create]
	eval [gub "
		;N FrtrC
		;    C
		;  FhthC
		;    C
		;  Fa
		;B
		[gubButtonGob -buttonchar B -modifiers Alt &Ok]
		.t -width 76 -height 20 -wrap none
		.B0 -command {after idle destroy /./}
		+{F<rha>} -text &Readme &Help &About
		:N0,t
		|.,F<rha> ; -.,F<rha>
#		&N0 <<NotebookTabChanged>> [cfg NS]::programInfoAdjustChangeViewMenuForText %W
		&B0 Control+q
		&B0 Escape
		/N0 book
		/B0 okbutton
		/Fa about
		/t<rh>
	" $w weegs]
	::ttk::notebook::enableTraversal [dict get $weegs book]
	setTextText [dict get $weegs tr] [readme]
	setTextText [dict get $weegs th] [usage]\n[help]
	::central::extras::CentralPackageBrowser new [dict get $weegs about]
	foreach t [list tr th] {
		[set t [dict get $weegs $t]] configure -state disabled 
		epmBindMenuTo $t [gui CVM]
	}
	focus [dict get $weegs okbutton]
	centreToplevel $w [gui W]
}

proc programInfoAdjustChangeViewMenuForText {w} {
	set ::ww [expr {[[$w select].t[string index [$w select] end] cget -wrap] eq "word"}]
}


#---------------------------------------------------------------------
# Project / Thatch
#---------------------------------------------------------------------

proc autoloadProjects {} {
	if {![cfg autoloadProjects]} { return }
	foreach p [cfg autoloadableProjects] {
		if {[lindex $p 1]} {
			projectOpen [lindex $p 0]
		}
	}
}

proc newProject {} {
	pjs!
	for {set i 0} {[dict exists $pjs [set id pj#$i]]} {incr i} {}
	dict set pjs $id [dict merge [cfg projectFields] [dict create id $id projectFile [string map {# ""} $id].tp et {} view {}]]
	return $id
}

proc addProjectToGUI {id} {
	pjs!
	adjMainGeom
	createProjectGUI [gui mainBook] $id
	[gui mainBook] add [set w [dict get $pjs $id w]] -text [dict get $pjs $id name]
	[gui mainBook] select $w
	return $id
}

proc projectOpen {{fn ""}} {
	pjs!
	if {![lassign [loadListFile $fn -filetypes [list [dict get [cfg filetypes] tp]]] fn data]} { return 0 }
	set id [newProject]
	dict set pjs $id [dict merge [dict get $pjs $id] [filterNv $data [dict keys [cfg projectFields]]]]
	dict set pjs $id projectFile $fn
	adjThatch $id
	addProjectToGUI $id
	project->gui $id
	thatch->gui $id
	return $id
}

proc projectSave {{fn {}}} {
	pjs!
	set id [getPjIdOfCurrentTab]
	gui->project $id
	if {$fn eq {}} {
		set fn [dict get $pjs $id projectFile]
	}
	if {$fn eq {}} {
		set fn ?
	}
	set l {}
	foreach f [cfg projectEditFields] {
		lappend l [list $f [dict get $pjs $id $f]]
	}
	if {![lassign [saveFile $fn [join $l \n]\n -filetypes [list [dict get [cfg filetypes] tp]]] fn]} { return 0 }
	dict set pjs $id projectFile $fn
	project->gui $id
	return 1
}

proc projectClose {} {
	pjs!
	set id [getPjIdOfCurrentTab]
	[gui mainBook] forget [dict get $pjs $id w]
	destroy [dict get $pjs $id w]
	dict unset pjs $id
	adjMainGeom
}

proc openThatch {{fn ""}} {
	pjs!
	if {$fn eq {} || $fn eq {?}} {
		gui!
		lappend filetypes [dict get [cfg filetypes] all]
		set fn [tk_getOpenFile -filetypes $filetypes -parent $gui(W) -title "Open Thatch"]
	}
	if {$fn eq {}} { return }
	set id [newProject]
	dict set pjs $id name $id
	dict set pjs $id outFile $fn
	adjThatch $id
	addProjectToGUI $id
	project->gui $id
	thatch->gui $id
	return $id
}

proc adjThatch {id} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {[file exists $fn]} {
		dict set pjs $id outFileSize [file size $fn]
		dict set pjs $id canvMapRatio [expr {(double([cfg mapCanvWidth] * [cfg mapCanvHeight])) / double([dict get $pjs $id outFileSize])}]
	} else {
		dict set pjs $id outFileSize 0
		dict set pjs $id canvMapRatio 0
	}
}

proc adjLineForDisp {id l} {
	pjs!
	dict set pjs $id mtimeFormat "%Y/%m/%d %T"
	if {[set f [dict get $pjs $id mtimeFormat]] ne ""} {
		set i [lsearch [cfg thatchTvFields] mtime]
		if {[set time [lindex $l $i]] ne ""} {
			lset l $i [clock format $time -format $f]
		}
	}
	return $l
}

proc project->gui {id} {
	pjs!
	foreach f [cfg projectEditFields] {
		[dict get $pjs $id w_$f] configure -validate none
		[dict get $pjs $id w_$f] delete 0 end
		[dict get $pjs $id w_$f] insert end [dict get $pjs $id $f]
		[dict get $pjs $id w_$f] configure -validate key
	}
	setProjectTitle $id [dict get $pjs $id name]
}

proc setProjectTitle {id t} {
	pjs!
	[gui mainBook] tab [dict get $pjs $id w] -text $t
}

proc gui->project {id} {
	pjs!
	foreach f [cfg projectEditFields] {
		dict set pjs $id $f [[dict get $pjs $id w_$f] get]
	}
	setProjectTitle $id [dict get $pjs $id name]
}

proc thatch->gui {id {focus {}}} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {![file exists $fn] } {return }
	set offset [dict get $pjs $id outFileSize]
	set tv [dict get $pjs $id w_tv]
	set c [dict get $pjs $id w_c]
	set tf [lrange [cfg thatchTvFields] 1 end-1]
	set pos [file size $fn]
	set thIndex [::thatch::thatchFullFileGlob $fn]
	dict set pjs $id thIndex $thIndex
	$tv delete [$tv children {}]
	$c delete all
	[dict get $pjs $id w_ename] delete 0 end
	[dict get $pjs $id w_t] delete 1.0 end
	defaults->editThatchedFileInfoGui $id
	dict set pjs $id fishMap [dict create]
	foreach {v k} [lreverse [dict get $pjs $id thIndex]] {
		set l [list $k]
		if {[dict exists $v checksums]} {
			set x [dict get $v checksums]
			set r {}
			foreach {q z} [split $x ,:] {
				lappend r $q
			}
			dict set v checksums $r
		}
		foreach f [list size mtime transforms checksums offset] {
			lappend l [dgwd "" $v $f]
		}
		set csize [expr {$pos - [dict get $pjs $id thIndex $k offset]}]
		dict set pjs $id $fn $k csize $csize
		incr pos -$csize
		lappend l $csize
		dict set pjs $id fishMap [$tv insert {} 0 -values [adjLineForDisp $id $l]] $k
		createMapRegion $id $fn $k $c
	}
	$c itemconfigure all -fill [cfg mapCanvBackground]
	adjTv $tv [cfg thatchTvFields]
	if {$focus ne {}} {
		$tv focus $focus
	}
}

proc thatchedFile->gui {w id {view {}}} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {![file exists $fn]} { return }
	dict set pjs $id view $view
	set sel [lindex [$w selection] 0]
	set name [dict get [dict get $pjs $id thIndex [dict get $pjs $id fishMap $sel]] name]
	set ext [file extension $name]
	set t [dict get $pjs $id w_t]
	set c [dict get $pjs $id w_c]
	thatchedFileInfo->editThatchedFileInfoGui $id
	set d [thatchExtract $fn [dict create $name [dict get $pjs $id thIndex $name]]]
	$t configure -state normal
	image delete {*}[$t image names]
	$t delete 1.0 end
	if {$ext in {".gif" ".png" ".ppm" ".xbm"}} {
		if {$view eq "hex"} {
			$t insert end [join [hexdump $d 32] \n]
			set title Img
		} else {
			$t image create end -image [image create [expr {$ext eq ".xbm" ? "bitmap" : "photo"}] -data $d]
			set title Hex
		}
	} else {
		set xfrms [dgwd {} $pjs $id thIndex $name transforms]
		if {"base64" in $xfrms} {
			if {$view eq {}} {
				set d [join [hexdump $d 32] \n]
			}
			set title [expr {$view eq {} ? "Txt" : "Hex"}]
		} else {
			if {$view eq "hex"} {
				set d [join [hexdump $d 32] \n]
			}
			set title [expr {$view eq {} ? "Hex" : "Txt"}]
		}
		$t insert end $d
	}
	$c itemconfigure all -fill [cfg mapCanvBackground]
	foreach item [$w selection] {
		set tag [dict get $pjs $id thIndex [dict get $pjs $id fishMap $item] name]
		$c itemconfigure $tag -fill [cfg mapCanvForeground]
		$c raise $tag
	}
	$t configure -state disabled
	[dict get $pjs $id w_bchgvw] configure -text [string cat \u2193 $title]
}

proc thatchedFileInfo->editThatchedFileInfoGui {id} {
	pjs!
	set w [dict get $pjs $id w_tv]
	set fi [$w set [lindex [$w selection] 0]]
	foreach tf [list name mtime] gf [list w_ename w_emtime] {
		[dict get $pjs $id $gf] delete 0 end
		[dict get $pjs $id $gf] insert end [dict get $fi $tf]
	}
	set x [dict get $fi transforms]
	foreach tf [list base64 zip] gf [list w_ebase64 w_ezip] {
		[dict get $pjs $id $gf] state [expr {$tf in $x ? "" : "!"}]selected
	}
	set x [dict get $fi checksums]
	foreach tf [list md5 sha256] gf [list w_emd5 w_esha256] {
		[dict get $pjs $id $gf] state [expr {$tf in $x ? "" : "!"}]selected
	}
	foreach tf [list size] gf [list w_esize] {
		set s [expr {[dict get $fi $tf] eq "" || [dict get $fi offset]  == 0 ? "!" : ""}]
		[dict get $pjs $id $gf] state ${s}selected
	}
}

proc defaults->editThatchedFileInfoGui {id} {
	pjs!
	foreach cf [list addDefSize addDefBase64 addDefZip addDefMd5 addDefSha256] gf [list w_esize w_ebase64 w_ezip w_emd5 w_esha256] {
		[dict get $pjs $id $gf] state [expr {[cfg $cf]?"":"!"}]selected
	}
	if {![cfg addDefMtime]} {
		[dict get $pjs $id w_emtime] delete 0 end
	}
}

proc getEditThatchedFileInfoGuiValues {id} {
	pjs!
	set d [dict create]
	foreach f [list name mtime] gf [list w_ename w_emtime] {
		dict set d $f [[dict get $pjs $id $gf] get]
	}
	foreach tf [list size base64 zip md5 sha256] gf [list w_esize w_ebase64 w_ezip w_emd5 w_esha256] {
		dict set d $tf [expr {"selected" in [[dict get $pjs $id $gf] state]}]
	}
	return $d
}

proc projectNew {} {
	pjs!
	set id [newProject]
	dict set pjs $id name $id
	adjThatch $id
	addProjectToGUI $id
	project->gui $id
	thatch->gui $id
	return $id
}

proc getPjIdOfCurrentTab {} {
	pjs!
	set w [[gui mainBook] select]
	dict for {id pj} $pjs {
		if {[dict get $pj w] eq $w} {
			return $id
		}
	}
	return {}
}

proc projectFieldChangedCallback {id} {
	pjs!
	dict set pjs $id et {}
	gui->project $id
}

proc projectFieldChanged {P} {
	pjs!
	set id [getPjIdOfCurrentTab]
	if {[dict get $pjs $id et] eq {}} {
		dict set pjs $id et [after idle [list [cfg NS]::projectFieldChangedCallback $id]]
	}
	return true
}

proc createProjectGUI {W id} {
	pjs!
	set ns [cfg NS]:
	set flds [cfg projectEditFields]
	set w [ttk::frame [randWindow $W.fF_]]
	set gob "
		;F c L E
		[string repeat ";    L E\n" [expr {[llength $flds]-1}]]
		;P F V C
		;    C
		;  F F L E L E K K K K K B B B Bv
		;    F t C
		;      C
		.c0 -width [cfg mapCanvWidth] -height [cfg mapCanvHeight] -background [cfg mapCanvBackground] -highlightthickness 0
		.{L} -text [lmap f $flds {string totitle $f}]
		.{L[expr {[llength $flds]}],[expr {[llength $flds]+1}]} -text Name Mtime
		.E0-[expr {[llength $flds]-1}] -validate key -validatecommand {$ns:projectFieldChanged %P}
		.{K} -text Size Base64 Zip Md5 Sha256
		.V0 -columns {[cfg thatchTvFields]} -show headings
		[gubButtonGob -buttonchar B -modifiers Alt Keep Remove Def \u2193Hex]
		.B0 -command {$ns:alterThatch $id}
		.B1 -command {$ns:removeFileFromThatch $id}
		.B2 -command {$ns:defaults->editThatchedFileInfoGui $id}
		.Bv -command {$ns:changeView $id} -width -5
		.B0-2 -width 0
		.t0 -height 10 -state disabled -wrap none
		:E -sticky ew
		:L -sticky w
		:c0 -rowspan [llength $flds]
		:F0,3,4,P0,V0,t0
		|.,F1,2,4;-F1,4
		|F0 2
		|F3 1
		-.,F0,2 1
		/c0 $id w_c
		/{E} {} [lmap f $flds { string cat $id " " w_ $f } ] {$id w_ename} {$id w_emtime}
		/V0 $id w_tv
		/t0 $id w_t
		/{K} {} {$id w_esize} {$id w_ebase64} {$id w_ezip} {$id w_emd5} {$id w_esha256}
		/Bv $id w_bchgvw
		/. {$id w}
	"
	eval [gub $gob $w pjs]

	set tv [dict get $pjs $id w_tv]
	foreach col [cfg thatchTvFields] {
		$tv heading $col -text [string totitle $col]
		$tv column $col -stretch 0 -anchor e
	}
	$tv column [lindex [cfg thatchTvFields] 0] -stretch 1 -anchor w
	$tv column [lindex [cfg thatchTvFields] end-1] -stretch 0 -anchor w
	adjTv $tv [cfg thatchTvFields]
	bind $tv <<TreeviewSelect>> [list $ns:thatchedFile->gui %W $id]
	foreach f [concat $flds [list ename emtime]] {
		epmBindMenuTo [dict get $pjs $id w_$f] [epmMenu]
	}
	foreach f [list esize ebase64 ezip emd5 esha256] {
		[dict get $pjs $id w_$f] state !alternate
	}
	epmBindMenuTo [dict get $pjs $id w_t] [gui CVM]

	return $id
}

proc createMapRegion {id fn k c} {
	pjs!
	lassign [lrepeat 3 [list]] topLine midRect botLine
	set csize [dict get $pjs $id $fn $k csize]
	incr csize -1
	while {[llength $topLine] < 1 && [llength $midRect] < 1 && [llength $botLine] < 1} {
		lassign [mapCalc [dict get $pjs $id canvMapRatio] [dict get $pjs $id thIndex $k offset] [incr csize] [cfg mapCanvWidth]] topLine midRect botLine
	}
	if {[llength $topLine] > 0} { $c create line $topLine -tag $k }
	if {[llength $botLine] > 0} { $c create line $botLine -tag $k }
	if {[llength $midRect] > 0} { $c create rectangle $midRect -outline {} -tag $k }
}

proc changeView {id} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {![file exists $fn]} { return }
	set w [dict get $pjs $id w_tv]
	set sel [lindex [$w selection] 0]
	if {[llength $sel] == 0} { return }
	set name [dict get [dict get $pjs $id thIndex [dict get $pjs $id fishMap $sel]] name]
	set view [dict get $pjs $id view]
	if {[file extension $name] in {".gif" ".png" ".ppm" ".xbm"}} {
		if {$view eq {}} {
			set view hex
		} else {
			set view {}
		}
		thatchedFile->gui $w $id $view
	} elseif {[file extension $name] in {".tcl" ".tm" ".txt" ".man" ".n" ".html" ".terms"}} {
		if {$view eq {}} {
			set view hex
		} else {
			set view {}
		}
		thatchedFile->gui $w $id $view
	}
}


#---------------------------------------------------------------------
# Thatch manip
#---------------------------------------------------------------------

proc alterThatch {id} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {$fn eq ""} { return }
	set gd [getEditThatchedFileInfoGuiValues $id]
	set name [dict get $gd name]
	if {$name eq ""} { return }
	set edit [dict exists $pjs $id thIndex $name]
	set ok 0
	set opts {}
	if {$edit} {
		set fid [dict get $pjs $id thIndex $name]
		foreach tf [list size] {
			set gv [dict get $gd $tf]
			set ke [dict exists $fid $tf]
			if {$ke == $gv} { continue }
			lappend opts -[expr {$gv ? "" : "no"}]$tf
		}
		foreach tf [list transforms checksums] stfl [list [list base64 zip] [list md5 sha256]] {
			set v [split [dgwd {} $fid $tf]]
			foreach f $stfl {
				set gv [dict get $gd $f]
				set ke [expr {$f in $v}]
				if {$ke != $gv} {
					lappend opts -[expr {$gv ? "" : "no"}]$f
				}
			}
		}
		file tempfile tmpfile thatch
		set ok [thatchCopy $fn $tmpfile {*}[join $opts] $name]
	} else {
		foreach tf [list size base64 zip md5 sha256] {
			set gv [dict get $gd $tf]
			lappend opts -[expr {$gv ? "" : "no"}]$tf
		}
		file tempfile tmpfile thatch
		set ok [thatchAdd $fn $tmpfile {*}[join $opts] $name]
	}
	if {$ok} {
		moveFile $tmpfile $fn
		adjThatch $id
		thatch->gui $id
	}
}

proc removeFileFromThatch {id} {
	pjs!
	set w [dict get $pjs $id w_tv]
	set sel [$w selection]
	if {[llength $sel] == 0} { return }
	set S [expr {[llength $sel] > 1 ? "s" : ""}]
	set msg {}
	set copyArgs {}
	foreach s $sel {
		set name [dict get [$w set $s] name]
		lappend msg $name
		lappend copyArgs -remove $name
	}
	if {![yesno "Remove File$S" [join [linsert $msg 0 "Remove File$S?"] \n]]} { return }
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	file tempfile tmpfile thatch
	if {![thatchCopy $fn $tmpfile {*}[join $copyArgs]]} { return }
	moveFile $tmpfile $fn
	adjThatch $id
	thatch->gui $id
}

proc preamble {} {
	return [::thatch::preamble [cfg wishpreamble]]
}

proc boot {} {
	return [string map [list @VER@ [package provide thatch] @THATCHOPTS@ [cfg thatchbootopts] @MAIN@ [cfg main]] $::thatch::bootstrap]
}

proc cgogo {} {
	switch -exact [cfg action] {
	list {
		set d [thatchGlob [cfg inFile]]
		if {[cfg verbosity] > 1} {
			dict for {k v} $d {
				puts stdout "$k $v"
			}
		} else {
			dict for {k v} $d {
				puts stdout $k
			}
		}
	}
	create {
		if {[cfg thatchboot]} {
			cfg cliFiles [list [cfg inFile] {*}[cfg cliFiles]]
			foreach c [chan pipe] n [list r w] {
				chan configure $c -encoding utf-8 -translation lf
				set $n $c
			}
			cfg inFile @$r
			if {[cfg preamble]} {
				puts $w [preamble]
			}
			puts $w [boot]
			close $w
		}
		thatch[string totitle [cfg action]] [cfg inFile] [cfg outFile] {*}[cfg cliFiles]
	}
	add - copy {
		thatch[string totitle [cfg action]] [cfg inFile] [cfg outFile] {*}[cfg cliFiles]
	}
	get {
#puts " [cfg inFile] [cfg cliFiles]"
		set res {}
		dict for {k v} [thatchGlob [cfg inFile] [cfg cliFiles]] {
			lappend res [thatchExtract [cfg inFile] [dict create $k $v]]
		}
		puts stdout [join $res \n]
	}
	}

	foreach fn [cfg toDel] {
		file delete -force $fn
	}
}


#---------------------------------------------------------------------
# Setup / Startup / Quit
#---------------------------------------------------------------------

proc fileExit {} {
	#exit
	destroy .
}

proc resolveThatch {fileNames} {
	set fns [list]
	foreach fn $fileNames {
		if {$fn ne ":thatch:"} {
			lappend fns $fn
			continue
		}
		if {[dict size [set fid [thatchGlob [info script] [list */thatch.tcl]]]] > 0} {
			if 0 {
			foreach c [chan pipe] n [list r w] {
				chan configure $c -encoding utf-8 -translation lf -buffersize 32768
				set $n $c
			}
			puts -nonewline $w [thatchExtract $::argv0 $fid]
			close $w
			lappend fns -rename thatch.tcl
			set fn @$r
			}
			if 1 {
			set c [file tempfile t thatch]
			chan configure $c -encoding utf-8 -translation lf
			puts -nonewline $c [thatchExtract $::argv0 $fid]
			close $c
			lappend fns -rename thatch.tcl
			set fn $t
			cfg+ toDel $t
			}
		} elseif {[info exists ::env(HOME)]} {
			set tfn $::env(HOME)/tcl/thatch/library/thatch.tcl
			if {[file exists $tfn]} {
				lappend fns -home [file dirname $tfn]
				set fn [file tail $tfn]
			}
		}
		lappend fns $fn
	}
	return $fns
}

proc args {args} {
	set c [dict create]
	if {[llength $args] == 0} { dict set c usage 1; return $c }
	dict set c verbosity [cfg verbosity]
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-l - -c - -a - -r - -e - -y {
				if {[dict exists $c action]} {
					puts stderr "options -lcarey are mutually exclusive"; exit 1
				}
				dict set c action [string map [list -l list -c create -a add -r remove -e get -y copy] $arg]
			}
			-tb { dict set c thatchboot 1 }
			-wp { dict set c wishpreamble 1 }
			-np { dict set c preamble 0 }
			-g  { dict set c gui 1 }
			-v  { dict incr c verbosity }
			-pp { dict set c printpreamble 1 }
			-pb { dict set c printboot 1 }
			-o - -h - -H - -m - -to { set state $arg }
			-?  { dict set c help 1 }
			default {
				if {![dict exists $c inFile]} {
					dict set c inFile $arg
				} else {
					dict lappend c cliFiles $arg
				}
			}
			}
		}
		-o  { set state opt; dict set c outFile $arg }
		-h  { set state opt; set home $arg }
		-H  { set state opt; set Home $arg }
		-m  { set state opt; dict set c main $arg }
		-to { set state opt; dict set c thatchbootopts $arg }
		}; #switch
	}
	if {[dict exists $c action] && [dict exists $c gui] && [dict get $c gui]} {
		puts stderr "options -lcarey and -g are mutually exclusive"; exit 1
	}
	if {[info exists Home]} {
		set home $Home
		if {![dict exists $c cliFiles]} { dict set c cliFiles [list] }
		dict set c cliFiles [linsert [dict get $c cliFiles] 0 -home $home]
	}
	if {[info exists home]} {
		dict set c inFile [file join $home [dict get $c inFile]]
	}
	return $c
}

proc setup {} {
	::aitcs::aitcs+ new cfg
	eval cfg [join [lrange [split {
		NS			[namespace current]
		name			ThatchTool
		ver			0.5
		author			"Stuart Cassoff"
		when			"Summer 2022"
		msgs			{}
		projectFields		[dict create
						name        ""
						projectFile ""
						home        ""
						inFile      ""
						outFile     ""
						lastFileDir ""
						mtimeFormat ""
					]
		projectEditFields	[list
						name
						projectFile
						home
						inFile
						outFile
					]
		thatchFields		[list
						name
						size
						mtime
						base64
						zip
						md5
						sha256
						offset
						csize
					]
		thatchTvFields		[list
						name
						size
						mtime
						transforms
						checksums
						offset
						csize
					]
		mapCanvWidth		40
		mapCanvHeight		60
		mapCanvForeground	pink
		mapCanvBackground	darkgreen
		theme			default
		autoloadableProjects	[list]
		autoloadProjects	0
		addDefSize		0
		addDefMtime		0
		addDefBase64		0
		addDefZip		0
		addDefMd5		0
		addDefSha256		0
		configFile		thatchtool.conf
		autoloadConfigFile	1
		configFileFields	[list
						theme
						autoloadableProjects
						autoloadProjects
						addDefSize
						addDefMtime
						addDefBase64
						addDefZip
						addDefMd5
						addDefSha256
					]
		configBoolFields	[list
						autoloadProjects
						addDefSize
						addDefMtime
						addDefBase64
						addDefZip
						addDefMd5
						addDefSha256
					]
		filetypes 		[dict create
						tp	{{Thatch Project Files} *.tp}
						config	{{Thatch Configuration Files} .conf}
						all	{{All Files} *}
					]
		gui			no
		usage			0
		help			0
		verbosity		1
		action			list
		inFile			""
		outFile			"-"
		cliFiles		{}
		toDel			{}
		printpreamble		0
		printboot		0
		main			""
		thatchboot		0
		thatchbootopts		""
		wishpreamble		0
		preamble		1
	} \n] 1 end-1] " \\\n"]

	cfg fullname [cfg name]\ [cfg ver]

	interp alias {} pjs! {} variable pjs
	pjs!
	set pjs [dict create]
}

proc setupGui {} {
	package require central::extras
	package require gub
	namespace import ::gub::*
	package require meb
	namespace import ::meb::meb
	package require tu::toplevels::tk
	namespace import ::tu::toplevels::tk::*
	package require tu::epm
	namespace import ::tu::epm::*
	package require tu::pretty
	namespace import ::tu::pretty::hexdump
	package require Tk
	package require autoscroll
	::aitcs::aitcs new gui
	createGui
}


#---------------------------------------------------------------------
# Go
#---------------------------------------------------------------------

proc go {} {
	package require Tcl 8.6
	package require thatch
	namespace import ::thatch::*
	package require aitcs
	namespace import ::aitcs::aitcs
	package require tu
	package require tu::file
	namespace import ::tu::*
	namespace import ::tu::file::*
	setup
	cfg {*}[args {*}$::argv]; set ::argv {}; set ::argc 0
#	if {[info exists [cfg file]]} { set cfg($cfg(what)file) $cfg(file) }
	if {[cfg usage]}         { puts [usage]; return }
	if {[cfg help]}          { puts [usage]\n[help]; if {[cfg verbosity] > 1} { puts \nREADME:\n[readme] }; if {[cfg verbosity] > 2} { puts \nLICENSE:\n[join [dict values [LICENSE]] \n]}; return }
	if {[cfg printpreamble]} { puts [preamble] }
	if {[cfg printboot]}     { puts [string map [list @VER@ [package provide thatch]] $::thatch::bootstrap] }
	if {[cfg printpreamble] || [cfg printboot]} { return }
	autoloadConfig
	cfg cliFiles [resolveThatch [cfg cliFiles]]
	if {[cfg gui]} {
		setupGui
		autoloadProjects
		if {[cfg inFile] ne ""} {
			if {[file extension [cfg inFile]] eq ".tp"} {
				projectOpen [cfg inFile]
			} else {
				openThatch [cfg inFile]
			}
		}
	} else {
		cgogo
	}
}

namespace current}]::go

# EOF

# THATCH /pkgs/dsc/aitcs-2.0b1.tm size 9704 mtime 1660225868 modbase /pkgs/dsc
#
# AitCS - Array in the Command's Shadow
#
# Stuart Cassoff
#
#
# Spring 2018
# Version 1.1
#
# Winter 2020
# Version 1.2
#
# Spring 2020
# Version 1.3, 1.4, 1.5, 1.6, 1.7, 1.8
#
# Fall 2020
# Version 1.9
# Version 2.0b1
#


if {$::tcl_platform(engine) eq "Tcl"} { package require Tcl 8.6 }

namespace eval aitcs {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	aitcs
		title	AitCS
		descr	{Array in the Command's Shadow}
		author	{Stuart Cassoff}
		when	{Fall 2020}
		version	2.0b1
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2020 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}


## Internal

if {$::tcl_platform(engine) eq "Jim"} {
	proc Alias {this that args} {        alias    $this    $that {*}$args }
} else {
	proc Alias {this that args} { interp alias {} $this {} $that {*}$args }
}

proc Randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

proc Randvar {name variable_namespace} {
	return [string trimright $variable_namespace :]::__${name}__[Randstr]_
}

proc MakeCommands {type name varname cns} {
	set jim [expr {$::tcl_platform(engine) eq "Jim"}]

	set cns [string trimright $cns :]
	set smae [expr {[namespace qualifiers $varname] eq $cns}]
	set cns $cns:

	set ops   [list ""   ?       -      ^    &    ->    ]
	set procs [list Vset Vexists Vunset Vgwd Vref Vaitcs]

	if {$type eq "eagle"} {
		lappend ops   =      +        +=
		lappend procs Vsetto Vlappend Vincr
	}

	if {$jim} {
		set hows [lrepeat [llength $ops] full]
	} else {
		if {$smae} {
			set hows [lrepeat [llength $ops] var]
		} else {
			set hows [lrepeat [llength $ops] full]
			set pl [list Vexists Vgwd]
			if {$type eq "eagle"} {
				lappend pl Vsetto
			}
			foreach p $pl {
				lset hows [lsearch $procs $p] partial
			}
		}
		lset hows [lsearch $procs Vref] full
	}

	foreach op $ops proc $procs how $hows {
		set pargs [info args $proc]
		if {!$jim} {
			set pargs [lmap arg $pargs {
				expr {[info default $proc $arg def] ? [list $arg $def] : $arg}
			}]
		}
		set body [info body $proc]
		if {[lindex $pargs 0] eq "varname"} {
			set pargs [lrange $pargs 1 end]
			# For itiming tests, consult Makefile.aitcs.in when modifying this line: set how ?
			set body [switch -exact -- $how {
				var { # Change "upvar ..." to "variable ..." and hardcode [namespace tail $varname] everywhere
					set vn [namespace tail $varname]
					string map [list {upvar 1 $varname _v_} "variable $vn" _v_ $vn {$varname} $vn {${varname}} $vn] $body
				}
				full { # Remove "upvar ..." and hardcode $varname everywhere
					string map [list {upvar 1 $varname _v_} {} _v_ $varname {$varname} $varname {${varname}} $varname] $body
				}
				partial { # Hardcode $varname where used
					string map [list {$varname} $varname {${varname}} $varname] $body
				}
				setvar { # Add "set varname $varname" to top of proc
					format {set varname {%s}%s%s} $varname \n $body
				}
				alias { # Use aliases
					if {$proc eq "Vaitcs"} {
						if {[llength [info commands Vaitcs_alias]] == 0} {
							proc Vaitcs_alias {varname elem args} [string map [list \
								{tailcall $_v_($elem)$op {*}$args} \
								{uplevel 1 $_v_($elem)$op $args}] $body]
						}
						#set proc [string cat $proc _alias]
						append proc _alias
					}
					Alias $cns:$name$op [namespace current]::$proc $varname
					continue
				}
			}]
		}

		# Debug
		#puts "$name$op {$pargs} {$body\n}"

		proc $cns:$name$op $pargs $body

		# Debug
		#puts $name$op\n[::tcl::unsupported::disassemble proc $cns:$name$op]
	}

	# Debug
	#exit

	Alias $cns:$name! upvar #0 $varname $name

	# For itiming tests, consult Makefile.aitcs.in when modifying these 3 lines
	# Alias $cns:$name&&    return -level 0 $cns:$name
	proc    $cns:$name&& {} [list return $cns:$name]

	set --- [join [list {} \
		"unset $varname" \
		{*}[lmap op [list {*}$ops ! && ---] {
			format {rename %s%-3s {}} $cns:$name $op
		}] \
	] \n\t]\n

	# For itiming tests, consult Makefile.aitcs.in when modifying these 3 lines
	# Alias $cns:$name--- eval ${---}
	proc    $cns:$name--- {}   ${---}
}

proc New {type args} {
	set name ""
	set cns ""
	set vns ""
	set argies [list]
	set nameset no
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-cns - -command_namespace - -vns - -variable_namespace { set state $arg }
			-- { set state -- }
			default { if {$nameset} { lappend argies $arg } else { set name $arg; set nameset yes } }
			}; # hctiws
		}
		-cns - -command_namespace  { set state opt; set cns $arg }
		-vns - -variable_namespace { set state opt; set vns $arg }
		-- { lappend argies $arg }
		default { error "Huh? <[expr {int(rand()*100000)}]>" }
		}; # hctiws
	}
	if {$state ni [list opt --]} { return -code error "no argument given for option \"$state\"" }
	if {[expr {[llength $argies] % 2}] != 0} { return -code error "initial values must be \"key value ...\"" }
	if {$vns eq ""} { set vns [uplevel 1 {namespace current}] }
	if {$cns eq ""} { set cns $vns }
	set vns [string trimright $vns :]
	set cns [string trimright $cns :]
	if {$name eq ""} { for {*}{
		{ set name [Randstr]; set i 0 }
		{ $i < 1000 && [llength [info commands ${cns}::$name]] > 0 }
		{ set name [Randstr]; incr i }
		{}
	}}
	if {[llength [info commands ${cns}::$name]] > 0} { return -code error "${cns}::$name already exists" }
	set varname [Randvar $name $vns]
	MakeCommands $type $name $varname $cns
	array set $varname $argies
	return ${cns}::$name
}


## Operators

# AitCS / AitCS+

#
proc Vset {varname elem args} {
	upvar 1 $varname _v_
	if {[llength $args] == 0} {
		set _v_($elem)
	} elseif {[llength $args] == 1} {
		set _v_($elem) [lindex $args 0]
	} else {
		array set _v_ [list $elem {*}$args]
		set _v_($elem)
	}
}

# -
proc Vunset {varname elem args} {
	upvar 1 $varname _v_
	foreach e $args { array unset _v_ $e }
	array unset _v_ $elem
}

# &
proc Vref {varname {elem ""}} {
	if {$elem eq ""} { return $varname }
	return ${varname}($elem)
}

# ?
proc Vexists {varname elem {myVar ""}} {
	upvar 1 $varname _v_
	if {![info exists _v_($elem)]} { return 0 }
	if {$myVar eq {}} { return 1 }
	upvar 1 $myVar var
	set var $_v_($elem)
	return 1
}

# ^
proc Vgwd {varname elem {default ""} {myVar ""}} {
	upvar 1 $varname _v_
	if {$myVar ne ""} { upvar 1 $myVar var }
	set var [expr {[info exists _v_($elem)] ? $_v_($elem) : $default}]
}

# ->
proc Vaitcs {varname elem args} {
	upvar 1 $varname _v_
	# Avoid mismatches by checking double-char operators before single-char operators
	if {[set op [string range $elem end-1 end]] in [list += && ->]} {
		set elem [string range $elem 0 end-2]
	} elseif {[set op [string index $elem end]] in [list ? - ^ & = + ! :]} {
		set elem [string range $elem 0 end-1]
	} elseif {[set op [string range $elem end-2 end]] eq "---"} {
		set elem [string range $elem 0 end-3]
	} else {
		set op ""
	}
	tailcall $_v_($elem)$op {*}$args
}

# AitCS+

# =
proc Vsetto {varname elem elemOrVal args} {
	upvar 1 $varname _v_
	set a [list $elem $elemOrVal {*}$args]
	set val [lindex $a end]
	foreach e [lrange $a 0 end-1] { set _v_($e) $val }
	return $val
}

# +
proc Vlappend {varname elem args} {
	upvar 1 $varname _v_
	lappend _v_($elem) {*}$args
}

# +=
proc Vincr {varname elem {incr {1}} args} {
	upvar 1 $varname _v_
	foreach i $args { incr _v_($elem) $i }
	incr _v_($elem) $incr
}


## Setup

if {$::tcl_platform(engine) eq "Jim"} {
	foreach plus [list "" +] type [list snake eagle] {
		proc aitcs$plus {cmd args} [string map [list @plus@ $plus @type@ $type] {
			if {$cmd eq "operators"} {
				tailcall ::aitcs::$cmd@plus@ {*}$args
			} elseif {$cmd eq "new"} {
				tailcall ::aitcs::New @type@ {*}$args
			} else {
				return -code error "unknown subcommand \"$cmd\": must be new, or operators"
			}
		}]
	}
} else {
	foreach plus [list "" +] type [list snake eagle] {
		namespace ensemble create -prefixes no \
			-command aitcs$plus \
			-map [dict create new [list New $type] operators operators$plus]
	}
}


# These things
proc operators  {} { return [list "" ? - ^ & -> ! && ---] }
proc operators+ {} { return [list {*}[operators] = + +=] }


namespace export aitcs aitcs+

package provide aitcs 2.0b1


}; # namespace eval aitcs


# EOF

# THATCH /pkgs/dsc/tu-0.3.tm size 6680 mtime 1662628341 modbase /pkgs/dsc
#
# Tu - Tcl Usefulness
#
#
# Stuart Cassoff
#
# Fall/Winter 2018
# Version 0.1
#
# Summer 2020
# Version 0.2
#
# Summer 2022
# Version 0.3
#

package require Tcl 8.5

namespace eval tu {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu
		title	Tu
		descr	{Tcl Usefulness}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.3
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2017,2018,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


# Tcl

proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

# dict get with default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

# In My Namespace
#
proc imn {what} {
	return [string trimright [uplevel 1 {namespace current}] :]::$what
}

#
# String Cat
#
proc strcat {args} { join $args "" }

#
# [lmap] replacement if not available
#
if {[llength [info commands lmap]] == 0} { proc lmap args {
	# 8.5 doesn't have lmap
	# https://wiki.tcl-lang.org/page/lmap+forward+compatibility
	set body [lindex $args end]
	set args [lrange $args 0 end-1]
	set n 0
	set pairs [list]
	foreach {varnames listval} $args {
		set varlist [list]
		foreach varname $varnames {
			upvar 1 $varname var$n
			lappend varlist var$n
			incr n
		}
		lappend pairs $varlist $listval
	}
	set temp [list]
	foreach {*}$pairs {
		lappend temp [uplevel 1 $body]
	}
	set temp
}}; # lmap

proc getNextOrPrevMatch {aList aString {curIdx 0} {backwards no}} {
	set idxs [lsearch -all -nocase $aList $aString*]
	if {[llength $idxs] == 0} { return -1 }
	if {[llength $idxs] == 1 && [lindex $idxs 0] == $curIdx} { return -1 }
	set p -1
	foreach idx $idxs {
		if {$idx < $curIdx} { set p $idx }
		if {$idx > $curIdx} { break }
	}
	if {$backwards} {
		if {$p == -1} {
			set newIdx [lindex $idxs end]
		} else {
			set newIdx $p
		}	
	} else {
		if {$idx == $curIdx} {
			set newIdx [lindex $idxs 0]
		} else {
			set newIdx $idx
		}
	}
	return $newIdx
}

proc makeRWAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " args " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			if {[llength $args] == 1} { set @m@ [lindex $args 0] }
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}
proc makeROAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " "{}" " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}

#
# Remove blank lines and comment lines, add space-backslash to all lines but the last.
# If a line contains ";#" (spaces ok too) then chares from those to the end of the line will be removed
# Useful for making big lists in code without all the unpleasant backslashes.
#
proc backslashify {blob} { return [join [lmap l [split $blob \n] { if {[regexp {^[[:space:]]*(;?#.*)?$} $l]} { continue } else { lindex [regexp -inline -- {^(.+?)(;\s*#.*)?$} $l] 1 } }] " \\\n"] }

proc maxstrlen {args} { return [::tcl::mathfunc::max 0 {*}[lmap s $args { string length $s }]] }

proc oneOf {l} {
	if {[llength $l] < 2} { return [join $l] }
	return [strcat "one of " [join [lrange $l 0 end-1] ", "] [expr {[llength $l] > 2 ? "," : ""}] " or " [lindex $l end]]
}


## Tk

proc bindings {tag} {
	set d [dict create]
	set l [expr {[string index $tag 0] eq "." ? [bindtags $tag] : [list $tag]}]
	foreach t $l { foreach b [bind $t] {
		dict set d $t $b [bind $t $b]
	}}
	return $d
}

proc setEntryText {w text} {
	$w delete 0 end
	if {$text ne ""} { $w insert end $text }
}

proc setTextText {w text} {
	if {[set state [$w cget -state]] ne "normal"} { $w configure -state normal }
	$w delete 1.0 end
	if {$text ne ""} { $w insert end $text }
	#if {$text ne ""} { $w insert end $text[expr {[string index $text end] eq "\n" ? "" : "\n"}] }
	if {$state ne "normal"} { $w configure -state $state }
}

# From the Tk widget demo "A multi-column list of countries"
# Code to do the sorting of the tree contents when clicked on
#
proc tvSortBy {tree col direction} {
	# Build something we can sort
	set data {}
	foreach row [$tree children {}] {
		lappend data [list [$tree set $row $col] $row]
	}

	set dir [expr {$direction ? "-decreasing" : "-increasing"}]
	set r -1

	# Now reshuffle the rows into the sorted order
	foreach info [lsort -dictionary -index 0 $dir $data] {
		$tree move [lindex $info 1] {} [incr r]
	}

	# Switch the heading so that it will sort in the opposite direction
	$tree heading $col -command [list [namespace current]::tvSortBy $tree $col [expr {!$direction}]]
}

#
# Underline Ampersand
# Convert ampersand in text to -underline option
# From Gub
#
proc ua {text} {	
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}


## Setup

namespace export makeRWAccessors makeRWOccessors
namespace export randstr dgwd imn backslashify maxstrlen oneOf
namespace export tvSortBy setEntryText setTextText ua bindings

package provide tu 0.3


}; # namespace eval tu

# EOF

# THATCH /pkgs/dsc/tu/epm-0.1.tm size 9770 mtime 1662579078 modbase /pkgs/dsc
#
# Tu Epm - Tcl Usefulness Edit Program Menu
#
# Keeps track of two menus that can be used by a program:
#   An epm menu with items for choosing files etc., and edit items.
#   An epm edit menu with only edit items.
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#

package require Tcl 8.5

namespace eval tu { namespace eval epm {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::epm
		title	{Tu Epm}
		descr	{Edit Program Menu}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


variable M {} ;# The epm menu
variable Mmap [dict create] ;# Named items and their indexes
variable E {} ;# The epm edit menu
variable Emap [dict create] ;# Named items and their indexes
# The 'mu' for E
variable epmmu [dict create -win {} -mmap {} -defs {} -ftmap {} -inserter {} -getwrap {} -setwrap {}]
variable epmww ;# Current word wrap state


# Create an epm menu
proc epmCreateMenu {{w ""} args} {
	variable M
	variable Mmap
	set M [menu $w.epm -tearoff 0 -postcommand [namespace current]::epmAdjustMenu {*}$args]
	foreach item [list "Choose &File" "Choose &Directory" "Choose C&olour" "&Set To Default" Cu&t &Copy &Paste C&lear "Select &All"] {
		$M add command -command [namespace current]::epm[string map {{ } {}} [string map {"&" ""} $item]] -label {*}[ua $item]
	}
	$M add checkbutton -command [namespace current]::epmAdjustWordWrap -variable [namespace current]::epmww -onvalue word -offvalue none -label {*}[ua "&Word Wrap"]
	$M insert 4 separator
	set Mmap [dict create cut 5 copy 6 paste 7 clear 8 wrap 10 setdef 3]
	return $M
}

# Return the epm menu, create if doesn't exist
proc epmMenu {} {
	variable M
	if {$M eq {}} { epmCreateMenu }
	return $M
}

# Create an edit menu
proc epmCreateAnEditMenu {{w ""} args} {
	set e [menu $w.eepm_[randstr] -tearoff 0 {*}$args]
	foreach item [list Cu&t &Copy &Paste C&lear "Select &All" "Sa&ve As"] {
		$e add command -command [namespace current]::epm[string map {{ } {}} [string map {"&" ""} $item]] -label {*}[ua $item]
	}
	$e add checkbutton -command [namespace current]::epmAdjustWordWrap -variable [namespace current]::epmww -onvalue word -offvalue none -label {*}[ua "&Word Wrap"]
	set emap [dict create cut 0 copy 1 paste 2 clear 3 saveas 5 wrap 6]
	epmSetEpmAdjustMenu $e $emap
	return [list $e $emap]
}

# Setup epm adjustment for a menu
proc epmSetEpmAdjustMenu {m mmap} {
	$m configure -postcommand [list [namespace current]::epmAdjustMenu $m $mmap]
}

# Create the epm edit menu
proc epmCreateEditMenu {{w ""}} {
	variable E
	variable Emap
	lassign [epmCreateAnEditMenu $w] E Emap
	return $E
}

# Return the epm edit menu, create if doesn't exist
proc epmEditMenu {} {
	variable E
	if {$E eq {}} { epmCreateEditMenu }
	return $E
}

# Have the epm menu popup for a widget
proc epmBindEpmMenuTo {w args} {
	variable Mmap
	# Call epmMenu to ensure M's menu exists
	tailcall epmBindMenuTo $w [epmMenu] {*}$args -mmap $Mmap
}

# Have the epm edit menu popup for a widget
proc epmBindEditMenuTo {w args} {
	variable Emap
	# Call epmEditMenu to ensure E's menu exists
	tailcall epmBindMenuTo $w [epmEditMenu] {*}$args -mmap $Emap
}

# Make a 'mu'
proc epmMakeMu {m args} {
	set ns [namespace current]:
	set mu [dict merge $args [dict create -menu $m \
		-mmap     [dgwd [dict create]      $args -mmap]     \
		-defs     [dgwd [dict create]      $args -defs]     \
		-ftmap    [dgwd [dict create]      $args -ftmap]    \
		-inserter [dgwd $ns:epmInserter    $args -inserter] \
		-getwrap  [dgwd $ns:epmGetWordWrap $args -getwrap]  \
		-setwrap  [dgwd $ns:epmSetWordWrap $args -setwrap]  \
	]]
	return $mu
}

# Have a menu popup for a widget
proc epmBindMenuTo {w m args} {
	set mu [epmMakeMu $m {*}$args]
	set mod   [dgwd ""        $args -mod]
	set seq   [dgwd "Button-" $args -seq]
	set bn    [dgwd 3         $args -bn]
	bind $w <$mod$seq$bn> [list [namespace current]::epmMenuGo %X %Y %W $mu]
}

# Popup activator
proc epmMenuGo {x y w mu} {
	variable epmmu [dict merge $mu [dict create -win $w -x $x -y $y]]
	tk_popup [dict get $epmmu -menu] $x $y
}

# Return the 'mu'
proc epmMu {} {
	variable epmmu
	return $epmmu
}

# Current window for a popup
proc epmW {args} {
	variable epmmu
	if {[llength $args] == 1} {
		dict set epmmu -win [lindex $args 0]
	}
	return [dict get $epmmu -win]
}

# Get the word wrap state for a window. Used by the default 'mu'.
proc epmGetWordWrap {w} {
	return [$w cget -wrap]
}

# Get the word wrap state for a window. Used by the default 'mu'.
proc epmSetWordWrap {w wrap} {
	$w configure -wrap $wrap
}

# Adjust a window's word wrap state according to epm state
proc epmAdjustWordWrap {} {
	variable epmmu
	variable epmww
	[dict get $epmmu -setwrap] [dict get $epmmu -win] $epmww
}

# Ajdust a menu according to a window's state
proc epmAdjustMenu {{m {}} {mmap {}}} {
	variable epmmu
	variable epmww
	set w [dict get $epmmu -win]
	if {$w eq "" || ![winfo exists $w]} {
		if {![catch {info level -1} res]} {
			::tsm::log "epmAdjustMenu !$res!" error
		}
		return
	}
	set defs [dict get $epmmu -defs]
	if {$m eq {}} {
		set m [dict get $epmmu -menu]
	}
	if {$mmap eq {}} {
		set mmap [dict get $epmmu -mmap]
	}
	set items [list]
	if {[dict size $mmap] > 0} { foreach what [list cut copy paste clear wrap setdef saveas] {
		if {![dict exists $mmap $what]} { continue }
		lappend items $what
	}}
	if {[llength $items] == 0} { return }
	set states [dict create]
	dict set states setdef [expr {[dict exists $defs $w] ? "normal" : "disabled"}]
	set wclass [winfo class $w]
	if {$wclass eq "Text"} {
		set epmww [[dict get $epmmu -getwrap] $w]
		if {$epmww eq "char"} { set epmww word }
		dict set states wrap normal
		set sel [expr {[llength [$w tag ranges sel]] > 0 ? "normal" : "disabled"}]
	} elseif {$wclass eq "TLabel"} {
		set epmww none
		dict set states wrap disabled
		set sel normal
	} else {
		set epmww none
		dict set states wrap disabled
		set sel [expr {[$w selection present] ? "normal" : "disabled"}]
	}
	dict set states copy $sel
	dict set states saveas $sel
	if {[$w cget -state] eq "normal"} {
		foreach item [list cut clear] {
			dict set states $item $sel
		}
		dict set states paste normal
	} else {
		foreach item [list cut clear paste setdef saeveas] {
			dict set states $item disabled
		}
	}
	if {[dict get $states paste] eq "normal" && "paste" in $items} {
		if {[catch {selection get -displayof $w -selection CLIPBOARD} s]} { set s "" }
		dict set states paste [expr {[string length $s] > 0 ? "normal" : "disabled"}]
	}
	foreach item $items {
		$m entryconfigure [dict get $mmap $item] -state [dict get $states $item]
	}
}

proc epmInserter {where what} {
	epmSelectAll ; epmClear
	$where insert end $what
}

proc epmChooseFile {} {
	variable epmmu
	set w [dict get $epmmu -win]
	set ftmap [dict get $epmmu -ftmap]
	set filetypes [dgwd {} $ftmap $w]
	if {[lsearch -exact -index {0 1} $filetypes [list *]] == -1} {
		lappend filetypes [list "All Files" [list *]]
	}
	if {[set v [::tu::file::loadFileDialog -initialdir [file dirname [$w get]] -initialfile [$w get] -filetypes $filetypes]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}

proc epmChooseDirectory {} {
	variable epmmu
	set w [dict get $epmmu -win]
	if {[set v [::tu::file::chooseDirectoryDialog -initialdir [$w get]]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}

proc epmChooseColour {} {
	variable epmmu
	set w [dict get $epmmu -win]
	if {[winfo class $w] eq "Text"} {
		set v ""
	} else {
		set v [string trim [$w get]]
	}
	if {$v eq ""} { set v #000000 }
	if {[catch {winfo rgb . $v}]} { set v olive }
	if {[set v [tk_chooseColor -initialcolor $v]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}

proc epmSaveAs {} {
	variable epmmu
	::tu::file::saveSelectionToFile [dict get $epmmu -win]
}

proc epmSetToDefault {} {
	variable epmmu
	set w [dict get $epmmu -win]
	set defs [dict get $epmmu -defs]
	if {![dict exists $defs $w]} { return }
	set v [dict get $defs $w]
	[dict get $epmmu -inserter] $w $v
}

foreach e [list Cut Copy Paste Clear SelectAll] {
	proc epm$e {} [string map [list @e@ $e] {
		variable epmmu
		set w [dict get $epmmu -win]
		event generate $w <<@e@>>
		focus $w
	}]
}; unset e


## Setup

package require tu

namespace path [namespace parent]

namespace export epmCreateMenu epmMenu
namespace export epmCreateEditMenu epmCreateAnEditMenu epmEditMenu
namespace export epmBindMenuTo epmBindEpmMenuTo epmBindEditMenuTo
namespace export epmMakeMu epmSetEpmAdjustMenu
namespace export epmMu epmW

package provide tu::epm 0.1

}}; # namespace tu::epm


# EOF

# THATCH /pkgs/dsc/tu/file-0.1.tm size 5076 mtime 1662579104 modbase /pkgs/dsc
#
# Tu File - Tcl Usefulness File Utils
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#

package require Tcl 8.5

namespace eval tu { namespace eval file {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::file
		title	{Tu File}
		descr	{File Utils}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


proc getFileFor {what {fn ""} args} {
	if {$fn ni {"" "?"}} { return $fn }
	if {[package provide Tk] eq ""} { return "" }
	return [${what}FileDialog {*}$args]
}

proc getLoadFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor load $fn {*}$args]
}

proc getSaveFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor save $fn {*}$args]
}

proc openFile {fn {mode r}} {
	if {$fn eq "-"} { return [expr {$mode eq "r" ? "stdin" : "stdout"}] }
	if {$fn in {stdin stdout stderr}} { return $fn }
	if {[catch {open $fn $mode} f]} { return "" }
	return $f
}

proc closeFile {f} {
	if {$f in {stdin stdout stderr}} { return }
	close $f
}

proc loadFile {{fn ""} args} {
	set fn [getLoadFile $fn {*}$args]
	if {$fn eq ""} { return [list "" "" 0] }
	if {[set f [openFile $fn]] eq ""} { return [list "" "" 0] }
	set data [read -nonewline $f]
	closeFile $f
	return [list $fn $data 1]
}

proc saveFile {fn data args} {
	set fn [getSaveFile $fn {*}$args]
	if {$fn eq ""} { return [list "" 0] }
	if {[set f [openFile $fn w]] eq ""} { return [list "" "" 0] }
	puts -nonewline $f $data
	closeFile $f
	return [list $fn 1]
}

proc loadListFile {{fn ""} args} {
	if {![lassign [loadFile $fn {*}$args] fn data]} { return [list $fn "" 0] }
	return [list $fn [processListFileBlob $data] 1]
}

proc processListFileBlob {blob} {
	set res [list]
	foreach l [split $blob \n] {
		set l [string trim $l]
		if {$l eq {} || [string index $l 0] eq {#}} { continue }
		if {[catch {set len [llength $l]}] || $len < 1} { continue }
		lappend res $l
	}
	return $res
}

proc filemode->text {mode} {
    foreach { mask pairs } {
	00400 { 00400 r }
	00200 { 00200 w }
	04100 { 04100 s 04000 S 00100 x }
	00040 { 00040 r }
	00020 { 00020 w }
	02010 { 02010 s 02000 S 00010 x }
	00004 { 00004 r }
	00002 { 00002 w }
	01001 { 01001 t 01000 T 00001 x }
    } {
	set value [expr $mode & $mask]
	set bit -
	foreach { x b } $pairs {
	    if { $value == $x } {
		set bit $b
	    }
	}
	append retval $bit
    }
    return $retval
}


## Tk

proc loadFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getOpenFile {*}$args
	} else {
		tailcall ttk::getOpenFile {*}$args
	}
}

proc saveFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getSaveFile {*}$args
	} else {
		tailcall ttk::getSaveFile {*}$args
	}
}

proc chooseDirectoryDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_chooseDirectory {*}$args
	} else {
		tailcall ttk::chooseDirectory {*}$args
	}
}

proc loadFileIntoText {w {fn ""} args} {
	lassign [loadFile $fn {*}$args] fn data
	if {$fn eq ""} { return "" }
	$w delete 1.0 end
	$w insert end $data
	return $fn
}

proc saveSelectionToFile {w {fn ""} args} {
	if {[llength [set sel [$w tag ranges sel]]] != 2} { return }
	set blob [string trim [$w get {*}$sel]]
	if {$blob eq ""} { return }
	saveFile $fn $blob {*}$args
	return $fn
}


## Setup

package require tu

namespace path [namespace parent]

namespace export openFile closeFile getLoadFile loadFile getSaveFile saveFile
namespace export loadListFile processListFileBlob
namespace export loadFileDialog saveFileDialog
namespace export loadFileIntoText saveSelectionToFile
namespace export filemode->text

package provide tu::file 0.1

}}; # namespace tu::file


# EOF

# THATCH /pkgs/dsc/tu/pretty-0.1.tm size 3295 mtime 1662630085 modbase /pkgs/dsc
#
# Tu Pretty - Tcl Usefulness Pretty Printing
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#

package require Tcl 8.5

namespace eval tu { namespace eval pretty {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::pretty
		title	{Tu Pretty}
		descr	{Pretty Printing}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


# Tcl

proc hexdump {blob {bits 8}} {
	set l [string length $blob]
	set dump {}
	for {set i 0; set ii [expr {$bits - 1}]; set addr 0} {$i < $l} {incr i $bits; incr ii $bits; incr addr $bits} {
		binary scan [string range $blob $i $ii] @0H*@0a* hex ascii
		regsub -all -- {[^[:graph:] ]} $ascii {.} ascii
		regsub -all -- {..} $hex {& } hex
		lappend dump [format "%04X  %-[expr {3 * $bits}]s %-${bits}s" $addr $hex $ascii]
	}
	return $dump
}

proc prettyBindings {bindings} {
	set q ""
	dict for {k v} $bindings {
		append q $k \n
		dict for {b s} $v {
			append q "  " [format "%-30s" $b] " " $s \n
		}
	}
	return $q
}

proc mpd {dict {pat {*}}} {
	set f %-[::tcl::mathfunc::max 0 {*}[lmap k [set keys [lsort -dictionary [dict keys $dict $pat]]] { string length $k }]]s\ \ %s
	return [lmap k $keys { format $f $k [dict get $dict $k] }]
}
interp alias {} makePrintDict {} mpd

proc pd {dict {pat {*}}} {
	return [join [mpd $dict $pat] \n]
}
interp alias {} printDict {} pd

proc pa {arr {pat {*}}} {
	return [pd [array get $arr] $pat]
}
interp alias {} printArray {} pa

proc pv {var {pat {*}}} {
	upvar 1 $var v
	if {[array exists v]} { return [pa $var $pat] }
	if {![catch {dict size $v}]} { return [pd $v $pat] }
	return $v
}
interp alias {} printVar {} pv


## Tk
proc pb {tag} { return [prettyBindings [bindings $tag]] }
interp alias {} prettyBindingsOfTag {} pb


## Setup

package require tu

namespace path [namespace parent]

namespace export hexdump prettyBindings
namespace export mpd pv pa pd pb
namespace export makePrintDict printVar printArray printDict
namespace export prettyBindingsOfTag

package provide tu::pretty 0.1

}}; # namespace tu::pretty


# EOF

# THATCH /pkgs/dsc/tu/toplevels-0.1.tm size 1885 mtime 1662630168 modbase /pkgs/dsc
#
# Tu Toplevels - Tcl Usefulness Toplevels
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#

package require Tcl 8.5

namespace eval tu { namespace eval toplevels {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::toplevels
		title	{Tu Toplevels}
		descr	{Toplevels}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}


## Setup

namespace export randstr

package provide toplevels 0.1

}}; # namespace tu::toplevels


# EOF

# THATCH /pkgs/dsc/tu/toplevels/tk-0.1.tm size 4331 mtime 1662630192 modbase /pkgs/dsc
#
# Tu Toplevels Tk - Tcl Usefulness Toplevels Tk
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#

package require Tcl 8.5

namespace eval tu { namespace eval toplevels { namespace eval tk {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::toplevels::tk
		title	{Tu Toplevels Tk}
		descr	{Tk Toplevels}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


variable defaultWindow .

proc randWindow {{w {.}}} {
	while {[winfo exists [set rw $w[randstr]]]} {}
	return $rw
}

proc windowIsToplevel {w} {
#	return [expr {![string match .?*.?* $w]}]
#	return [expr {[winfo toplevel $w] eq $w}]
	return [expr {[winfo toplevel $w] eq $w || [winfo class $w] eq "Toplevel"}]
}

proc newToplevel {{w ""} args} {
	if {$w eq ""} { set w [randWindow] }
	toplevel $w {*}$args
	return $w
}

proc groupToplevel {w {W ""}} {
	variable defaultWindow
	if {$W eq ""} { set W $defaultWindow }
	wm group $w $W
}

proc destroyToplevel {w} {
	destroy $w
}

proc transientizeToplevel {w {W ""}} {
	variable defaultWindow
	if {$W eq ""} { set W $defaultWindow }
	wm transient $w $W
}

proc withdrawToplevel {w} {
	wm withdraw $w
}

proc forgetToplevel {w} {
	wm forget $w
}

proc deiconifyToplevel {w} {
	wm deiconify $w
}

proc toplevelIsZoomable {w} {
	return [expr {"-zoomed" in [dict keys [wm attributes $w]]}]
}

proc toplevelZoomed {w {zoomed {}}} {
	if {$zoomed ne {}} { wm attributes $w -zoomed $zoomed }
	return [wm attributes $w -zoomed]
}

proc toplevelFullscreen {w {fullscreen {}}} {
	if {$fullscreen ne {}} { wm attributes $w -fullscreen $fullscreen }
	return [wm attributes $w -fullscreen]
}

proc toplevelGeometry {w args} {
	if {[llength $args] == 1} { wm geometry $w [lindex $args 0] }
	return [wm geometry $w]
}

proc manageToplevel {w} {
	wm manage $w
}

proc toplevelResizable {w args} {
	if {[llength $args] == 2} { wm resizable $w {*}$args }
	return [wm resizable $w]
}

proc setToplevelTitle {w t} {
	wm title $w $t
}

proc getToplevelTitle {w} {
	return [wm title $w]
}

proc centreToplevel {w {W ""}} {
	variable defaultWindow
	if {$W eq ""} { set W $defaultWindow }
	::tk::PlaceWindow $w widget $W
}
interp alias {} [namespace current]::centerToplvel {} [namespace current]::centreToplvel

proc onToplevelDelete {w s} {
	if {[winfo toplevel $w] eq $w} {
		wm protocol $w WM_DELETE_WINDOW $s
	} else {
		bind $w <Destroy> $s
	}
}

proc createToplevelGui {{title ""} args} {
	set w [newToplevel {} {*}$args]
	groupToplevel $w
	if {$title ne ""} { setToplevelTitle $w $title }
	return $w
}


## Setup

package require tu::toplevels

namespace path [namespace parent]

namespace export \
	randWindow windowIsToplevel newToplevel groupToplevel destroyToplevel \
	transientizeToplevel centreToplevel centerToplevel \
	withdrawToplevel forgetToplevel deiconifyToplevel \
	onToplevelDelete setToplevelTitle getToplevelTitle \
	toplevelIsZoomable toplevelZoomed toplevelFullscreen \
	toplevelGeometry manageToplevel toplevelResizable \
	createToplevelGui

package provide tu::toplevels::tk 0.1

}}}; # namespace tu::toplevels::tk


# EOF

# THATCH /pkgs/dsc/central-1.0b3.tm size 10351 mtime 1662903281 modbase /pkgs/dsc
#
# Central - Central Coordination Bureau
#
#
# Stuart Cassoff
#
# Summer 2018
# Version 0.1
#
# Fall 2020
# Version 1.0b1
#
# Summer 2022
# Version 1.0b2
# Version 1.0b3
#


namespace eval central {

if {$::tcl_platform(engine) eq "Jim"} {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	central
		title	Central
		descr	{Central Coordination Bureau}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	1.0b3
	}]
	dict set pkgconfig importfrom [namespace current]
	dict set pkgconfig initializer Central_Init
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2020,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}; # pkgconfig

} else { # !Jim

package require Tcl 8.5

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	central
		title	Central
		descr	{Central Coordination Bureau}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	1.0b3
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig initializer Central_Init
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2020,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig

}; # Jim/Tcl


## Central

proc IsGlobPat {p} { regexp {[*?[\]]} $p }


if {$::tcl_platform(engine) eq "Jim"} {

proc RnsSepArge {} { list "::" "::" "?arg...?" }

} else { # !Jim

proc RnsSepArge {} [list list ""  " " \
	[expr {[info tclversion] eq "8.5" ? "..." : "?arg ...?"}]]

}; # Jim/Tcl


variable Packages [list]


proc packages {} {
	variable Packages
	return $Packages
}

proc packagekeys {package {key {*}}} {
	if {$package ni [packages]} {
		return -code error "package \"$package\" not known"
	}
	return [lsearch -glob -inline -all [${package}::pkgconfig list] $key]
}

proc package {{package {*}} {key {*}}} {
	set pkgexact [expr {![IsGlobPat $package]}]
	set keyexact [expr {![IsGlobPat $key]}]
	if {$pkgexact} {
		if {$package ni [packages]} {
			return -code error "package \"$package\" not known"
		}
		if {$keyexact} {
			if {$key ni [packagekeys $package]} {
				return -code error "key \"$key\" not known"
			}
			return [${package}::pkgconfig get $key]
		}
	}
	set res [dict create]
	foreach pkg [lsearch -glob -inline -all [packages] $package] {
		foreach k [lsearch -glob -inline -all [packagekeys $pkg] $key] {
			dict set res $pkg $k [${pkg}::pkgconfig get $k]
		}
	}
	if {!$pkgexact || [dict size $res] == 0} {
		return $res
	} else {
		return [dict get $res $package]
	}
}

proc find {{key {*}} {value {*}}} {
	set glob [expr {[IsGlobPat $key] || [IsGlobPat $value]}]
	set found [list]
	foreach pkg [central packages] {
		foreach k [lsearch -glob -inline -all [central packagekeys $pkg] $key] {
			if {[string match $value [set val [central package $pkg $k]]]} {
				lappend found [list $pkg $k $val]
			}
		}
	}
	if {$glob || [llength $found] > 1} {
		return $found
	} else {
		return [lindex $found 0 0]
	}
}

proc centralized {tclpackage} {
	set none [list {} {}]
	set package [find "name" $tclpackage]
	if {$package eq {}} { return $none }
	if {"importfrom" ni [packagekeys $package]} { return $none }
	return [list $package [package $package importfrom]]
}

proc exports {tclpackage} {
	lassign [centralized $tclpackage] package importfrom
	if {$package eq {} || $importfrom eq {}} { return {} }
	return [namespace eval $importfrom { namespace export }]
}

proc import {args} {
	set force ""
	if {[llength $args] > 0 && [lindex $args 0] eq "-force"} {
		set args [lassign $args force]
	}
	if {[llength $args] == 0} {
		lassign [RnsSepArge] rns sep
		return -code error "wrong # args: should be \"${rns}central${sep}import ?-force? tclpackage ?pattern ...?\""
	}
	set args [lassign $args tclpackage]
	if {[llength $args] == 0} {
		set args [list *]
	}
	lassign [centralized $tclpackage] package importfrom
	if {$package eq {} || $importfrom eq {}} {
		return [list]
	}
	set toimport ${importfrom}::[join $args " ${importfrom}::"]
	uplevel 1 [list namespace import {*}$force {*}$toimport]
	return $toimport
}

proc initializer {tclpackage} {
	set package [find "name" $tclpackage]
	if {$package eq {}} { return {} }
	if {"initializer" ni [packagekeys $package]} { return {} }
	set init [package $package initializer]
	if {[string index $init 0] eq ":"} {
		return $init
	} elseif {"importfrom" ni [packagekeys $package]} {
		return {}
	} else {
		return [package $package importfrom]::$init
	}
}

proc initialize {tclpackage args} {
	set package [find "name" $tclpackage]
	if {$package eq {}} { return -code error "package \"$tclpackage\" not known" }
	set code error
	set res "Tcl package \"$tclpackage\" doesn't have an initialization command"
	if {"initializer" ni [packagekeys $package]} { return -code $code $res }
	set init [initializer $tclpackage]
	if {$init ne {}} {
		set code [expr {[catch {uplevel 1 [linsert $args 0 $init]} res] ? "error" : "ok"}]
	}
	return -code $code $res
}

proc require {args} {
	if {[llength $args] == 0} {
		lassign [RnsSepArge] rns sep
		set    msg "wrong # args: should be \"${rns}central${sep}require ?-exact?"
		append msg " ?-noinit? ?-noimport? ?-import {command ...}? ?-init command? -allinfo"
		append msg " tclpackage ?requirement ...?\""
		return -code error $msg
	}
	set exact ""
	set tclpackage ""
	set requirements [list]
	set init yes
	set import yes
	set initter {}
	set importing [list]
	set allinfo no
	set tclpackage_set no
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-exact    { set exact   -exact }
			-noinit   { set init    no     }
			-noimport { set import  no     }
			-allinfo  { set allinfo yes    }
			-init - -import { set state $arg }
			default { if {$tclpackage_set} {
					lappend requirements $arg
				  } else {
					set tclpackage $arg
					set tclpackage_set yes
				  }}
			}; #switch
		}
		-init   { set state opt; set initter   $arg }
		-import { set state opt; set importing $arg }
		}; #switch
	}
	set reqRes [uplevel 1 [list package require {*}$exact {*}$tclpackage {*}$requirements]]
	uplevel 1 [list [namespace current]::central discover]
	if {$init}   { set initRes   [uplevel 1 [list {*}$initter [initializer $tclpackage]]]
	} else       { set initRes   {} }
	if {$import} { set importRes [uplevel 1 [list [namespace current]::central import $tclpackage {*}$importing]]
	} else       { set importRes {} }
	if {$allinfo} {
		return [dict create require $reqRes init $initRes import $importRes]
	} else {
		return $reqRes
	}
}

if {$::tcl_platform(engine) eq "Jim"} {

proc discover {{exclude {}}} {
	variable Packages
	set pkgs [list]
	foreach p [info commands ::*::pkgconfig] {
		set ns [namespace parent $p]
		if {$ns ni $exclude} {
			lappend pkgs $ns
		}
	}
	set Packages [lsort $pkgs]
	return $Packages
}

} else { # !Jim

proc discover {{exclude {}}} {
	variable Packages
	set pkgs [list]
	set nsl [list ::]
	while {[llength $nsl] > 0} {
		set nsl [lassign $nsl[set nsl {}] ns]
		if {$ns in $exclude} continue
		if {[llength [info commands ${ns}::pkgconfig]] == 1} {
			lappend pkgs $ns
		}
		lappend nsl {*}[namespace children $ns]
	}
	set Packages [lsort $pkgs]
	return $Packages
}

}; # Jim/Tcl


## Setup

proc Central_Init {} { return 1.0b3 }; # For testing and demonstration


variable CSCmds [list centralized discover exports find import initialize initializer package packagekeys packages require]


if {$::tcl_platform(engine) eq "Jim"} {

proc central {cmd args} {
	variable CSCmds
	if {$cmd in $CSCmds} {
		tailcall ::central::$cmd {*}$args
	} else {
		return -code error "unknown subcommand \"$cmd\": must be [join [lrange $CSCmds 0 end-1] ", "], or [lindex $CSCmds end]"
	}
}

} else { # !Jim

namespace ensemble create -prefixes no -command central -subcommands $CSCmds

unset CSCmds

}; # Jim/Tcl


namespace export central

# Careful!
::package provide central 1.0b3


}; # namespace eval central


# EOF

# THATCH /pkgs/dsc/central/extras-1.0b3.tm size 9901 mtime 1662903082 modbase /pkgs/dsc
#
# Central Extras - Central Coordination Bureau: Extras
#
#
# Stuart Cassoff
#
# Summer 2018
# Version 0.1
#
# Fall 2020
# Version 1.0b1
#
# Summer 2022
# Version 1.0b2
# Version 1.0b3
#


if {$::tcl_platform(engine) ne "Jim"} {

package require Tcl 8.5

}; # !Jim


namespace eval central { namespace eval extras {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	central::extras
		title	{Central Extras}
		descr	{Central Coordination Bureau: Extras}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	1.0b3
	}]
	dict set pkgconfig importfrom [namespace current]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2020,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}; # pkgconfig


## Central Extras

proc Maxstrlen {args} {
	set max 0
	foreach str $args {
		if {[set len [string length $str]] > $max} {
			set max $len
		}
	}
	return $max
}

proc textDump {{myPackages {}}} {
	if {[llength $myPackages] == 0} { set myPackages [central packages] }
	set fPackages [list]
	foreach package $myPackages {
		set keys [central packagekeys $package]
		set max [Maxstrlen 0 {*}$keys]
		set l [list]
		foreach key $keys {
			lappend l [format "%-${max}s %s" $key [central package $package $key]]
		}
		lappend fPackages [join $l \n] {} {}
	}
	return [join $fPackages \n]
}


if {$::tcl_platform(engine) ne "Jim"} {

proc SetTextText {w text} {
	if {[set state [$w cget -state]] ne "normal"} { $w configure -state normal }
	$w delete 1.0 end
	if {$text ne ""} { $w insert end $text }
	if {$state ne "normal"} { $w configure -state $state }
}

proc packages->textWidget {w {packagetag {}} {myPackages {}}} {
	if {[llength $myPackages] == 0} { set myPackages [central packages] }
	set i 0
	foreach package $myPackages {
		if {[incr i] > 1} {
			$w insert end \n\n
		}
		$w insert end $package $packagetag \n\n
		set keys [central packagekeys $package]
		set max [Maxstrlen 0 {*}$keys]
		foreach key $keys {
			$w insert end [format "%-${max}s %s%s" $key [central package $package $key] \n]
		}
	}
}


if {[catch {package require TclOO}]} {

variable TclOO no

} else { # TclOO

variable TclOO yes

oo::class create CentralPackageBrowser {
	variable w W
	variable packagesList packagesText licenseFrame
	variable packageFields
	foreach field [list name version title author when descr license,type] { variable $field }
	variable myNs
constructor {window {toolkit {ttk}}} {
	if {$toolkit ni [list tk ttk]} { return -code error "bad toolkit \"$toolkit\": must be one of tk, or ttk" }
	namespace import ::central::central
	set packageFields [list name version title author when descr license,type license,text]
	set w [string trimright [set W $window] .]
	my CreateGui $w $toolkit
	central discover
	set myNs [central find name central::extras]
	foreach package [central packages] {
		if {"title" in [central packagekeys $package]} {
			set pkgtitle [central package $package title]
		} else {
			set pkgtitle $package
		}
		if {$toolkit eq "ttk"} {
			$packagesList insert {} end -values [list $pkgtitle $package]
		} else {
			$packagesList insert end $pkgtitle
		}
	}
	if {$toolkit eq "ttk"} {
		set item [lindex [$packagesList children {}] 0]
		$packagesList focus $item
		$packagesList selection set [list $item]
	} else {
		$packagesList selection set 0
		event generate $packagesList <<ListboxSelect>>
	}
}
destructor {
}
method CreateGui {w toolkit} {
	set k $toolkit:
	grid [$k:panedwindow $w.p0 -orient horizontal] -sticky nsew -row 0 -column 0
	$w.p0 add [$k:labelframe $w.p0.m0 -text Packages]
	$w.p0 add [$k:panedwindow $w.p0.p1 -orient vertical]
	$w.p0.p1 add [$k:labelframe $w.p0.p1.m1 -text Info]
	$w.p0.p1 add [set licenseFrame [$k:labelframe $w.p0.p1.m2 -text License]]

	if {$toolkit eq "ttk"} {
		set packagesList [$k:treeview $w.p0.m0.v0 -columns {title package} -displaycolumns {title} \
				  -selectmode browse -show {} \
				  -xscrollcommand [list $w.p0.m0.u1 set] -yscrollcommand [list $w.p0.m0.u0 set]]
	} else {
		set packagesList [$k:listbox $w.p0.m0.v0 -width 23 \
				  -xscrollcommand [list $w.p0.m0.u1 set] -yscrollcommand [list $w.p0.m0.u0 set]]
	}
	grid $packagesList -sticky nsew -row 0 -column 0
	grid [$k:scrollbar $w.p0.m0.u0 -orient vertical -command [list $w.p0.m0.v0 yview]] -sticky ns -row 0 -column 1
	grid [$k:scrollbar $w.p0.m0.u1 -orient horizontal -command [list $w.p0.m0.v0 xview]] -sticky ew -row 1 -column 0


	grid [$k:label $w.p0.p1.m1.l0  -text Title]   -sticky w -row 0 -column 0
	grid [$k:label $w.p0.p1.m1.l2  -text Author]  -sticky w -row 0 -column 2
	grid [$k:label $w.p0.p1.m1.l4  -text Name]    -sticky w -row 1 -column 0
	grid [$k:label $w.p0.p1.m1.l6  -text When]    -sticky w -row 1 -column 2
	grid [$k:label $w.p0.p1.m1.l8  -text Version] -sticky w -row 2 -column 0
	grid [$k:label $w.p0.p1.m1.l10 -text License] -sticky w -row 2 -column 2
	grid [$k:label $w.p0.p1.m1.l12 -text Descr]   -sticky w -row 3 -column 0

	set lopts {-width 20 -anchor w}
	grid [$k:label $w.p0.p1.m1.l1  -textvariable [my varname title]   {*}$lopts] -sticky w -row 0 -column 1
	grid [$k:label $w.p0.p1.m1.l3  -textvariable [my varname author]  {*}$lopts] -sticky w -row 0 -column 3
	grid [$k:label $w.p0.p1.m1.l5  -textvariable [my varname name]    {*}$lopts] -sticky w -row 1 -column 1
	grid [$k:label $w.p0.p1.m1.l7  -textvariable [my varname when]    {*}$lopts] -sticky w -row 1 -column 3
	grid [$k:label $w.p0.p1.m1.l9  -textvariable [my varname version] {*}$lopts] -sticky w -row 2 -column 1
	grid [$k:label $w.p0.p1.m1.l11 -textvariable [my varname license,type]]      -sticky w -row 2 -column 3
	grid [$k:label $w.p0.p1.m1.l13 -textvariable [my varname descr]]             -sticky w -row 3 -column 1 -columnspan 3

	grid [set packagesText [text $w.p0.p1.m2.t0 -width 20 -state disabled -height 10 -wrap none -xscrollcommand [list $w.p0.p1.m2.u3 set] -yscrollcommand [list $w.p0.p1.m2.u2 set]]] -sticky nsew -row 0 -column 0
	grid [$k:scrollbar $w.p0.p1.m2.u2 -orient vertical -command [list $w.p0.p1.m2.t0 yview]] -sticky ns -row 0 -column 1
	grid [$k:scrollbar $w.p0.p1.m2.u3 -orient horizontal -command [list $w.p0.p1.m2.t0 xview]] -sticky ew -row 1 -column 0
	grid rowconfigure $w.p0.m0 0 -weight 1
	grid rowconfigure $w.p0.p1.m2 0 -weight 1
	grid columnconfigure $w.p0.m0 0 -weight 1
	grid columnconfigure $w.p0.p1.m1 1 -weight 1
	grid columnconfigure $w.p0.p1.m2 0 -weight 1

	if {$toolkit eq "ttk"} {
		$packagesList column title -width 185
	}
	bind $packagesList <<[expr {$toolkit eq "ttk" ? "Treeview" : "Listbox"}]Select>> [list [self] showPackageInfo $toolkit]
	bind $w <Map> [list focus $packagesList]
}
method showPackageInfo {toolkit} {
	if {$toolkit eq "ttk"} {
		if {[llength [set sel [$packagesList selection]]] == 0} { return }
		set item [lindex $sel 0]
		set package [$packagesList set $item package]
	} else {
		if {[llength [set sel [$packagesList curselection]]] == 0} { return }
		set package [lindex [central packages] $sel]
	}
	set packageInfo [central package $package]
	set nData 0
	foreach field $packageFields {
		if {[set $field [my dgwd {} $packageInfo $field]] ne {}} {
			incr nData
		}
	}
	${myNs}::SetTextText $packagesText ${license,text}
	if {$nData > 0} {
		$licenseFrame configure -text License
	} else {
		$licenseFrame configure -text Info
		${myNs}::SetTextText $packagesText [${myNs}::textDump $package]
	}
}
method dgwd {def dict key args} {
	if {![dict exists $dict $key {*}$args]} { return $def }
	return [dict get $dict $key {*}$args]
}
}; # oo::class::create CentralPackageBrowser

}; # TclOO

}; # !Jim


## Setup

package require central

variable CNS [namespace parent]

namespace import ${CNS}::central


if {$::tcl_platform(engine) eq "Jim"} {

variable Cmd {}

foreach Cmd [list textDump] {
	alias ${CNS}::$Cmd [namespace current]::$Cmd
	lappend ${CNS}::CSCmds $Cmd
}

set ${CNS}::CSCmds [lsort [set ${CNS}::CSCmds]]

unset CNS Cmd

} else { # !Jim

variable CESCmds [list textDump packages->textWidget]
if {$TclOO} { lappend CESCmds CentralPackageBrowser }
variable Cmd {}
variable Map [namespace ensemble configure ${CNS}::central -map]

namespace ensemble configure ${CNS}::central -subcommands \
	[lsort [concat [namespace ensemble configure ${CNS}::central -subcommands] $CESCmds]]

foreach Cmd $CESCmds { dict set Map $Cmd [namespace current]::$Cmd }

namespace ensemble configure ${CNS}::central -map $Map

unset CESCmds CNS Cmd Map TclOO

}; # Jim/Tcl


package provide central::extras 1.0b3

}}; # namespace central::extras


# EOF

# THATCH /pkgs/dsc/gub-1.0a1.tm size 68287 mtime 1662214944 modbase /pkgs/dsc
#
# Gub - GUiBuilder
#
# "The world's fastest gui builder."
# "An nroff for Tk."
#
# Stuart Cassoff
#
# Fall/Winter 2008
# Version 0.3
#
# Summer 2009
# Version 0.4
#
# Fall 2011, Winter 2012, Spring 2012
# Version 0.5
#
# Summer 2012, Winter 2013, Spring 2015
# Version 0.6
#
# Spring 2015
# Version 0.7
#
# Spring 2015
# Version 0.8
#
# Fall/Winter/Spring 2018
# Version 0.9
#
# Summer 2022
# Version 1.0a1
#

[namespace eval gub {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	gub
		title	Gub
		descr	{Gui Builder}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	1.0a1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2008,2011,2012,2015,2018,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


#---------------------------------------------------------------------
# Package declaration, exports, variables
#---------------------------------------------------------------------

proc provide {} {
	package provide gub 1.0a1
	return on
}

proc require {} {
	package require Tcl 8.5
	return true
}

proc export {} {
	namespace export {*}{
		gub gubParse gubAssemble
		guhb guhbParse guhbAssemble guhbGubParse
		gubWidgetMap gubTkWidgetMap gubAutoscrollbarWidgetMap
		gubSaveGob gubButtonGob gubOpGob gubPropGob
		gubBindGob gubAddGob gubWrapToggleGob
		gubOpListGob gubMultiGob
		gubGoobKeys gubGoobWidgetKeys
		gubWidgetMapKeys gubWidgetMapElementKeys
		gubAutoUnderlines
		gubLangs gubTLangs
		gubUA
		gubOffsets
		gubExpandEventKeys
		gubParseArgs
		gubOperators gubOperatorNames
		guhbOperators guhbOperatorNames
		gubSubOperators gubSubOperatorNames
		gubBoilerplate gubBoilerplateHead gubBoilerplateTail gubWrapSource
	}
	#namespace export ngub ogub
	return yes
}

proc initvars {} {
	variable WidgetMap       [gubTkWidgetMap]
	variable GubOperators    [gubOperators]
	variable GubSubOperators [gubOperators]
	variable GubScrollbarSelectors [dict create - H | V]
#	variable lsl [dict keys $lsd]
	return   true
}


#---------------------------------------------------------------------
# Info
#---------------------------------------------------------------------

proc gubOperators {} {
	list @ / ^ & - | . : + ! %
}

proc gubOperatorNames {} {
	dict create {*}{
		@  label
		/  save
		^  sibling
		&  bind
		-  rowconfigure
		|  columnconfigure
		.  property
		:  grid
		+  add
		!  alternatefunctionality
		%  scrollbarlink
	}
}

proc guhbOperators {} {
	list >
}

proc guhbOperatorNames {} {
	dict create > in
}

proc gubSubOperators {} {
	list / : + ! % ~
}

proc gubSubOperatorNames {} {
	dict create {*}{
		/  altsave
		:  grid
		+  add
		!  name
		%  disableautoscrollbarlinking
		~  widgetmap
	}
}

proc gubGoobKeys {} {
	list in widgets rccs binds adds labels N F E offsets
}

proc gubGoobWidgetKeys {} {
	list widgetchar n path options row col savename gridoptions links
}

proc gubWidgetMapKeys {} {
	list widgetchar widgetmapelement
}

proc gubWidgetMapElementKeys {} {
	list command properties
}

proc gubAutoUnderlines {} {
	list -text
}

proc gubTLangs {} { list &Tcl R&uby &Lua &Perl P&ython }
proc gubLangs  {} { list  tcl  ruby  lua  perl  python }


#---------------------------------------------------------------------
# Widget map
#---------------------------------------------------------------------

proc gubTkWidgetMap {} {
	dict create {*}{
		b  {command  tk::button}
		c  {command  tk::canvas		properties scrollable}
		k  {command  tk::checkbutton}
		e  {command  tk::entry}
		f  {command  tk::frame		properties container}
		l  {command  tk::label}
		m  {command  tk::labelframe	properties container}
		o  {command  tk::listbox	properties scrollable}
		u  {command  tk::menubutton}
		p  {command  tk::panedwindow	properties adder}
		r  {command  tk::radiobutton}
		a  {command  tk::scale}
		s  {command  tk::scrollbar	properties scroller}
		i  {command  tk::spinbox}
		t  {command  tk::text		properties scrollable}
		B  {command ttk::button}
		K  {command ttk::checkbutton}
		X  {command ttk::combobox}
		E  {command ttk::entry}
		F  {command ttk::frame		properties container}
		L  {command ttk::label}
		M  {command ttk::labelframe	properties container}
		U  {command ttk::menubutton}
		N  {command ttk::notebook	properties adder}
		P  {command ttk::panedwindow	properties adder}
		O  {command ttk::progressbar}
		R  {command ttk::radiobutton}
		A  {command ttk::scale}
		S  {command ttk::scrollbar	properties scroller}
		T  {command ttk::separator}
		G  {command ttk::sizegrip}
		I  {command ttk::spinbox}
		V  {command ttk::treeview	properties scrollable}
	}
}

proc gubAutoscrollbarWidgetMap {{wc {u}}} {
	set wc [string tolower $wc]
	set WC [string toupper $wc]
	return [dict create \
		$wc {command ::gub::tk::autoscrollbar  properties scroller} \
		$WC {command ::gub::ttk::autoscrollbar properties scroller} \
	]
}

proc gubWidgetMap {{newWidgetMap {}}} {
	variable WidgetMap
	set oldWidgetMap $WidgetMap
	if {[dict size $newWidgetMap] > 0} { set WidgetMap $newWidgetMap }
	return $oldWidgetMap
}


#---------------------------------------------------------------------
# Gub Level 0
#---------------------------------------------------------------------

proc gubGuh {} { dict merge [dict create {*}{
	in		.
	lang		tcl
	varname		gui
	vartype		dict
	setcmd		set
	widgetmap	{}
	offsets		{}
	offsetrow	0
	offsetcol	0
	widgetchar_in	""
	autoul		!gub!
	compact		yes
	eolChar		\n
	guhb		auto
	gob		""
	rgob		{}
	inline_put	yes
	inline_link	yes
	inline_save	yes
	sections	{create put add link rcc bind save}
	wrap		no
}] [dict create langs [gubLangs]]}

proc gubParseArgs {args} {
	set opts    [list -i        -gb        -gh        -lg        -wr \
			  -a        -sc        -vn        -co        -au \
			  -ec       -os        -or        -oc        -iw \
			  -rg       -wm                              -sp \
			                                             -ss \
			                                             -sl]

	set options [list -in       -gob       -guhb      -lang      -wrap           \
			  -array    -setcmd    -varname   -compact   -autounderlines \
			  -eolchar  -offsets   -offsetrow -offsetcol -inwidgetchar   \
			  -readygob -widgetmap                       -separateput    \
			                                             -separatesave   \
			                                             -separatelink]

	set binopts    [list -a     -co      -wr   -sp          -ss           -sl          ]
	set binoptions [list -array -compact -wrap -separateput -separatesave -separatelink]

	set binvals [dict create -a array -co yes -wr yes -sp no -ss no -sl no]

	set elems [dict create -a vartype -co compact -wr wrap -i in -lg lang -vn varname \
		-sc setcmd -wm widgetmap -sp inline_put -ss inline_save -sl inline_link \
		-os offsets -or offsetrow -oc offsetcol \
		-iw widgetchar_in -au autoul -ec eolChar -gh guhb -gb gob -rg rgob]

	set optIs gob
	set xtra -3
	set state opt
	set guh [gubGuh]

	foreach arg $args {
		if {$state in $opts} {
			dict set guh [dict get $elems $state] $arg
			set state opt
		} elseif {[set i [lsearch -exact $binopts $arg]] != -1 || [set i [lsearch -exact $binoptions $arg]] != -1} {
			set arg [lindex $binopts $i]
			dict set guh [dict get $elems $arg] [dict get $binvals $arg]
		} elseif {[set i [lsearch -exact $opts $arg]] != -1 || [set i [lsearch -exact $options $arg]] != -1} {
			set state [lindex $opts $i]
		} else {
			dict set guh $optIs $arg; set optIs [string map [list gob in in varname varname 0 $xtra [incr xtra]] $optIs]
		}
	}
	return $guh
}

proc gub {args} {
	# Backcompat for programs using old-style [gub].
	#if {[llength $args] >= 4 && [lindex $args 2] eq {}} { return [ogub {*}$args] }
	# New style
	set guh [gubParseArgs {*}$args]
	CheckArg $guh guhb
	if {[GobIsGohb [dict get $guh gob]] && [dict get $guh guhb] in [list auto yes]} { tailcall Guhb $guh }
	dict set guh goob [gubParse $guh]
	set src [dict get [gubAssemble $guh] [dict get $guh lang]]
	return [if {[dict get $guh wrap]} {
		gubWrapSource $src [dict get $guh lang]
	} else {
		set src
	}]
}

if 0 {
# Backcompat for programs using deprecated [ngub].
interp alias {} [namespace current]::ngub {} [namespace current]::gub

# Backcompat for programs using old-style [gub].
proc ogub {gob {in {.}} {widgetMap {}} {varname {gui}} {vartype {dict}} {autoul {!gub!}}} {
	set guh [gubParseArgs $gob -in $in -wm $widgetMap -au $autoul -vn $varname {*}[expr {$vartype eq "array" ? "-a" : ""}]]
	dict set guh goob [gubParse $guh]
	return [dict get [gubAssemble $guh] [dict get $guh lang]]
}
}

#---------------------------------------------------------------------
# Gub Level 1
#---------------------------------------------------------------------

proc getPathForSavename {wigs savename} {
	foreach widget $wigs {
		if {[dgwd {} $widget savename] eq $savename} {
			return [dict get $widget path]
		}
	}
	return {}
}

proc getWigPropertyForPath {wigs path {property {widgetchar}}} {
	foreach widget $wigs {
		if {[dgwd {} $widget path] eq $path} {
			return [dgwd {} $widget $property]
		}
	}
	return {}
}

proc guhbResolveReferences {guh src} {
	set wigs [dict get $guh wigs]
	set labs [dict get $guh labs]
	set refmap [dict create]
	foreach {ref name} [regexp -inline -all {/>(.+?)/} $src] {
		if {[dict exists $refmap $ref]} continue
		if {[set path [getPathForSavename $wigs $name]] eq "" && [set path [dgwd "" $labs $name]] eq ""} continue
		dict set refmap $ref $path
	}
	if {[dict size $refmap] > 0} {
		return [string map $refmap $src]
	} else {
		return $src
	}
}

#
# Separate the gobs
#
proc guhbParse {guh} {
	set g [list]
	set in .
	set rgob [list]
	set args [list]
	foreach l [makeReady [dict get $guh gohb]] {
		if {[string index $l 0] eq ">"} {
			lappend g $in $rgob $args
			set args [lassign [string trim [string range $l 1 end]] in]
			set rgob [list]
		} else {
			lappend rgob $l
		}
	}
	lappend g $in $rgob $args
	set goohb [list]
	foreach {in rgob args} $g {
		if {[llength $args] == 0} {
			lappend goohb $in $rgob
			continue
		}
		optsLenMustBeEven guhb $args
		unset -nocomplain keys values
		set options [list -keys -values]
		set state opt
		foreach arg $args {
			if {$state in $options} {
				set [string range $state 1 end] $arg
				set state opt
			} elseif {$arg in $options} {
				set state $arg
			}
		}
		if {!([info exists keys] && [info exists values] && [llength $keys] > 0)} {
			lappend goohb $in $rgob
			continue
		}
		set vlen [llength $values]
		set vi 0
		set ghi 0
		while {$vi < $vlen} {
			set keyvals [list]
			foreach key $keys {
				lappend keyvals $key [lindex $values $vi]
				incr vi
			}
			lappend goohb $in [guhbSubI [string map $keyvals $rgob] $ghi]
			incr ghi
		}
	}
	return $goohb
}

#
# Parse each gob
#
proc guhbGubParse {guh} {
	set gooohb [list]
	set offsets [dict get $guh offsets]
#dict set offsets b 0
	set rofs [dict create [dict get $guh in] [dict get $guh offsetrow]]
	set wigs [list [dict create savename . path [dict get $guh in]]]
	set labs [list]
	set gubargs [list \
		-in [dict get $guh in] \
		-wm [dict get $guh widgetmap] \
		-vn [dict get $guh varname] \
		-sc [dict get $guh setcmd] \
		-au [dict get $guh autoul] \
	]
	if {[dict get $guh vartype] eq "array"} { lappend gubargs -a  }
	if {[dict get $guh compact]}            { lappend gubargs -co }
	if {![dict get $guh inline_put]}        { lappend gubargs -sp }
	if {![dict get $guh inline_link]}       { lappend gubargs -sl }
	if {![dict get $guh inline_save]}       { lappend gubargs -ss }
	set lang [dict get $guh lang]
	set goohb [dict get $guh goohb]
	set done [list]
	set ntodo [expr {[llength $goohb]/2}]
	set ndone -1
	while {$ndone != 0 && $ndone != $ntodo} { set ndone 0; set i -1; foreach {in rgob} $goohb {
		incr i
#puts rofs:$rofs
		if {$i in $done} continue
		if {[set path [getPathForSavename $wigs $in]] eq "" && [set path [dgwd "" $labs $in]] eq ""} continue
		if {![dict exists $rofs $path]} { dict set rofs $path 0 }
		set gguh [gubParseArgs {*}$gubargs -in $path -iw [getWigPropertyForPath $wigs $path widgetchar] -offsets $offsets -offsetrow [dict get $rofs $path] -rg $rgob -lg $lang]
		dict set gguh goob [gubParse $gguh]
		lappend gooohb $gguh
		set offsets [dict get $gguh goob offsets]
		lappend wigs {*}[dgwd {} $gguh goob widgets]
		lappend labs {*}[dgwd {} $gguh goob labels]
#puts p:[dict get $rofs $path]
		dict incr rofs $path
#puts in_row:[dict get $gguh goob in_row]
		foreach {path offset} [dict get $gguh goob in_row] {
			dict set rofs $path [expr {$offset + 1}]
		}
		lappend done $i
		incr ndone
	}}
	return [list $gooohb $wigs [dict create {*}$labs]]
}

proc guhbAssemble {guh} {
	set nl [expr {[dict get $guh compact] ? "" : "\n"}]
	set src ""
	foreach gguh [dict get $guh gooohb] {
		append src [dict get [gubAssemble $gguh] [dict get $gguh lang]] \n $nl
	}
	set res [dict create tcl {} ruby {} lua {} perl {} python {}]
	dict set res [dict get $guh lang] [guhbResolveReferences $guh $src]
	return $res
}

proc Guhb {guh} {
	set in [string trimright [dict get $guh in]]
	dict set guh in [expr {$in eq "" ? "." : $in}]
	dict set guh gohb [dict get $guh gob]
	dict unset guh gob
	dict set guh goohb [guhbParse $guh]
	lassign [guhbGubParse $guh] gooohb wigs labs
	dict set guh gooohb $gooohb
	dict set guh wigs $wigs
	dict set guh labs $labs
	return [dict get [guhbAssemble $guh] [dict get $guh lang]]
}
proc guhb {args} {
	#puts !$args!
	Guhb [gubParseArgs {*}$args]
}


#---------------------------------------------------------------------
# Parse
#---------------------------------------------------------------------

proc gubParse {guh} {
	variable WidgetMap
	variable GubOperators
	variable GubSubOperators
	variable GubScrollbarSelectors

	# Create as local vars some elements from the guh and adjust as needed
	foreach k [list in widgetmap offsetrow offsetcol widgetchar_in autoul compact gob rgob] {
		set $k [dict get $guh $k]
	}
	if {$in eq ""} { set in [dict get [gubParseArgs] in] }
	if {$autoul eq "!gub!" } { set autoul [gubAutoUnderlines] }
	set wm [expr {[dict size $widgetmap] == 0 ? $WidgetMap : $widgetmap}]
	incr offsetrow -1

	# Create arrays
	foreach a {N O F E C R S P G V H X B L D Q U Z M A} { array set $a {} }

	# N widget counters
	# O widget counter offsets
	# F column offsets
	# E row offsets
	# C columnconfigures
	# R rowconfigures
	# S saves
	# P widget opts
	# G grid opts
	# V vertiscrolls
	# H horiscrolls
	# X alternate functionality
	# B binds
	# L labels
	# D label definitions (unexpanded widgetSpecs)
	# Q add opts
	# U putcmds
	# Z altsaves
	# M moveups
	# A names

	# Initialize widget counters
	foreach wc [dict keys $wm] { set N([string tolower $wc]) -1 }

	# Bring in offsets
	dict for {wc v} [dict get $guh offsets] {
		if {![info exists N($wc)]} { return -code error "bad offset \"$wc\"" }
		if {$v < 1} continue
		set O($wc) $v
	}

	# Setup "in"
	set orig_in $in
	set in [string trimright $in .]
	set goob [dict create]
	dict set goob in [expr {$in eq "" ? "." : $in}]

	# Widget map changes
	set wmchgs [dict create]


	# Parse: separate layout and process operators

	set layout [list]

	# Brace expansion
	set braceMax -1
	set braceNum  0
	set braceParm ""

	# Enable auto scrollbar linking
	# "aslH", "aslV": autoScrollLink (H)orizontal, (V)ertical
	set aslH 1
	set aslV 1

	# Use gob if supplied regardless if rgob exists
	if {$gob ne ""} {
		set rgob [makeReady $gob]
	}

	# Run through the lines
	foreach line $rgob {
		# Clean and separate
		set ln [string trimleft $line]
		if {$ln eq ""} { explode }
		set op [string index $ln 0]
		if {$op eq "#"} { explode }
		if {$op ni $GubOperators} {
			lappend layout $line
			continue
		}

		# Tabs count as spaces. Tricky.
		set ln [string map [list \t " "] $ln]
		# name, value
		if {[set i [string first " " $ln]] == -1} {
			set n [string range $ln 1 end]
			set v ""
		} else {
			set n [string range $ln 1 $i-1]
			set v [string range $ln $i+1 end]
		}
		set vv [string trim $v]

		# Bind and break
		if {$op eq "&"} {
#			if {$n eq ""} continue
			set c [string index $n 0]
			if {$c eq "&"} {
				error ampersand
			} elseif {$c eq "!"} {
				set bab [expr {[string first " " $vv] == -1 ? "" : ";break"}]
				set n [string range $n 1 end]
			} else {
				set bab ""
			}
		}

		# Scrollbar linking
		if {$op eq "%"} {
			if {[string length $n] < 2} continue
			set selector [string index $n 0]
#			if {$selector ni $lsl} continue
			set whichLink [dict get $GubScrollbarSelectors $selector]
			set n [string range $n 1 end]
		}

		# Alternate functionality
		if {$op eq "!"} {
			if {$n ne "" && [set subop [string index $n 0]] in $GubSubOperators} {
				set n [string range $n 1 end]
			} else {
				set subop ""
			}
		}
		if {$op eq "!" && $subop ne ""} {
			if {$subop eq ":"} { #grid
				if {[string index $n 0] in [list - |]} { #row/column offset
					# "ofsA": name of offset array
					set ofsA [string map [list | F - E] [string index $n 0]]
					set n [string range $n 1 end]
					if {$n eq ""} continue
				} else { #putcmd
					set ofsA ""
				}
			} elseif {$subop eq "%"} { #disable autoscrollbarlinking
				set selector [string index $n 0]
				if {$selector eq ""} {
					set aslH 0
					set aslV   0
				} else {
					checkOpt selector $selector
					set n [string range $n 1 end]
					set linkDir [dict get [dict create - H | V] $selector]
					set asl$linkDir 0
				}
				continue
			} elseif {$subop eq "~"} { #widget map
				if {[set i [string first " " $v]] == -1} {
					set wc $v
					set wp ""
				} else {
					set wc [string range $v 0 $i-1]
					set wp [string range $v $i+1 end]
				}
				if {$n ne "" && $wc ne ""} {
					set nme [dict create command $wc]
					if {[llength $wp] > 0} {
						dict set nme properties $wp
					}
					set wmchgs [dict merge $wmchgs [dict create $n $nme]]
					set wm [dict merge $wm $wmchgs]
				}
				continue
			}
		}; #alternatefunctionality if

		if {$n eq ""} { set n all } ;# Nothing is everything

		# Label definitions (unexpanded widgetSpecs)
#!!!!!!!!!!
		if {$op eq "@" && $vv ne ""} {
			#if {$vv eq "sall"} continue; # Forbidden!
			set D($vv) $n
		}

		# Setup brace expansion
		if {[string index $n 0] eq "\{" && [string index $n end] eq "\}"} {
			set n [string range $n 1 end-1]
			if {[info exists D($n)]} { set n $D($n) }
			set vv [lassign $vv[set vv {}] braceParm]
			set braceMax [llength $vv]
			set braceCount 0
			set braceNum 0
			# Drive the brace loop with a ranged widgetSpec
			if {[string length $n] == 1} { # Single widgetchar; counter set to # of items
				set n ${n}0-[expr {$braceMax - 1}]
			} elseif {[regexp {^(.)(\d+)-$} $n -> c s]} { # Widgetchar and num with trailing '-'; start counting at num, counter set to # of items
				set n ${c}${s}-[expr {$braceMax - 1 + $s}]
			}
		} else {
			# Reset or a prior brace expansion will resume if the 
			# number of widgets was less than the brace expansion args length
			set braceMax -1
		}

		# Operators
		foreach n [Expand $n[set n {}]] {
			if {$vv eq "" && $op in [list . @]} continue; # These don't use an empty $vv so skip them
			# Brace expansion
			if {$braceNum < $braceMax} {
#puts braceMax=$braceMax:$braceCount=$braceCount:[expr {$braceCount % $braceMax}]:braceNum=$braceNum
				set v [gubSubI [list {*}$braceParm [lindex $vv [expr {$braceCount % $braceMax}]]] $braceCount]
				incr braceCount
				if {[incr braceNum] == $braceMax} { set braceNum 0 } ;# Wrap around
			} else {
				set v $vv
			}
			# Operators
			switch -exact -- $op {
			@ {     set L($n) $v; #label }
			& { lappend B($n) $v$bab; #bind }
			- { lappend R($n) $v; #rowconfigure }
			| { lappend C($n) $v; #columnconfigure }
			. { lappend P($n) {*}[continueIfLenOdd $v]; #property }
			/ { set S($n) [expr {[llength $v] == 1 && [llength [lindex $v 0]] > 1 ? [join $v] : $v}]; #save }
			^ { set M($n) [expr {$v eq "" ? "1" : $v}]; #sibling }
			: { set G($n) [expr {$v eq "" ? "-sticky nsew" : [continueIfLenOdd $v]}]; #grid }
			+ { set Q($n) [continueIfLenOdd $v]; #add }
			% { set ${whichLink}($n) $v; #scrollbarlinking }
			! { #alternatefunctionality
			          if {$subop eq ":" && $ofsA ne ""} { set ${ofsA}($n) $v; #row/column offset
			    } elseif {$subop eq ":"} { set U($n) $v; #grid
			    } elseif {$subop eq "+"} { set U($n) $v; #add
			    } elseif {$subop eq "!"} { set A($n) $v; #name
			    } elseif {$subop eq "/"} { if {$n eq "!" && $v ne ""} { dict set guh varname $v } else { set Z($n) $v }; #save
			    } else                   { set X($n) $v; #alternatefunctionality }
			}
			default {}
			}; #operators switch
		}; #operators loop
	}; #lines loop


	# Process layout; create widget hierarchy, resolve widget options

	array set paths {}
	set paths(.)  [dict get $goob in]
	set paths(..) [parentPath $paths(.)]
	set paths(\\.) "."

	set widgets     [dict create]
	set adders      [dict create]
	set auto_adds   [dict create]
	set scrollers   [dict create]
	set scrollables [list]

	array set in_x {}
	array set in_row {}
	array set in_col {}
	array set is_first_child {}

	set initial_in $in

	set in_x(-1) $in
	set in_row($in) $offsetrow
	set in_col($in) $offsetcol
	set is_first_child($in) yes
	set y -1
	set container 0

	if {$widgetchar_in ne "" && [Is adder [dict get $wm $widgetchar_in]]} {
		dict set adders $initial_in .
	}

	foreach l $layout {
		incr y
		set x -2
		foreach zw [ArrayValues in_x] {
			set in_col($zw) $offsetcol
		}
		foreach child [array names is_first_child] {
			set is_first_child($child) yes
		}
		foreach {c m} [split $l ""] {
			incr x 2;
			foreach xp [lsort -integer -decreasing [array names in_x]] {
				if {$x > $xp} {
					set in $in_x($xp)
					break
				}
			}
			if {[dict exists $wm $c]} {
				# lc widgetchar and num for path et al
				set lc [string tolower $c]
				set n [incr N($lc)]
				set ofs_n $n
				if {[info exists O($lc)]} {
					incr ofs_n $O($lc)
				}

				# these
				# ?alllabel? all ?wcharlabel? wchar ??wnamlabel? wnam? ?wnumlabel? wnum
				set these [list all $c]
				if {$m ne " "} { lappend these $c$m }
				lappend these $c$n
				foreach what $these[set these {}] {
					if {[info exists L($what)]} {
						lappend these $L($what)
					}
					lappend these $what
				}

				# those
				# wnum ?wnumlabel? ?wnam ?wnamlabel?? wchar ?wcharlabel? all ?alllabel?
				set those [list $c$n]
				if {$m ne " "} { lappend those $c$m }
				lappend those $c all
				foreach what $those[set those {}] {
					lappend those $what
					if {[info exists L($what)]} {
						lappend those $L($what)
					}
				}

				# Flying Teapot
				foreach what $these {
					if {![info exists M($what)]} continue
					if {[info exists paths($M($what))]} {
						set in $paths($M($what))
						set in [join [lrange [split $in .] 0 end-1] .]
						incr in_col($in)
					} elseif {[scan $M($what) %d i] == 1} {
						for {} {$i > 0} {incr i -1} {
							set in [join [lrange [split $in .] 0 end-1] .]
						}
						incr in_col($in)
					}
				}

				# Widget Pathfinder (Fourth Sector ;)
				set wpath ""
				set wut ""
				foreach what $those {
					if {[info exists A($what)]} {
						set wut $what
						break
					}
				}
				if {$wut ne ""} { # A widget by any other path would smell just as sweet
					set wpath $A($wut)
					if {$wpath eq ""} {
						if {[info exists L($what)]} {
							set wpath $L($what)
						}
					}
					if {$wpath eq ""} {
						set wpath [namespace tail [dict get $wm $c command]]
						append wpath %gwi%
						#puts wpath!
						#set wpath %gwc%%gwi%
					}
					if {$wpath ne ""} {
						# Substitutions:
						# %gwi%   Gub Widget Integer (widget counter)
						# %gwii%  Gub Widget Integer Integer (gwi + 1)
						# %gwc%   Gub Widget Command
						# %gwC%   Gub Widget Command (titled)
						# %gwn%   Gub Widget Name ("wNamepaceCommand0")
						# %grs%   Gub Random String (6 char random lowercase string)
						# The "!!" (name) operator can set this up
						if {[string first "%g" $wpath] != -1} {
							#set wpath [string map [list %n $n] $wpath]
							set cmd [dict get $wm $c command]
							set cmd [namespace tail $cmd]
							set wpath [string map [list \
								%gwi%  $ofs_n \
								%gwii% [expr {$ofs_n + 1}] \
								%gwc%  $cmd \
								%gwC%  [string totitle $cmd] \
								%gwn%  w[gubMakeName [dict get $wm $c command]]$ofs_n \
								%grs%  [randstr] \
							] $wpath]
						}
					}
				}
				if {$wpath eq ""} {
					#set wpath $lc$n
					set wpath $lc$ofs_n
				}
				set path $in.$wpath
			}; #c in wm if


			if {$c eq " " && $m eq " "} {
				incr in_col($in)
				continue
			}

			if {$is_first_child($in)} {
				incr in_row($in)
				set is_first_child($in) no
			}

			if {![dict exists $wm $c]} continue


			# Start making widget
			set widget [dict create widgetchar $c n $ofs_n path $path]

			# Paths
			set paths($c$n) $path
			if {$m ne " "} {
				set paths($c$m) $path
			}

			# Options, gridoptions, addoptions
			foreach arr {p g q t} optarr {P G Q T} key {options gridoptions addoptions {}} {
				unset -nocomplain $arr
				if {$arr eq "p"} { foreach what $these {
					if {[info exists ${optarr}($what)]} {
						array set $arr [set ${optarr}($what)]
					}
				}} else { foreach what $those {
					if {[info exists ${optarr}($what)]} {
						array set $arr [set ${optarr}($what)]
						break
					}
				}}
				if {![info exists $arr]} continue

				foreach toul $autoul {
					if {![info exists ${arr}($toul)] || [info exists ${arr}(-underline)]} continue
					set optVal [gubUA [set ${arr}($toul)]]
					lassign $optVal ${arr}($toul) _ underline
					if {[llength $optVal] > 1} {
						set ${arr}(-underline) $underline
					}
				}

				dict set widget $key [array get $arr]
				unset $arr; # p,q are used elsewhere as scalars! :/
			}

			# save
			foreach what $those {
				if {![info exists S($what)]} continue
				if {$S($what) eq ""} {
					if {$what in [list $c all]} {
						set name $c$n
					} else {
						set name $what
					}
				} else {
					set name $S($what)
				}
				dict set widget savename $name
				if {[info exists Z($what)]} {
					dict set widget altsave 1
				}
				break
			}

			# Binds for all
			# Copy to widget-specific element so bind processing finds them.
			foreach what [list all] {
				if {![info exists B($what)]} continue
				foreach wut $those {
					if {$wut eq $c} { break } ;# widgetchar has higher specificity
					if {[info exists B($wut)]} continue
					set B($c$n) $B($what)
				}
			}

			# putcmd
			foreach what $those {
				if {![info exists U($what)]} continue
				dict set widget putcmd $U($what)
			}

			# Row force
			if {[dict exists $widget gridoptions -row]} {
				set in_row($in) [dict get $widget gridoptions -row]
				dict unset widget gridoptions -row
			}

			# Column force
			if {[dict exists $widget gridoptions -column]} {
				set in_col($in) [dict get $widget gridoptions -column]
				dict unset widget gridoptions -column
			}

			# Row offset
			foreach what $these {
				if {![info exists E($what)]} continue
				if {$E($what) eq {}} { #snap-back
					set l [string length $in]
					set max -1
					foreach k [dict keys $widgets $in.*] {
						if {[string last . $k] != $l} continue
						if {[dict get $widgets $k col] != $in_col($in)} continue
						if {[dict get $widgets $k row] <= $max} continue
						set max [dict get $widgets $k row]
					}
					set in_row($in) [incr max]
				} elseif {[string index $E($what) 0] in [list - +]} { #relative
					incr in_row($in) $E($what)
				} else { #absolute
					set in_row($in) $E($what)
				}
			}
			# Column offset
			foreach what $these {
				if {![info exists F($what)]} continue
				if {$F($what) eq {}} { #snap-back
					set l [string length $in]
					set max -1
					foreach k [dict keys $widgets $in.*] {
						if {[string last . $k] != $l} continue
						if {[dict get $widgets $k row] != $in_row($in)} continue
						if {[dict get $widgets $k col] <= $max} continue
						set max [dict get $widgets $k col]
					}
					set in_col($in) [incr max]
				} elseif {[string index $F($what) 0] in [list - +]} { #relative
					incr in_col($in) $F($what)
				} else { #absolute
					set in_col($in) $F($what)
				}
			}

			dict set widget col $in_col($in)
			dict set widget row $in_row($in)

			dict set widgets [dict get $widget path] $widget

			if {[Is scrollable [dict get $wm $c]]} {
				lappend scrollables $c$n $x $y
			}
			if {[Is scroller [dict get $wm $c]]} {
				dict set scrollers $x $y $c$n
			}

			# Add to auto_adds if parent is an adder and widget don't have no putcmd ("" counts as a putcmd).
			if {[dict exists $adders $in] && (![dict exists $widget putcmd] || [dict get $widget putcmd] ne "")} {
				dict lappend auto_adds $paths([dict get $adders $in]) $paths($c$n)
			}

			# Add to adders if an adder
			if {[Is adder [dict get $wm $c]]} {
				dict set adders $path $c$n
			}

			# Container ops
			if {[info exists in_x($x)]} {
				set w $in_x($x)
				set al [array get in_x]
				foreach i [lsearch -glob -all $al $w*] {
					set ii [lindex $al [incr i -1]]
					unset in_row($in_x($ii))
					unset in_col($in_x($ii))
					unset in_x($ii)
				}
			}
			# Container override
			if {[set container [Is container [dict get $wm $c]]]} {
				foreach what $those {
					if {[info exists X($what)]} {
						set container false
						break
					}
				}
			}
			if {$container} {
				set in $path
				set in_x($x) $in
				set in_row($in) 0
				set in_col($in) 0
				set is_first_child($in) no
			} else {
				incr in_col($in)
			}
		}; #cm loop
	}; #layout loop

	if {$container} {
		unset in_row($in)
		unset in_col($in)
	}

	# Scrollbars, row/col configures, binds, adds
	array set LL [lreverse [array get L]]

	# Auto-link (setup adjacent) scrollables and scrollers
	# Horizontal scrollbars
	if {$aslH} { foreach {w x y} $scrollables {
		set i [expr {$y + 1}]
		if {![dict exists $scrollers $x $i]} continue
		set s [dict get $scrollers $x $i]
		if {[info exists H($s)]} continue
		if {![info exists paths($s)]} continue
		set p $paths($s)
#		if {![dict exists $widgets $p]} continue
		set widget [dict get $widgets $p]
		set H($s) $w
		dict set widgets $p options     [dict merge [dgwd {} $widget options]     [dict create -orient horizontal]]
		dict set widgets $p gridoptions [dict merge [dgwd {} $widget gridoptions] [dict create -sticky ew]]
#		dict set widgets $p link,options     [dict create -orient horizontal]
#		dict set widgets $p link,gridoptions [dict create -sticky ew]
	}}
	# Vertical scrollbars
	if {$aslV} { foreach {w x y} $scrollables {
		set i [expr {$x + 2}]
		if {![dict exists $scrollers $i $y]} continue
		set s [dict get $scrollers $i $y]
		if {[info exists V($s)]} continue
		if {![info exists paths($s)]} continue
		set p $paths($s)
#		if {![dict exists $widgets $p]} continue
		set widget [dict get $widgets $p]
		set V($s) $w
		dict set widgets $p options     [dict merge [dgwd {} $widget options]     [dict create -orient vertical]]
		dict set widgets $p gridoptions [dict merge [dgwd {} $widget gridoptions] [dict create -sticky ns]]
#		dict set widgets $p link,options     [dict create -orient vertical]
#		dict set widgets $p link,gridoptions [dict create -sticky ns]
	}}

	# Link scrollers (k) and scrollables (v)
	foreach a {H V} d {x y} { foreach {k v} [array get $a] {
		foreach q {k v} { if {![info exists paths([set $q])] && [info exists LL([set $q])]} { set $q $LL([set $q]) } }
		if {!([info exists paths($k)] && [info exists paths($v)])} continue
		set kp $paths($k); set vp $paths($v)
#		if {![dict exists $widgets $kp] || ![dict exists $widgets $vp]} continue
		set kwidget [dict get $widgets $kp]
		set vwidget [dict get $widgets $vp]
		dict set widgets $kp links [list {*}[dgwd {} $kwidget links] -command $paths($v) ${d}view]
		dict set widgets $vp links [list {*}[dgwd {} $vwidget links] -${d}scrollcommand $paths($k) set]
#		dict set widgets $kp link,options [list {*}[dgwd {} $kwidget link,options] -command $paths($v) ${d}view]
#		dict set widgets $vp link,options [list {*}[dgwd {} $vwidget link,options] -${d}scrollcommand $paths($k) set]
	}}

	# Row and column configures
	set rccs [list]
	foreach a {R C} which {row column} { foreach {k v} [array get $a] { foreach o $v {
		if {![info exists paths($k)] && [info exists LL($k)]} { set k $LL($k) }
		if {[info exists paths($k)]} {
			set wl [list $paths($k)]
		} else {
			set wl [lsort -dictionary [dict values [array get paths $k*]]]
		}
		if {[llength $wl] == 0} continue
		if {$o eq ""} { set o "0" }
		if {[set i [string first " " $o]] == -1} {
			set oo " -weight 1"
			set i end
		} else {
			set oo [string range $o $i end]
			incr i -1
		}
		if {$compact} {
			foreach w $wl {
				lappend rccs [list $which $w "[list [ExpandRange [string range $o 0 $i]]]$oo"]
			}
		} else {
			set er [ExpandRange [string range $o 0 $i]]
			foreach w $wl { foreach o $er {
				lappend rccs [list $which $w "$o$oo"]]
			}}
		}
	}}}


	# Binds
	set binds [list]
	foreach {k vv} [array get B] {
		if {![info exists paths($k)] && [info exists LL($k)]} { set k $LL($k) }
		if {![info exists paths($k)]} continue
		if {[info exists X($k)]} continue
		foreach v $vv {
			if {[set i [string first " " $v]] == -1} {
				set seq $v
				set tag [dict get $goob in]
			} else {
				set tag [string range $v 0 $i-1]
				set seq [string trim [string range $v $i+1 end]]
				if {$seq eq ""} {
					set seq $tag
					set tag [dict get $goob in]
				}
			}
			if {[string index $tag 0] in [list . /]} {
				set script "$paths($k) invoke;break"
				if {[string index $seq 0] in [list . /]} {
					set tag $seq
					set seq ""
				}
if 0 {
				if {    ($seq eq "" || [string index $seq end] in [list - +])
				     && [dict exists $widgets $paths($k) options -text]
				     && [dict exists $widgets $paths($k) options -underline]
				} {
					if {$seq eq ""} { set seq "+" }
					set options [dict get $widgets $paths($k) options]
					append seq [string index [dict get $options -text] [dict get $options -underline]]
				}
}
			} else {
				set script $seq
				if {$script eq "{}"} { set script {} }
				set seq $tag
				set tag $paths($k)
			}
if 1 {
			if {    ($seq eq "" || [string index $seq end] in [list - +])
			     && [dict exists $widgets $paths($k) options -text]
			     && [dict exists $widgets $paths($k) options -underline]
			} {
				if {$seq eq ""} { set seq "+"}
				set options [dict get $widgets $paths($k) options]
				append seq [string index [dict get $options -text] [dict get $options -underline]]
			}
}
			if {$seq eq ""} {
				continue
			}
			foreach sequence [gubExpandEventKeys $seq] {
		                lappend binds [list $tag $sequence $script]
			}
		}
	}


	# Adds
	set adds [dict create]
	dict for {k vv} $auto_adds { foreach v $vv {
		if {[dict exists $adds $k] && [lsearch -index 0 [dict get $adds $k] $v] != -1} continue
		dict lappend adds $k [list $v [dgwd {} $widgets $v addoptions]]
	}}

	# References, labels, refmap
	set refmap {}
	set labels [dict create]
	foreach {k v} [array get L] {
		if {![info exists paths($k)]} continue
		dict lappend labels $v $paths($k)
		lappend refmap /$v/ $paths($k)
	}
	foreach {k v} [array get paths] {
		lappend refmap /$k/ $v
	}


	# Save "in"
	set what .
	set these [list $what]
	if {[info exists L($what)]} {
		lappend these $L($what)
	}
	foreach what $these {
		if {![info exists S($what)]} continue
		if {[set s $S($what)] eq ""} { set s in }
		dict set goob save_in $s
		break
	}


	# Put it all together

	if {[dict size $widgets] > 0} {
		dict set goob widgets [dict values $widgets]
	}
	foreach what {rccs binds} {
		if {[llength [set $what]] > 0} { dict set goob $what [set $what] }
	}
	foreach what {adds labels wmchgs} {
		if {[dict size [set $what]] > 0} { dict set goob $what [set $what] }
	}
	foreach what {N F E} {
		dict set goob $what [dict create {*}[array get $what]]
	}

	dict set goob offsets [dict create]
	dict for {k v} [dict get $goob N] {
		if {[info exists O($k)]} { incr v $O($k) }
		if {$v == -1} continue
		dict set goob offsets $k [incr v]
	}

	foreach what {in_row in_col} {
		dict set goob $what [dict create {*}[array get $what]]
	}
	if {[dict exists $goob in_row $initial_in] && $initial_in ne $orig_in} {
		dict set goob in_row $orig_in [dict get $goob in_row $initial_in]
		dict unset goob in_row $initial_in
	}
	if {[dict exists $goob in_col $initial_in] && $initial_in ne $orig_in} {
		dict set goob in_col $orig_in [dict get $goob in_col $initial_in]
		dict unset goob in_col $initial_in
	}

	return [string map $refmap $goob]
}


#---------------------------------------------------------------------
# Assemble
#---------------------------------------------------------------------

proc gubAssemble {guh} {
	variable WidgetMap

	set wm [expr {[dict size [dict get $guh widgetmap]] > 0 ? [dict get $guh widgetmap] : $WidgetMap}]

	set goob [dict get $guh goob]
	if {[dict exists $goob wmchgs]} {
		set wm [dict merge $wm[set wm {}] [dict get $goob wmchgs]]
	}

	set in [dict get $goob in]

	set inline_put  [dict get $guh inline_put]
	set inline_link [dict get $guh inline_link]

	set langs [dict get $guh langs]
	foreach lang $langs {
		set gen_$lang [dict create]
		foreach e [dict get $guh sections] {
			dict set gen_$lang $e [list]
		}
	}

	foreach e [dict get $guh sections] {
		set g_$e [expr {1 ? $e : "create"}]
	}

	set adds [dict create]
	if {[dict exists $goob adds]} { dict for {w toadds} [dict get $goob adds] {
		foreach toadd $toadds {
			dict set adds [lindex $toadd 0] [list $w [lindex $toadd 1]]
		}
	}}

	set slinks [dict create]
	foreach lang $langs {
		dict set slinks $lang [list]
	}

	set put [dict create]
	set add [dict create]

if 0 {
	set weegs [dict create]
	set chils [dict create]
	if {[dict exists $goob widgets]} { foreach widget [dict get $goob widgets] {
		set path [dict get $widget path]
		dict set weegs $path $widget
		dict lappend chils [parentPath $path] $path
	}}
	set order "natural"
#	set order "tree"
	if {$order eq "natural"} {
		set todo [dict keys $weegs]
	} elseif {$order eq "tree"} {
		set stk [list [dict get $guh in]]
		#set stk [dict get $chils [dict get $guh in]]
		set out [list]
		while {[llength $stk] > 0} {
			set w [lindex $stk 0]
			set stk [lrange $stk 1 end]
			lappend out $w
			if {[dict exists $chils $w]} {
				lappend stk {*}[dict get $chils $w]
			}
		}
		set todo $out
	} else {
		return -code error "unknown order: $order"
	}
	foreach weeg $todo {}
}
	if {[dict exists $goob widgets]} { foreach widget [dict get $goob widgets] {
		#if {![dict exists $weegs $weeg]} { continue }
		#set widget  [dict get $weegs $weeg]
		set path    [dict get $widget path]
		set upath   [d2u $path]
		set ppath   [parentPath $path]
		set uppath  [d2u $ppath]
		set command [dict get $wm [dict get $widget widgetchar] command]
		set options [dgwd [list] $widget options]

		set cmds [dict create]
		dict set cmds tcl    $command
		dict set cmds ruby   [tc2rc $command]
		dict set cmds lua    [tc2lc $command]
		dict set cmds perl   [tc2pc $command]
		dict set cmds python [tc2yc $command]

		set ppaths [dict create]
		dict set ppaths tcl    $ppath
		dict set ppaths ruby   $uppath
		dict set ppaths lua    $uppath
		dict set ppaths perl   [tp2pp $ppath]
		dict set ppaths python $uppath

		set paths [dict create]
		dict set paths tcl    $path
		dict set paths ruby   $upath
		dict set paths lua    $upath
		dict set paths perl   [tp2pp $path]
		dict set paths python $upath

		set links [dict create]
		foreach lang $langs {
			dict set links $lang [list]
		}
		set lnks $links
		if {[dict exists $widget links]} { foreach {opt win cmd} [dict get $widget links] {
			set uwin [d2u $win]
			dict lappend links tcl    $opt $win\ $cmd
			dict lappend links ruby   $opt proc\{|*args|$uwin.$cmd\(*args)\}
			dict lappend links lua    $opt $uwin..\"\ $cmd\"
			dict lappend links perl   $opt \[\$$uwin,'$cmd'\]
			dict lappend links python $opt $uwin.$cmd
		}}
		if {$inline_link} {
			set lnks $links
		} else {
			foreach lang $langs {
				if {[llength [set l [dict get $links $lang]]] == 0} { continue }
				dict lappend slinks $lang [dict get $paths $lang] $l
			}
		}

		set opts [dict create]
		dict set opts tcl    [to2to [concat $options [dict get $lnks    tcl]]]
		dict set opts ruby   [to2ro [concat $options [dict get $lnks   ruby]]]
		dict set opts lua    [to2lo [concat $options [dict get $lnks    lua]]]
		dict set opts perl   [to2po [concat $options [dict get $lnks   perl]] {}]
		dict set opts python [to2yo [concat $options [dict get $lnks python]]]

		set wigs [dict create]
		dict set wigs tcl    [dict get $cmds tcl]\ [dict get $paths tcl][dict get $opts tcl]
		dict set wigs ruby   [dict get $cmds ruby]([dict get $ppaths ruby][dict get $opts ruby])
		dict set wigs lua    ltk.[dict get $cmds lua]\{[dict get $ppaths lua][dict get $opts lua]\}
		dict set wigs perl   \$[dict get $ppaths perl]->[dict get $cmds perl]([dict get $opts perl])
		dict set wigs python [dict get $cmds python]([dict get $ppaths python][dict get $opts python])

		dict for {lang sv} [GubAssembleSetVar $widget $wigs $guh] {
			set w_$lang [dict get $sv create]
			set    save [dict get $sv save]
			if {$save ne ""} {
				dict lappend gen_$lang ${g_save} $save
			}
		}
		set putcmd [dgwd grid $widget putcmd]
		if {$inline_put && $putcmd ne ""} {
			set w_tcl "\[$w_tcl\]"
		}

		set putpaths [dict create]
		if {$inline_put} {
			foreach lang $langs {
				dict set putpaths $lang [set w_$lang]
			}
		} else {
			dict set putpaths tcl    $path
			dict set putpaths ruby   $upath
			dict set putpaths lua    $upath
			dict set putpaths perl   [tp2pp $path]
			dict set putpaths python $upath
		}

		foreach lang $langs {
			dict set put $lang ""
			dict set add $lang ""
		}

		if {[dict exists $adds $path]} {
			lassign [dict get $adds $path] add_w add_opts
			set uadd_w [d2u $add_w]
			dict set add tcl    "$add_w add [dict get $putpaths tcl]"
			dict set add ruby   "$uadd_w.add(($upath=[dict get $putpaths ruby])"
			dict set add lua    "$uadd_w.add\{[dict get $putpaths lua]"
			dict set add perl   "\$[tp2pp $add_w]->add\("
			dict set add python "$uadd_w.add("
			if {[Is container [dict get $wm [dict get $widget widgetchar]]]} {
				dict append add perl   "my \$$upath=[dict get $putpaths perl]"
				dict append add python "$upath:=[dict get $putpaths python]"
			} else {
				dict append add perl   [dict get $putpaths perl]
				dict append add python [dict get $putpaths python]
			}
			if {$add_opts ne ""} {
				dict append add tcl    [to2to $add_opts]
				dict append add ruby   [to2ro $add_opts]
				dict append add lua    [to2lo $add_opts]
				dict append add perl   [to2po $add_opts]
				dict append add python [to2yo $add_opts]
			}
			dict append add ruby   ")"
			dict append add lua   "\}"
			dict append add perl   ")"
			dict append add python ")"
		} elseif {$putcmd eq ""} {
			if 1 {
			foreach lang $langs {
				dict lappend gen_$lang create [set w_$lang]
				#dict lappend gen_$lang create [set putpath_$lang]
			}
			} else {
			dict lappend gen_tcl    create $w_tcl
			dict lappend gen_ruby   create $w_ruby
			dict lappend gen_lua    create $w_lua
			dict lappend gen_perl   create $w_perl
			dict lappend gen_python create $w_python
			}
		} else {
			set row [dict get $widget row]
			set col [dict get $widget col]
			if {[dict exists $widget gridoptions]} {
				set gridoptions [dict get $widget gridoptions]
			} else {
				set gridoptions ""
			}
			dict set put tcl  "$putcmd [dict get $putpaths tcl][to2to $gridoptions] -row $row -column $col"
			dict set put ruby "($upath=[dict get $putpaths ruby]).$putcmd\(:row=>$row,:column=>$col[to2ro $gridoptions])"
			dict set put lua  "$upath=[dict get $putpaths lua] ltk.$putcmd\{$upath,row=$row,column=$col[to2lo $gridoptions]\}"
			if {1 ||[Is adder [dict get $wm [dict get $widget widgetchar]]]} {
				dict set put perl   "(my \$$upath=[dict get $putpaths perl])"
				dict set put python "($upath:=[dict get $putpaths python])"
			} else {
				dict set put perl   [dict get $putpaths perl]
				dict set put python [dict get $putpaths python]
			}
			dict append put perl   "->g_$putcmd\(-row=>$row,-column=>$col[to2po $gridoptions])"
			dict append put python ".$putcmd\(row=$row,column=$col[to2yo $gridoptions])"
		}

		if {$inline_put} {
			set cmds [list ${g_create} ${g_create}]
		} else {
			foreach lang $langs {
				dict lappend gen_$lang ${g_create} [set w_$lang]
			}
			set cmds [list ${g_put} ${g_add}]
		}
		set hows [list put add]
		foreach lang $langs { foreach how $hows cmd $cmds {
			set howd [set $how]
			if {[dict get $howd $lang] ne ""} {
				dict lappend gen_$lang $cmd [dict get $howd $lang]
			}
		}}
	}}; # widgets loop

	if {!$inline_link} {
		foreach {wig opts} [dict get $slinks tcl] {
			#dict lappend gen_tcl ${g_link} "$wig configure [join $opts]"
			dict lappend gen_tcl ${g_link} "$wig configure[to2to $opts]"
		}
		foreach {wig opts} [dict get $slinks ruby] {
			foreach {n v} $opts {
				dict lappend gen_ruby ${g_link} "$wig\['[string range $n 1 end]'\]=$v"
			}
		}
		foreach {wig opts} [dict get $slinks lua] {
			dict lappend gen_lua ${g_link} "ltk.wcmd($wig)\{'configure'[to2lo $opts]\}"
		}
		foreach {wig opts} [dict get $slinks perl] {
			dict lappend gen_perl ${g_link} "\$$wig->configure([to2po $opts {}])"
		}
		foreach {wig opts} [dict get $slinks python] {
			dict lappend gen_python ${g_link} "$wig.configure([to2yo $opts {}])"
		}
	}

	if {[dict exists $goob rccs]} { foreach rcc [dict get $goob rccs] {
		lassign $rcc which window options
		set w [expr {$window eq "" ? $in : ""}]
		#set w [expr {$window eq "" && $w eq "" ? "." : $w}]
		set wwindow ${w}$window
		set uwwindow [d2u $wwindow]
		dict lappend gen_tcl    ${g_rcc} "grid ${which}configure ${w}${window} $options"
		dict lappend gen_ruby   ${g_rcc} "TkGrid.${which}configure($uwwindow,[lindex $options 0][to2ro [lrange $options 1 end]])"
		dict lappend gen_lua    ${g_rcc} "ltk.grid{'${which}configure','${w}${window}',[lindex $options 0][to2lo [lrange $options 1 end]]}"
		dict lappend gen_perl   ${g_rcc} "\$[tp2pp $wwindow]->g_grid_${which}configure([lindex $options 0][to2po [lrange $options 1 end]])"
		dict lappend gen_python ${g_rcc} "$uwwindow.${which}configure([lindex $options 0][to2yo [lrange $options 1 end]])"
	}}; # rccs

	if {[dict exists $goob binds]} { foreach binding [dict get $goob binds] {
		lassign $binding window seq script
		if {$window eq ""} {
			set win $in
		} elseif {$window eq ".."} {
			set win [parentPath $in [expr {[string length $window]-1}]]
		} elseif {$window eq "\\."} {
			set win "."
		} else {
			set win $window
		}
		dict lappend gen_tcl    ${g_bind} "bind $win $seq [braceIfNeeded $script]"
		dict lappend gen_ruby   ${g_bind} "[d2u $win].bind('[string range $seq 1 end-1]') {[d2u $script]}"
		dict lappend gen_lua    ${g_bind} "ltk.bind{'$win','$seq',{$script}}"
		dict lappend gen_perl   ${g_bind} "\$[tp2pp $win]->g_bind('$seq',sub{[d2u $script]})"
		dict lappend gen_python ${g_bind} "[d2u $win].bind('$seq',lambda e:{[d2u $script]})"
	}}; # binds


	if {[dict exists $goob save_in]} {
		set emptyWigs [dict create]
		foreach lang $langs {
			dict set emptyWigs $lang {}
		}
		dict for {lang sv} [GubAssembleSetVar [dict create path $in savename [dict get $goob save_in] altsave 1] $emptyWigs $guh] {
			dict lappend gen_$lang save [dict get $sv save]
		}
	}

	if 1 {
	set gen [dict create]
	foreach lang $langs { dict set gen $lang [set gen_$lang] }
	return [GubAssembleScript $gen $guh]
	} else {
	set src [dict create]
	foreach lang $langs { dict set src $lang [GubAssembleScripta [set gen_$lang] $guh $lang] }
	return $src
	}
}

proc GubAssembleScripta {gen guh {lang {}}} {
	if {$lang eq {}} {
		set lang [dict get $guh lang]
	}
	set c [dict get $guh eolChar]
	set lc ""
	if {$lang eq "perl"} {
		set c ";$c[set c {}]"
		set lc ";"
	}
	return [join [lmap e [dict get $guh sections] { set l [dict get $gen $e]; if {[llength $l] == 0} continue; join $l $c }] $c]$lc
}

proc GubAssembleScript {gen guh} {
	set src [dict create]
	set sections [dict get $guh sections]
	foreach lang [dict get $guh langs] {
		set c [dict get $guh eolChar]
		set lc ""
		if {$lang eq "perl"} {
			set c ";$c"
			set lc ";"
		}
		dict set src $lang [join [lmap s $sections { set l [dict get $gen $lang $s]; if {[llength $l] == 0} continue; join $l $c }] $c]$lc
	}
	return $src
}

proc GubAssembleSetVar {widget wigs guh} {
	set langs [dict get $guh langs]
	set create [dict create]
	set   save [dict create]
	foreach lang $langs {
		dict set create $lang [dict get $wigs $lang]
		dict set   save $lang ""
	}
	if {![dict exists $widget savename]} {
		set sv [dict create]
		foreach lang $langs {
			dict set sv $lang [dict create create [dict get $create $lang] save [dict get $save $lang]]
		}
		return $sv
	}
	set varname     [dict get $guh varname]
	set vartype     [dict get $guh vartype]
	set setcmd      [dict get $guh setcmd] 
	set inline_save [dict get $guh inline_save]
	set savename    [dict get $widget savename]
	set altsave     [dgwd 0 $widget altsave]
	set path        [dict get $widget path]
	if {$inline_save && !$altsave} {
		dict set create tcl "\[[dict get $create tcl]\]"
	}
	if {$varname eq ""} {
		dict set save tcl    "$setcmd $savename "
		dict set save ruby   "($savename="
		dict set save python "($savename:="
		if {$inline_save && !$altsave} {
			dict append save tcl    [dict get $create tcl]
			dict set  create tcl    [dict get $save tcl]
			dict set    save tcl    ""
			dict append save ruby   [dict get $create ruby])
			dict set  create ruby   [dict get $save ruby]
			dict set    save ruby   ""
			dict append save python [dict get $create python])
			dict set  create python [dict get $save python]
			dict set    save python ""
		} else {
			dict append save tcl    $path
			dict append save ruby   "'$path')"
			dict append save python "'$path')"
		}
	} else {
		if {$vartype eq "dict"} {
			set setcmd_tcl "dict set"
			dict set save tcl "$setcmd_tcl $varname $savename"
			if {$inline_save && !$altsave} {
				dict append save tcl " " [dict get $create tcl]
			} else {
				dict append save tcl " " $path
			}
			if {!$altsave} {
				if {(![dict exists $widget putcmd] || [dict get $widget putcmd] ne "") && $inline_save && [dict get $guh inline_put]} {
					dict set save tcl "dict get \[[dict get $save tcl]\] $savename"
				}
			}
			if {$inline_save && !$altsave} {
				dict set create tcl [dict get $save tcl]
				dict set save tcl ""
			}
		} elseif {$vartype eq "array"} {
			dict set save tcl "$setcmd $varname\($savename)"
			if {$inline_save && !$altsave} {
				dict set create tcl "[dict get $save tcl] [dict get $create tcl]"
				dict set save tcl ""
			} else {
				dict append save tcl " " $path
			}
		} else {
			return -code error "unknown vartype: $vartype"
		}
		dict set save ruby   "($varname\[\"$savename\"\]="
		dict set save lua    "$varname\[\"$savename\"\]="
		dict set save perl   "\$$varname\{'$savename'\}="
		dict set save python "($varname\[\"$savename\"\]:="
		if {$inline_save && !$altsave} {
			dict set create ruby   [dict get $save   ruby][dict get $create   ruby])
			dict set create lua    [dict get $save    lua][dict get $create    lua]
			dict set create perl   [dict get $save   perl][dict get $create   perl]
			dict set create python [dict get $save python][dict get $create python])
			dict set   save ruby   ""
			dict set   save lua    ""
			dict set   save perl   ""
			dict set   save python ""
		} else {
			dict append save ruby   "'$path'"
			dict append save lua    "'$path'"
			dict append save perl   "'$path'"
			dict append save python "'$path'"
		}
	}
	set sv [dict create]
	foreach lang $langs {
		dict set sv $lang [dict create create [dict get $create $lang] save [dict get $save $lang]]
	}
	return $sv
}


#---------------------------------------------------------------------
# Misc
#---------------------------------------------------------------------

#
# Dict Get With Default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

#
# Underline Ampersand
# Convert ampersand in text to -underline option
#
proc gubUA {text} {
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}

#
# Underlined char or "" if none
#
proc gubUlChar {s} {
	set u [gubUA $s]
	if {[llength $u] == 1} {
		return ""
	} else {
		return [string index [lindex $u 0] [lindex $u end]]
	}
}

#
# "one of" message
#
proc oneOf {l} {
	if {[llength $l] < 2} { return [join $l] }
	return [strcat "one of " [join [lrange $l 0 end-1] ", "] [expr {[llength $l] > 2 ? "," : ""}] " or " [lindex $l end]]
}

#
# Random [a-z] string
#
proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

#
# Default value of arg of caller
#
proc MyArgDef {arg} {
	set proc [lindex [info level -1] 0]
	if {$arg ni [info args $proc]} { return "" }
	if {[info default $proc $arg val]} { return $val }
	return ""
}

proc ArrayValues {array {pattern *}} {
	upvar 1 $array a
	set res [list]
	foreach key [array names a $pattern] {
		lappend res $a($key)
	}
	return $res
}

proc Is {what widgetmapelement} {
	if {![dict exists $widgetmapelement properties]} { return 0 }
	set is [expr {$what in [dict get $widgetmapelement properties]}]
	# Special case: adders are also containers
	set is [expr {$is || ($what eq "container" && [Is adder $widgetmapelement])}]
	return $is
}

proc LooksLikeARange {s} {
	expr {[string first "-" $s] > 1 || [string first "*" $s] != -1 || [string first "+" $s] != -1}
}

proc ExpandRange {range} {
	set expanded [list 0]
	foreach chunk [split $range ,] {
		if {[set inc [lassign [split $chunk *] chunk]] eq ""} { set inc 1 }
		if {[string first "+" $chunk] != -1} {
			lassign [split $chunk -] c1 c2
			if {[string index $c1 0] eq "+"} { set c1 [expr {[lindex $expanded end] + [string range $c1 1 end]}] }
			if {[string index $c2 0] eq "+"} { set c2 [expr {$c1+ [string range $c2 1 end]}] }
			set chunk [join [list $c1 {*}$c2] -]
		}
		foreach hi [lassign [split $chunk -] chunk] {
			for {set lo $chunk; set chunk {}} {$lo <= $hi} {incr lo $inc} { lappend chunk $lo }
		}
		lappend expanded {*}$chunk
	}
	return [lrange $expanded 1 end]
}

proc ExpandImplied {n} {
	set cen {}
	set wc "" ;# malformed input may cause error below
	if {[string index $n 0] eq "+"} { return $n }
	foreach s [split $n ,] {
		set ch [string index $s 0]
		if {$ch eq "+"} {
			set comma [expr {[llength $cen] == 0 ? "" : ","}]
			lset cen end [lindex $cen end]$comma$s
			continue
		}
		if {[string is digit $ch] || $ch eq "<"} {
			set s $wc$s
		} else {
			set wc [string index $s 0]
		}
		lappend cen $s
	}
	set en {}
	foreach s $cen {
		if {![LooksLikeARange $s]} { lappend en $s; continue }
		set wc [string index $s 0]
		foreach i [ExpandRange [string range $s 1 end]] { lappend en $wc$i }
	}
	return [join $en ,]
}

proc ExpandLtGt {s} {
	set s ,$s,
	while {[regexp -indices {,([^,]+?)<(.+?)>,} $s alli nami vali]} {
		lassign $nami l r; set nam [string range $s $l $r]
		lassign $vali l r; set val [string range $s $l $r]
		set n {}
		foreach b [split $val ""] {
			lappend n $nam$b
		}
		lassign $alli l r
		set s [string replace $s [incr l] [incr r -1] [join $n ,]]
	}
	return [string range $s 1 end-1]
}

proc Expand {n} {
	split [ExpandLtGt [ExpandImplied $n]] ,
}

proc gubExpandEventKeys {event} {
	set n [expr {[string match "<<?*>>" $event] ? 2 : 1}]
	lassign [list [string repeat "<" $n] [string repeat ">" $n]] lb rb
	set event [string trim $event "<>"]
	if {[llength [set l [split $event "+"]]] > 1 && [string length [lindex $l end]] == 1} {
		lassign $l mod key
		if {$mod ne ""} { append mod "-" }
		set f "$lb$mod%s$rb"
		set k [string tolower $key]
		set e [list [format $f $k]]
		if {[set K [string toupper $key]] ne $k} {
			lappend e [format $f $K]
		}
		return $e
	}
	return [list $lb$event$rb]
}

proc gubOffsets {{wpaths {}}} {
	set offsets [dict create]
	foreach w $wpaths {
 		if {![regexp {^(.)(\d+)$} [lindex [split $w .] end] -> wc wn]} continue
		if {$wn >= [dgwd -1 $offsets $wc]} { dict set offsets $wc [incr wn] }
	}
	return $offsets
}

proc parentPath {win {level 1}} {
	return .[string trimleft [join [lrange [split $win .] 0 end-$level] .] .]
}

proc GobIsGohb {gob} { regexp {(?:^|\n|;)[[:space:]]*>[[:space:]]*?.} $gob }

proc cleanGob {gob} {
	# Remove escaped newlines
	set a [regsub -all {((?:[^\\]|^)(?:\\\\)*)\\(?:\n|$)} $gob {\1}]
	# Remove comments
	set b [regsub -all -line {(?:^|;)[[:space:]]*?#.*?$} $a {}]
	# Replace unescaped semicolons with newlines
	set c [regsub -all {((?:[^\\]|^)(?:\\\\)*);} $b "\\1\n"]
	# Replace escaped semicolons with semicolons
	#set d [regsub -all {((?:[^\\]|^)(?:\\\\)*)\\;} $c {\1;}]
	set d [regsub -all {((?:[^\\]|^|)(?:\\\\)*)\\;} $c {\1;}]
	# Remove leading tabs
	set e [regsub -all -line {^\t+} $d ""]
	# Remove trailing whitespace (side effect: compress runs of newlines)
	set f [regsub -all -line {[[:space:]]+$} $e ""]
	# Remove first char if newline
	set g [string trimleft $f "\n"]
	return $g
}

proc makeReady {gob} { split [cleanGob $gob] \n }

proc optsLenMustBeEven {name opts} {
	if {[llength $opts] % 2 == 0} { return $opts }
	return -code error "$name options ($opts) must have an even number of elements"
}

proc continueIfLenOdd {opts} {
	if {[llength $opts] % 2 == 0} { return $opts }
	return -code continue
}

proc checkOpt {name opt} {
	if {	($name eq "guhb"     && $opt ni [set o [list auto yes no]])
	     || ($name eq "selector" && $opt ni [set o [list - |]])
	} {
		return -code error "unknown option \"$opt\": must be [oneOf $o]"
	}
}

proc CheckArg {guh name} {
	set opt [dict get $guh $name]
	if {$name eq "guhb"} { checkOpt $name $opt }
}

proc gubMakeName {cmd} {
	join [lmap e [split $cmd :] {
		if {$e eq {}} continue
		string totitle $e
	}] ""
}

proc braceIfNeeded {script} {
	if {[string index $script 0] ne "\{" && [string index $script end] ne "\}"} {
		return "{$script}"
	} else {
		return $script
	}
}

if 0 {proc ladj {s} {
	if {[string first "\{" $s] == -1} {
		string map [list \\ \\\\] $s
	} else {
		regsub -all {(\}.*?)\\(.*?\{)} $s {\1\\\\\2}
	}
}}

proc gubConfigureWidget {w args} {
	if {[llength $args] > 0} { $w configure {*}$args }
	return $w
}

#
# 8.5 doesn't have [string cat]
#
proc strcat {args} {
	return [join $args ""]
}

#
# 8.5 doesn't have [lmap]
#
if {[llength [info commands lmap]] == 0} { proc lmap args {
	# 8.5 doesn't have lmap
	# https://wiki.tcl-lang.org/page/lmap+forward+compatibility
	set body [lindex $args end]
	set args [lrange $args 0 end-1]
	set n 0
	set pairs [list]
	foreach {varnames listval} $args {
		set varlist [list]
		foreach varname $varnames {
			upvar 1 $varname var$n
			lappend varlist var$n
			incr n
		}
		lappend pairs $varlist $listval
	}
	set temp [list]
	foreach {*}$pairs {
		lappend temp [uplevel 1 $body]
	}
	set temp
}}; # lmap


#---------------------------------------------------------------------
# Lang
#---------------------------------------------------------------------

#
# Dot To Underscore
#
proc d2u {s} {
	return [string map [list . _] $s]
}

#
# Tcl Option To Tcl Option
#
proc to2to {options {firstChar { }}} {
	if {[llength $options] == 0} { return "" }
	return $firstChar$options
}

#
# Tcl Option To Ruby Option
#
proc to2ro {options {firstChar {,}}} {
	if {[llength $options] == 0} { return "" }
	return ${firstChar}[join [lmap {n v} $options {
		set q [expr {$n in [list -weight -command -xscrollcommand -yscrollcommand] ? "" : "'"}]
		strcat ":[string range $n 1 end]=>$q$v$q"
	}] ,]
}

#
# Tcl Option To Lua Option
#
proc to2lo {options {firstChar {,}}} {
	if {[llength $options] == 0} { return "" }
	return ${firstChar}[join [lmap {n v} $options {
		set q [expr {$n in [list -weight -command -xscrollcommand -yscrollcommand] ? "" : "'"}]
		strcat "[string range $n 1 end]=$q$v$q"
	}] ,]
}

#
# Tcl Option To Perl Option
#
proc to2po {options {firstChar {,}}} {
	if {[llength $options] == 0} { return "" }
	return ${firstChar}[join [lmap {n v} $options {
		set q [expr {$n in [list -weight -command -xscrollcommand -yscrollcommand] ? "" : "'"}]
		strcat "$n=>$q$v$q"
	}] ,]
}

#
# Tcl Option To Python Option
#
proc to2yo {options {firstChar {,}}} {
	if {[llength $options] == 0} { return "" }
	return ${firstChar}[join [lmap {n v} $options {
		set q [expr {$n in [list -weight -sticky -orient -command -xscrollcommand -yscrollcommand] ? "" : "'"}]
		if {$n in [list -orient -sticky]} { set v [string toupper $v[set v {}]] }
		strcat "[string range $n 1 end]=$q$v$q"
	}] ,]
}

#
# Tcl Command To Ruby Call
#
proc tc2rc {c} {
	set c [string map {:: .} [string trim $c[set c {}] :]]
	set l [split $c .]
	lset l end [string totitle [lindex $l end]]
	set ns [lindex $l 0]
	if {$ns eq "tk"} {
		set l [lrange $l[set l {}] 1 end]
	} elseif {$ns eq "ttk"} {
		lset l 0 Tile
		if {[lindex $l end] in [list Button]} {
			lset l end T[lindex $l end]
		}
	}
	return Tk::[join $l ::].new
}

#
# Tcl Command To Lua Call
#
proc tc2lc {c} {
	set c [string map {:: .} [string trim $c[set c {}] :]]
	set l [split $c .]
	set ns [lindex $l 0]
	if {$ns eq "tk"} {
		set l [lrange $l[set l {}] 1 end]
	} elseif {$ns eq "ttk"} {
		set l [lrange $l[set l {}] 1 end]
	}
	return [join $l "moo"]
}

#
# Tcl Command To Perl Call
#
proc tc2pc {c} {
	return new_[string map {: _} $c]
}

#
# Tcl Command To Python Call
#
proc tc2yc {c} {
	set c [string map {:: .} [string trim $c :]]
	set l [split $c .]
	lset l end [string totitle [lindex $l end]]
	set ns [lindex $l 0]
	if {$ns eq "tk"} {
		set l [lrange $l[set l {}] 1 end]
		if {[llength $l] == 1 && [lindex $l 0] eq "Panedwindow"} {
			set l [list "PanedWindow"]
		}
	}
	return [join $l .]
}


#
# Tcl Path To Perl Path
#
proc tp2pp {path} {
	return [expr {$path eq "." ? "root" : [d2u $path]}]
}

#
# Tcl Namespace To Python Namespace
#
#proc tns2yns {ns} { string map {:: .} [string trim $ns :] }

proc gubBoilerplateHead {lang} {
	return [switch -exact -- $lang {
	tcl    { strcat {package require Tk} }
	ruby   { strcat {require 'tk'} \n {require 'tkextlib/tile'} \n {gui={}} \n {_ = TkRoot.new()} }
	lua    { strcat {require 'ltk'} \n {_ = nil} }
	python { strcat {from tkinter import *} \n {from tkinter import ttk} \n {gui={}} \n {_ = Tk()} }
	perl   { strcat {use strict;} \n {use Tkx;} \n {my @gui;} \n {my $root = Tkx::widget->new(".");} }
	default explode
	}]
}

proc gubBoilerplateTail {lang} {
	return [switch -exact -- $lang {
	tcl    { strcat {#vwait forever} }
	ruby   { strcat {Tk.mainloop()} }
	lua    { strcat {ltk.mainloop()} }
	python { strcat {_.mainloop()} }
	perl   { strcat {Tkx::MainLoop();} }
	default explode
	}]
}

proc gubBoilerplate {lang} {
	strcat [gubBoilerplateHead $lang] \n\n {#@source@} \n\n [gubBoilerplateTail $lang]
}

proc gubWrapSource {src lang} {
	string map [list #@source@ $src] [gubBoilerplate $lang]
}


#---------------------------------------------------------------------
# Gub Level -1
#---------------------------------------------------------------------

proc gubSaveGob {widgetChar args} {
	set s ""; set i -1
	foreach a $args { append s / $widgetChar [incr i] " " $a \n }
	return $s
}

proc gubButtonGob {text args} {
	set args [list $text {*}$args]
	set index 0
	set window ""
	set modifiers [list ]
	set buttonchar b
	set gob ""
	set ai -1
	set state opt
	foreach arg $args {
		incr ai
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-first      -
			-window     -
			-modifiers  -
			-buttonchar { set state $arg }
			--          { incr ai; break }
			default     { break }
			}; # hctiws
		}
		-first      { set state opt; set index      $arg }
		-window     { set state opt; set window     $arg }
		-modifiers  { set state opt; set modifiers  $arg }
		-buttonchar { set state opt; set buttonchar $arg }
		}; # hctiws
	}
	set window [string trim $window]
	if {$window eq ""} {
		set window " "
	} else {
		set window " $window "
	}
	incr index -1
	foreach text [lrange $args $ai end] {
		incr index
		set buttonOpts [gubUA $text]
		append gob . ${buttonchar} ${index} " -text " $buttonOpts ";"
		if {[llength $buttonOpts] > 1 && [llength $modifiers] > 0} {
			set key [string index [lindex $buttonOpts 0] [lindex $buttonOpts end]]
			foreach mod $modifiers {
				append gob & ${buttonchar} ${index} ${window} ${mod}+${key} ";"
			}
		}
	}
	return $gob
}

proc gubSubI {s {i 0} {x ""}} {
	if {[string first "%g" $s] == -1} { return $s }
	# Substitutions:
	# %g?wi%   Gub Widget Integer
	# %g?wii%  Gub Widget Integer Integer (gwi + 1)
	return [string map [list \
		%g${x}i%  $i \
		%g${x}ii% [expr {$i + 1}] \
	] $s]
}

proc guhbSubI {s {i 0}} { gubSubI $s $i h }

proc gubOpListGob {op todo property properties} {
	set ldiff [expr {[llength $todo] - [llength $properties]}]
	if {$ldiff > 0} { lappend properties {*}[lrepeat $ldiff [lindex $properties end]] }
	set l {}
	set gi 0
	foreach td $todo p $properties {
		lappend l [strcat $op $td " " $property " " [gubSubI $p $gi]]
		incr gi
	}
	return [join $l \n]
}

proc gubPropGob {widgetchar property args} {
	set l {}
	set gi 0
	foreach a $args {
		lappend l [strcat . $widgetchar $gi " " $property " " [gubSubI $a $gi]]
		incr gi
	}
	return [join $l \n]
}

proc gubBindGob {w ev z script} {
	set l {}
	set gi 0
	foreach q $z {
		lappend l [strcat & $w " " $ev$q " " [gubSubI $script $gi]]
		incr gi
	}
	return [join $l \n]
}

proc gubAddGob {to toadd property properties} {
	set l {}
	set gi 0
	foreach ta $toadd p $properties {
		lappend l [strcat +$to " " $ta " " $property " " [gubSubI $p $gi]]
		incr gi
	}
	return [join $l \n]
}

proc gubWrapToggleGob {w t args} {
	set l {}
	foreach {w t} [list $w $t {*}$args] {
		lappend l [strcat . $w " " "-command {/$t/ configure -wrap \[expr {\[/$t/ cget -wrap\] eq \"none\" ? \"word\" : \"none\"}\]}"]
	}
	return [join $l \n]
}

proc gubMultiGob {n g {pre ""}} {
	set gob $pre
	regexp {^ *} $pre spaces
	if {$pre ne ""} {
		append gob " "
		append spaces "  "
	}
	append gob $g
	append gob \n
	for {set i 1} {$i < $n} {incr i} {
		append gob ";$spaces$g" \n
	}
	return $gob
}


#---------------------------------------------------------------------
# Autoscrollbar procs
#---------------------------------------------------------------------

proc initas {} { foreach k {tk ttk} wc {s S} {
	namespace eval $k [string map [list @wc@ $wc] {
		proc autoscrollbar {w args} {
			set w [[dict get [dict get [[namespace parent]::gubWidgetMap] @wc@] command] $w {*}$args]
			::autoscroll::autoscroll $w
			return $w
		}
	}]}
	return yes
}


#---------------------------------------------------------------------
# Setup
#---------------------------------------------------------------------

proc setup {} { if {!([require] && [initas] && [initvars] && [export] && [provide])} { error "Something went wrong." } }

namespace current}]::setup


# EOF

# THATCH /pkgs/dsc/meb-0.5.tm size 16603 mtime 1659823182 modbase /pkgs/dsc
#
# Meb - MEnuBuilder
#
# Stuart Cassoff
#
# Spring 2012
# Version 0.1
#
# Summer 2012
# Spring 2015
# Version 0.2
#
# Spring 2015
# Version 0.3
#
# Winter 2018
# Spring 2018
# Version 0.4
#
# Summer 2022
# Version 0.5
#

[namespace eval meb {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	meb
		title	Meb
		descr	{Menu Builder}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.5
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2012,2015,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


#---------------------------------------------------------------------
# Package declaration, exported procs
#---------------------------------------------------------------------

proc provide {} {
	package provide meb 0.5
	return on
}

proc require {} {
	package require Tcl 8.6
	return true
}

proc export {} {
	namespace export {*}{
		meb
		mebMoobKeys mebItemKeys
		mebLangs mebTLangs
		mebUA
	}
	return yes
}

#---------------------------------------------------------------------
# Meb
#---------------------------------------------------------------------

proc mebMoobKeys {} {
	return [list items script saves]
}

proc mebItemKeys {} {
	return [list autocmd index inName inNumber level options save spaces type]
}

proc mebTLangs {} { list &Tcl R&uby &Lua &Perl P&ython }
proc mebLangs  {} { list  tcl  ruby  lua  perl  python }

proc meb {args} {
	if {[llength $args] == 0} { return "" }
	set ns ::
	set menuname .m
	set names 0
	set autocmd 1
	set menucmd menu
	set menu 1
	set menuopts {}
	set varname gui
	set vartype dict
	set state opt
	set mob [lindex $args end]
	foreach arg [lrange $args 0 end-1] {
		switch -- $state {
		opt {
			switch -- $arg {
			-a  - -array     { set vartype array }
			-n  - -names     { set names   1     }
			-na - -noautocmd { set autocmd 0     }
			-nm - -nomenu    { set menu    0     }
			-mc - -menucmd   { set state -mc     }
			-mn - -menuname  { set state -mn     }
			-mo - -menuopts  { set state -mo     }
			-ns - -namespace { set state -ns     }
			-vn - -varname   { set state -vn     }
			default { error "unknown opt \"$arg\"" }
			}
		}
		-mc { set state opt; set menucmd  $arg }
		-mn { set state opt; set menuname $arg }
		-mo { set state opt; set menuopts $arg }
		-ns { set state opt; set ns       $arg }
		-vn { set state opt; set varname  $arg }
		}
	}
	set ns ::[string trimleft $ns :]
	set ns [string trimright $ns :]::
	set items {}
	array set vars {}
	set spaces {}

	# Parse
	foreach l [split $mob \n] {
		set leadin ""; set rest ""; set space ""; set text ""; set comments ""
		regexp {^([[:space:]]*[;.])?(.*)$} $l -> leadin rest
		regexp {^([[:space:]]*)(.*)$} $rest -> space text
		set text [regsub -all -- {(-[^\s]+?)\s+#([0-9a-fA-F]{6})} $text {\1 `\2}]
		regexp {^(.*?)(#.*)$} $text -> text comments
		set text [regsub -all -- {(-[^\s]+?)\s+`([0-9a-fA-F]{6})} $text {\1 #\2}]
		if {$text eq ""} { continue }
		set space [string trimleft $space \t]
		set text [string trimright $text]
		set type command
		set usertype ""
		set options [dict create]
		set opts {}
		set ol [dict create spaces [string length $space]]

		if {[regexp {^[-=]+$} $text]} {
			set type separator
		} else {
			regexp {^(.+?)[[:space:]]+([-+].*)?$} $text -> text opts
			foreach {n v} [linsert [mebUA $text] 0 -label] {
				dict set options $n $v
			}
		}
		foreach {n v} $opts {
			if {[string index $n 0] eq "+"} {
				switch -exact -- [string range $n 1 end] {
				mo - menuopts - menuoptions {dict set ol menuoptions $v}
				save { dict set ol save $v }
				type { set usertype $v }
				}
				continue
			}
			switch -exact -- $n {
			-value {set type radiobutton}
			-onvalue - -offvalue {set type checkbutton}
			-menu {set type cascade}
			-variable {
				set type checkbutton
				incr vars($v)
			}
			}
			dict set options $n $v
		}
		if {$usertype ne ""} {
			set type $usertype
			dict set ol usertype $usertype
		}
		dict set ol type $type
		dict set ol options $options
		lappend items $ol
	}

	# Auto Radiobuttons
	foreach z $items[set items {}] {
		if {![dict exists $z usertype] && [dict get $z type] eq "checkbutton" \
				&& [dict exists $z options -variable] \
				&& $vars([dict get $z options -variable]) > 1} {
			dict set z type radiobutton
		}
		lappend items $z
	}

	# Levels
	set s {}
	foreach z $items {
		lappend s [dict get $z spaces]
	}
	set i -1
	set l [lsort -integer -unique $s]
	foreach n $l {
		set levs($n) [incr i]
	}
	foreach z $items[set items {}] {
		dict set z level $levs([dict get $z spaces])
		lappend items $z
	}

	# Cascades
	set ritems {}
	set l 0
	foreach z [lreverse $items] {
		if {[dict get $z level] < $l} {
			dict set z type cascade
		}
		set l [dict get $z level]
		lappend ritems $z
	}
	set items [lreverse $ritems]

	set olev 0

	# Paths, numbers, names, indexes, autocommands
	foreach z $items[set items {}] {
		set lev [dict get $z level]
		if {[dict get $z type] eq "cascade"} {
			if {![info exists cnt($lev)]} {
				set cnt($lev) 0
			} else {
				incr cnt($lev)
			}
			foreach e [array names cnt] {
				if {$e > $lev} {
					unset cnt($e)
				}
			}
		}

		while {$olev > $lev} {
			unset idx($olev)
			incr olev -1
		}
		set olev $lev

		if {![info exists idx($lev)]} {
			set idx($lev) 0
		} else {
			incr idx($lev)
		}
		set index $idx($lev)
		set L([dict get $z level]) [string totitle [regsub -all {[[:space:]]} [dgwd {} $z options -label] {}]]
		set inNumber $menuname
		set inName   $menuname
		set ac ""
		for {set i 0} {$i < [dict get $z level]} {incr i} {
			append inNumber .m$cnt($i)
			append inName .m$L($i)
			if {$i == 0} {
				append ac [string tolower $L($i)]
			} else {
				append ac $L($i)
			}
		}
		if {[info exists L($i)]} {
			if {$i == 0} {
				append ac [string tolower $L($i)]
			} else {
				append ac $L($i)
			}
		}
		dict set z inNumber $inNumber
		dict set z inName   $inName
		dict set z autocmd  $ac
		dict set z index    $index

		# Also determine whether or not to create a menu for a cascade
		if {[dict get $z type] eq "cascade"} {
			dict set z addmnu [expr {![dict exists $z options -menu]}]
		}

		lappend items $z
	}

	# Attach cascades
	set ritems {}
	set l 0
	foreach z [lreverse $items] {
		if {[dict get $z type] eq "cascade"} {
			dict set z menuNumber $menuNumber
			dict set z menuName $menuName
		}
		set menuNumber [dict get $z inNumber]
		set menuName [dict get $z inName]
		lappend ritems $z
	}
	set items [lreverse $ritems]

	# Menu, opts
	set menu0 "$menucmd $menuname"
	set addtro 1
	foreach {n v} $menuopts[set menuopts {}] {
		if {$n eq "-tearoff"} {
			set addtro 0
		}
		lappend menuopts $n $v
	}
	if {$addtro} {
		lappend menuopts -tearoff 0
	}
	append menu0 " " $menuopts

	set ruby_menucmd Tk[string totitle $menucmd].new
	set ruby_menu0 [d2u $menuname]=$ruby_menucmd\([to2ro $menuopts]\)

	set lua_menucmd ltk.$menucmd
	set lua_menu0 [d2u $menuname]=$lua_menucmd\{[to2lo $menuopts]\}

	set perl_menucmd new_$menucmd
	set perl_menu0 my\ \$[d2u $menuname]=\$root->$perl_menucmd\([to2po $menuopts])

	set python_menucmd [string totitle $menucmd]
	set python_menu0 [d2u $menuname]=$python_menucmd\(root,[to2yo $menuopts]\)

	# Assemble
	set        o {}
	set   ruby_o {}
	set    lua_o {}
	set   perl_o {}
	set python_o {}
	set int [lindex {inNumber inName} $names]
	set mnt [lindex {menuNumber menuName} $names]
	if {$menu} {
		if {[string length $menu0] > 0} {
			lappend o $menu0
		}
		if {[string length $ruby_menu0] > 0} {
			lappend ruby_o $ruby_menu0
		}
		if {[string length $lua_menu0] > 0} {
			lappend lua_o $lua_menu0
			lappend lua_o [d2u $menuname]_wc=ltk.wcmd([d2u $menuname])
		}
		if {[string length $perl_menu0] > 0} {
			lappend perl_o $perl_menu0
		}
		if {[string length $python_menu0] > 0} {
			lappend python_o $python_menu0
		}
	}
	foreach z $items {
		set type [dict get $z type]
		set zint [dict get $z $int]
		set uzint [d2u $zint]
		set        l   "$zint add $type"
		set    lua_l   "$uzint\_wc\{'add','$type'"
		set   ruby_l   "$uzint.add_$type\("
		set   perl_l "\$$uzint->add_$type\("
		set python_l   "$uzint.add_$type\("
		set addcmd [expr {$autocmd && $type eq "command"}]
		set addtro 1
		set   ruby_options {}
		set    lua_options {}
		set   perl_options {}
		set python_options {}
		dict for {n v} [dict get $z options] {
			if {[regexp {[[:space:]]} $v]} {
				if {$n eq "-command"} {
					set v "\{$v\}"
				}
				if {$n eq "-label"} {
					set v "\{$v\}"
				}
			}
			if {$n eq "-command"} {
				set addcmd 0
				if {[string range $v 0 1] ne "::" && [string index $v 0] ne "\{"} {
					set v "$ns$v"
				}
			}
			if {$n eq "-variable"} {
				if {[string range $v 0 1] ne "::"} {
					set v "$ns$v"
				}
			}

			append l " " $n " " $v

			set cv $v
			if {[regexp {[[:space:]]} $cv]} {
				if {[string index $cv 0] eq "\{" && [string index $cv end] eq "\}"} {
					set cv [string range $cv 1 end-1]
				}
			}
			if {[string index $cv 0] ne "\"" && [string index $cv end] ne "\""} {
				if {$n in [list "-command"]} {
					set cv "\"$cv\""
				}
			}

			if {$n ni [list "-variable" "-command"]} {
				lappend lua_options $n $cv
			}

			if {$n ni [list "-variasble"]} {
				if {$n in [list "-command"]} {
					lappend python_options $n [tns2yns $cv]
				} else {
					#lappend python_options $n $cv
					lappend python_options $n [tns2yns $cv]
				}
			}

			lappend ruby_options $n $cv
			lappend perl_options $n $cv

		}
		if {$addcmd} {
			set ac [dict get $z autocmd]
			append l " " -command " " $ns$ac
			lappend    lua_options -command [tns2yns $ns$ac]
			lappend python_options -command "\"[tns2yns $ns$ac]\""
		}
		if {[dict get $z type] eq "cascade"} {
			set menuopts [dgwd {} $z menuoptions]
			set nmenuopts {}
			set addtro 1
			foreach {n v} $menuopts {
				if {$n eq "-tearoff"} {
					set addtro 0
				}
				if {$n eq "-postcommand"} {
					set addcmd 0
					if {[string range $v 0 1] ne "::" && [string index $v 0] ne "\{"} {
						set v $ns$v
					}
				}
				lappend nmenuopts $n $v
			}
			set menuopts $nmenuopts
			if {$addtro} {
				lappend menuopts -tearoff 0
			}
			if {[dict get $z addmnu]} {
				append l " " -menu " "
				append l "\["
				append l $menucmd " " [dict get $z $mnt]
				append l " " $menuopts
				append l "\]"
				set umnt [d2u [dict get $z $mnt]]
				lappend   ruby_options -menu   $umnt
				lappend    lua_options -menu   $umnt
				lappend   perl_options -menu \$$umnt
				lappend python_options -menu   $umnt
			}
		}
		lappend o $l
		append ruby_l [to2ro $ruby_options]
		append ruby_l "\)"
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend ruby_o [d2u [dict get $z $mnt]]=$ruby_menucmd\([d2u [dict get $z $int]],[to2ro $menuopts]\)
		}
		lappend ruby_o $ruby_l
		if {[llength $lua_options] > 0} {append lua_l "," }
		append lua_l [to2lo $lua_options]
		append lua_l "\}"
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend lua_o [d2u [dict get $z $mnt]]=$lua_menucmd\{[to2lo $menuopts]\}
		}
		lappend lua_o $lua_l
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend lua_o [d2u [dict get $z $mnt]]_wc=ltk.wcmd([d2u [dict get $z $mnt]])
		}
		append perl_l [to2po $perl_options]
		append perl_l "\)"
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend perl_o my\ \$[d2u [dict get $z $mnt]]=\$[d2u [dict get $z $int]]->$perl_menucmd\([to2po $menuopts]\)
		}
		lappend perl_o $perl_l
		append python_l [to2yo $python_options]
		append python_l "\)"
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend python_o [d2u [dict get $z $mnt]]=$python_menucmd\([d2u [dict get $z $int]],[to2yo $menuopts]\)
		}
		lappend python_o $python_l
	}

	# Saves
	foreach z $items {
		if {![dict exists $z save]} { continue }
		set save [dict get $z save]
		if {$save eq ""} {
			set save [dict get $z autocmd]
		}
		if {$vartype eq "array"} {
			set s "set ${varname}($save)"
		} else {
			set s "dict set $varname $save"
		}
		append s " "
		append s "\["
		append s "dict create in [dict get $z $int] index [dict get $z index]"
		dict set saves [dict get $z $int] [dict create index [dict get $z index]]
		if {[dict get $z type] eq "cascade"} {
			append s " " cascade " " [dict get $z $mnt]
			dict set saves [dict get $z $int] cascade [dict get $z $mnt]
		}
		append s "\]"
		lappend o $s
	}

	set d [dict create items $items]
	dict set d tcl    [dict create src [join $o        \n]]
	dict set d ruby   [dict create src [join $ruby_o   \n]]
	dict set d lua    [dict create src [join $lua_o    \n]]
	dict set d perl   [dict create src [join $perl_o \;\n]\;]
	dict set d python [dict create src [join $python_o \n]]
	if {[info exists saves]} {
		dict set d saves $saves
	}
	return $d
}

#---------------------------------------------------------------------
# Misc
#---------------------------------------------------------------------

#
# Dict Get With Default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

#
# Underline Ampersand
# Convert ampersand in text to -underline option
#
proc mebUA {text} {
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}

#---------------------------------------------------------------------
# Lang
#---------------------------------------------------------------------

#
# Dot To Underscore
#
proc d2u {s} {
	return [string map [list . _] $s]
}

#
# Tcl Option To Ruby Option
#
proc to2ro {options} {
	set z {}
	foreach {n v} $options {
		switch -- $n {
		-command - -tearoff - -underline - -menu { set l ""; set r "" }
		default { set l "'"; set r "'" }
		}
		lappend z ":[string range $n 1 end]=>$l$v$r"
	}
	return [join $z ,]
}

#
# Tcl Option To Lua Option
#
proc to2lo {options} {
	set z {}
	foreach {n v} $options {
		switch -- $n {
		-command - -tearoff - -underline - -menu { set l ""; set r "" }
		default { set l "'"; set r "'" }
		}
		lappend z [string range $n 1 end]=$l$v$r
	}
	return [join $z ,]
}

#
# Tcl Option To Perl Option
#
proc to2po {options} {
	set z {}
	foreach {n v} $options {
		switch -- $n {
		-command - -tearoff - -underline - -menu { set l ""; set r "" }
		default { set l "\""; set r "\"" }
		}
		lappend z $n=>$l$v$r
	}
	return [join $z ,]
}

#
# Tcl Option To Python Option
#
proc to2yo {options} {
	set z {}
	foreach {n v} $options {
		switch -- $n {
		-command - -tearoff - -underline - -menu { set l ""; set r "" }
		default { set l "\""; set r "\"" }
		}
		lappend z [string range $n 1 end]=$l$v$r
	}
	return [join $z ,]
}

#
# Tcl Namespace To Python Namespace
#
proc tns2yns {ns} {
	set q 0
	if {[string index $ns 0] eq "\"" && [string index $ns end] eq "\""} {
		set ns [string range $ns 1 end-1]
		set q 1
	}
	set ns [string map {:: _} [string trim $ns :]]
	if {$q} {
		set ns "\"$ns\""
	}
	return $ns
}

#---------------------------------------------------------------------
# Setup
#---------------------------------------------------------------------

proc setup {} {
	if {![require] || ![export] || ![provide]} { error "Something went wrong." }
}

namespace current}]::setup

# EOF

# THATCH /pkgs/tklib/autoscroll-1.1.tm size 7111 mtime 1659183312 modbase /pkgs/tklib
# autoscroll.tcl --
#
#       Package to create scroll bars that automatically appear when
#       a window is too small to display its content.
#
# Copyright (c) 2003    Kevin B Kenny <kennykb@users.sourceforge.net>
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: autoscroll.tcl,v 1.8 2005/06/01 02:37:51 andreas_kupries Exp $

package require Tk
package provide autoscroll 1.1

namespace eval ::autoscroll {
    namespace export autoscroll unautoscroll
    bind Autoscroll <Destroy> [namespace code [list destroyed %W]]
    bind Autoscroll <Map> [namespace code [list map %W]]
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::autoscroll --
 #
 #       Create a scroll bar that disappears when it is not needed, and
 #       reappears when it is.
 #
 # Parameters:
 #       w    -- Path name of the scroll bar, which should already exist
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       The widget command is renamed, so that the 'set' command can
 #       be intercepted and determine whether the widget should appear.
 #       In addition, the 'Autoscroll' bind tag is added to the widget,
 #       so that the <Destroy> event can be intercepted.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::autoscroll { w } {
    if { [info commands ::autoscroll::renamed$w] != "" } { return $w }
    rename $w ::autoscroll::renamed$w
    interp alias {} ::$w {} ::autoscroll::widgetCommand $w
    bindtags $w [linsert [bindtags $w] 1 Autoscroll]
    eval [list ::$w set] [renamed$w get]
    return $w
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::unautoscroll --
 #
 #       Return a scrollbar to its normal static behavior by removing
 #       it from the control of this package.
 #
 # Parameters:
 #       w    -- Path name of the scroll bar, which must have previously
 #               had ::autoscroll::autoscroll called on it.
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       The widget command is renamed to its original name. The widget
 #       is mapped if it was not currently displayed. The widgets
 #       bindtags are returned to their original state. Internal memory
 #       is cleaned up.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::unautoscroll { w } {
    if { [info commands ::autoscroll::renamed$w] != "" } {
        variable grid
        rename ::$w {}
        rename ::autoscroll::renamed$w ::$w
        if { [set i [lsearch -exact [bindtags $w] Autoscroll]] > -1 } {
            bindtags $w [lreplace [bindtags $w] $i $i]
        }
        if { [info exists grid($w)] } {
            eval [join $grid($w) \;]
            unset grid($w)
        }
    }
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::widgetCommand --
 #
 #       Widget command on an 'autoscroll' scrollbar
 #
 # Parameters:
 #       w       -- Path name of the scroll bar
 #       command -- Widget command being executed
 #       args    -- Arguments to the commane
 #
 # Results:
 #       Returns whatever the widget command returns
 #
 # Side effects:
 #       Has whatever side effects the widget command has.  In
 #       addition, the 'set' widget command is handled specially,
 #       by gridding/packing the scroll bar according to whether
 #       it is required.
 #
 #------------------------------------------------------------

proc ::autoscroll::widgetCommand { w command args } {
    variable grid
    if { $command == "set" } {
        foreach { min max } $args {}
        if { $min <= 0 && $max >= 1 } {
            switch -exact -- [winfo manager $w] {
                grid {
                    lappend grid($w) "[list grid $w] [grid info $w]"
                    grid forget $w
                }
                pack {
                    foreach x [pack slaves [winfo parent $w]] {
                        lappend grid($w) "[list pack $x] [pack info $x]"
                    }
                    pack forget $w
                }
            }
        } elseif { [info exists grid($w)] } {
            eval [join $grid($w) \;]
            unset grid($w)
        }
    }
    return [eval [list renamed$w $command] $args]
}


 #----------------------------------------------------------------------
 #
 # ::autoscroll::destroyed --
 #
 #       Callback executed when an automatic scroll bar is destroyed.
 #
 # Parameters:
 #       w -- Path name of the scroll bar
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       Cleans up internal memory.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::destroyed { w } {
    variable grid
    catch { unset grid($w) }
    rename ::$w {}
}


 #----------------------------------------------------------------------
 #
 # ::autoscroll::map --
 #
 #       Callback executed when an automatic scroll bar is mapped.
 #
 # Parameters:
 #       w -- Path name of the scroll bar.
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       Geometry of the scroll bar's top-level window is constrained.
 #
 # This procedure keeps the top-level window associated with an
 # automatic scroll bar from being resized automatically after the
 # scroll bar is mapped.  This effect avoids a potential endless loop
 # in the case where the resize of the top-level window resizes the
 # widget being scrolled, causing the scroll bar no longer to be needed.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::map { w } {
    wm geometry [winfo toplevel $w] [wm geometry [winfo toplevel $w]]
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::wrap --
 #
 #       Arrange for all new scrollbars to be automatically autoscrolled
 #
 # Parameters:
 #       None.
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       ::scrollbar is overloaded to automatically autoscroll any new
 #          scrollbars.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::wrap {} {
    if {[info commands ::autoscroll::_scrollbar] != ""} {return}
    rename ::scrollbar ::autoscroll::_scrollbar
    proc ::scrollbar {w args} {
        eval ::autoscroll::_scrollbar [list $w] $args
        ::autoscroll::autoscroll $w
        return $w
    }
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::unwrap --
 #
 #       Turns off automatic autoscrolling of new scrollbars. Does not
 #         effect existing scrollbars.
 #
 # Parameters:
 #       None.
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       ::scrollbar is returned to its original state
 #
 #----------------------------------------------------------------------

proc ::autoscroll::unwrap {} {
    if {[info commands ::autoscroll::_scrollbar] == ""} {return}
    rename ::scrollbar {}
    rename ::autoscroll::_scrollbar ::scrollbar
}

# THATCH /pkgs/tklib/license.terms size 2156 mtime 1659183312
This software is copyrighted by Ajuba Solutions and other parties.
The following terms apply to all files associated with the software unless
explicitly disclaimed in individual files.

The authors hereby grant permission to use, copy, modify, distribute,
and license this software and its documentation for any purpose, provided
that existing copyright notices are retained in all copies and that this
notice is included verbatim in any distributions. No written agreement,
license, or royalty fee is required for any of the authorized uses.
Modifications to this software may be copyrighted by their authors
and need not follow the licensing terms described here, provided that
the new terms are clearly indicated on the first page of each file where
they apply.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

GOVERNMENT USE: If you are acquiring this software on behalf of the
U.S. government, the Government shall have only "Restricted Rights"
in the software and related documentation as defined in the Federal 
Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
are acquiring the software on behalf of the Department of Defense, the
software shall be classified as "Commercial Computer Software" and the
Government shall have only "Restricted Rights" as defined in Clause
252.227-7013 (c) (1) of DFARs.  Notwithstanding the foregoing, the
authors grant the U.S. Government and others acting in its behalf
permission to use and distribute the software in accordance with the
terms specified in this license. 

# THATCH /pkgs/tcllib/md5/pkgIndex.tcl size 195 mtime 1659183356
if {![package vsatisfies [package provide Tcl] 8.2]} {return}
package ifneeded md5 2.0.8 [list source [file join $dir md5x.tcl]]
package ifneeded md5 1.4.5 [list source [file join $dir md5.tcl]]

# THATCH /pkgs/tcllib/md5/md5x.tcl size 25283 mtime 1659183356
# md5.tcl - Copyright (C) 2003 Pat Thoyts <patthoyts@users.sourceforge.net>
#
# MD5  defined by RFC 1321, "The MD5 Message-Digest Algorithm"
# HMAC defined by RFC 2104, "Keyed-Hashing for Message Authentication"
#
# This is an implementation of MD5 based upon the example code given in
# RFC 1321 and upon the tcllib MD4 implementation and taking some ideas
# from the earlier tcllib md5 version by Don Libes.
#
# This implementation permits incremental updating of the hash and 
# provides support for external compiled implementations either using
# critcl (md5c) or Trf.
#
# -------------------------------------------------------------------------
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# -------------------------------------------------------------------------

package require Tcl 8.2;                # tcl minimum version

namespace eval ::md5 {
    variable  accel
    array set accel {critcl 0 cryptkit 0 trf 0}

    namespace export md5 hmac MD5Init MD5Update MD5Final

    variable uid
    if {![info exists uid]} {
        set uid 0
    }
}

# -------------------------------------------------------------------------

# MD5Init --
#
#   Create and initialize an MD5 state variable. This will be
#   cleaned up when we call MD5Final
#
proc ::md5::MD5Init {} {
    variable accel
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token state

    # RFC1321:3.3 - Initialize MD5 state structure
    array set state \
        [list \
             A [expr {0x67452301}] \
             B [expr {0xefcdab89}] \
             C [expr {0x98badcfe}] \
             D [expr {0x10325476}] \
             n 0 i "" ]
    if {$accel(cryptkit)} {
        cryptkit::cryptCreateContext state(ckctx) CRYPT_UNUSED CRYPT_ALGO_MD5
    } elseif {$accel(trf)} {
        set s {}
        switch -exact -- $::tcl_platform(platform) {
            windows { set s [open NUL w] }
            unix    { set s [open /dev/null w] }
        }
        if {$s != {}} {
            fconfigure $s -translation binary -buffering none
            ::md5 -attach $s -mode write \
                -read-type variable \
                -read-destination [subst $token](trfread) \
                -write-type variable \
                -write-destination [subst $token](trfwrite)
            array set state [list trfread 0 trfwrite 0 trf $s]
        }
    }
    return $token
}

# MD5Update --
#
#   This is called to add more data into the hash. You may call this
#   as many times as you require. Note that passing in "ABC" is equivalent
#   to passing these letters in as separate calls -- hence this proc 
#   permits hashing of chunked data
#
#   If we have a C-based implementation available, then we will use
#   it here in preference to the pure-Tcl implementation.
#
proc ::md5::MD5Update {token data} {
    variable accel
    upvar #0 $token state

    if {$accel(critcl)} {
        if {[info exists state(md5c)]} {
            set state(md5c) [md5c $data $state(md5c)]
        } else {
            set state(md5c) [md5c $data]
        }
        return
    } elseif {[info exists state(ckctx)]} {
        if {[string length $data] > 0} {
            cryptkit::cryptEncrypt $state(ckctx) $data
        }
        return
    } elseif {[info exists state(trf)]} {
        puts -nonewline $state(trf) $data
        return
    }

    # Update the state values
    incr state(n) [string length $data]
    append state(i) $data

    # Calculate the hash for any complete blocks
    set len [string length $state(i)]
    for {set n 0} {($n + 64) <= $len} {} {
        MD5Hash $token [string range $state(i) $n [incr n 64]]
    }

    # Adjust the state for the blocks completed.
    set state(i) [string range $state(i) $n end]
    return
}

# MD5Final --
#
#    This procedure is used to close the current hash and returns the
#    hash data. Once this procedure has been called the hash context
#    is freed and cannot be used again.
#
#    Note that the output is 128 bits represented as binary data.
#
proc ::md5::MD5Final {token} {
    upvar #0 $token state

    # Check for either of the C-compiled versions.
    if {[info exists state(md5c)]} {
        set r $state(md5c)
        unset state
        return $r
    } elseif {[info exists state(ckctx)]} {
        cryptkit::cryptEncrypt $state(ckctx) ""
        cryptkit::cryptGetAttributeString $state(ckctx) \
            CRYPT_CTXINFO_HASHVALUE r 16
        cryptkit::cryptDestroyContext $state(ckctx)
        # If nothing was hashed, we get no r variable set!
        if {[info exists r]} {
            unset state
            return $r
        }
    } elseif {[info exists state(trf)]} {
        close $state(trf)
        set r $state(trfwrite)
        unset state
        return $r
    }

    # RFC1321:3.1 - Padding
    #
    set len [string length $state(i)]
    set pad [expr {56 - ($len % 64)}]
    if {$len % 64 > 56} {
        incr pad 64
    }
    if {$pad == 0} {
        incr pad 64
    }

    #puts "P $pad|bits=[expr {8 * $state(n)}]"

    append state(i) [binary format a$pad \x80]

    # RFC1321:3.2 - Append length in bits as little-endian wide int.
    append state(i) [binary format ii [expr {8 * $state(n)}] 0]

    #puts DATA=[Hex $state(i)]([string length $state(i)])

    # Calculate the hash for the remaining block.
    set len [string length $state(i)]
    for {set n 0} {($n + 64) <= $len} {} {
        MD5Hash $token [string range $state(i) $n [incr n 64]]
    }

    #puts md5-post__________________________________________
    #parray ::${token}

    # RFC1321:3.5 - Output
    set r [bytes $state(A)][bytes $state(B)][bytes $state(C)][bytes $state(D)]
    unset state

    #puts HASH=[Hex $r]
    return $r
}

# -------------------------------------------------------------------------
# HMAC Hashed Message Authentication (RFC 2104)
#
# hmac = H(K xor opad, H(K xor ipad, text))
#

# HMACInit --
#
#    This is equivalent to the MD5Init procedure except that a key is
#    added into the algorithm
#
proc ::md5::HMACInit {K} {

    # Key K is adjusted to be 64 bytes long. If K is larger, then use
    # the MD5 digest of K and pad this instead.
    set len [string length $K]
    if {$len > 64} {
        set tok [MD5Init]
        MD5Update $tok $K
        set K [MD5Final $tok]
        set len [string length $K]
    }
    set pad [expr {64 - $len}]
    append K [string repeat \0 $pad]

    # Cacluate the padding buffers.
    set Ki {}
    set Ko {}
    binary scan $K i16 Ks
    foreach k $Ks {
        append Ki [binary format i [expr {$k ^ 0x36363636}]]
        append Ko [binary format i [expr {$k ^ 0x5c5c5c5c}]]
    }

    set tok [MD5Init]
    MD5Update $tok $Ki;                 # initialize with the inner pad
    
    # preserve the Ko value for the final stage.
    # FRINK: nocheck
    set [subst $tok](Ko) $Ko

    return $tok
}

# HMACUpdate --
#
#    Identical to calling MD5Update
#
proc ::md5::HMACUpdate {token data} {
    MD5Update $token $data
    return
}

# HMACFinal --
#
#    This is equivalent to the MD5Final procedure. The hash context is
#    closed and the binary representation of the hash result is returned.
#
proc ::md5::HMACFinal {token} {
    upvar #0 $token state

    set tok [MD5Init];                  # init the outer hashing function
    MD5Update $tok $state(Ko);          # prepare with the outer pad.
    MD5Update $tok [MD5Final $token];   # hash the inner result
    return [MD5Final $tok]
}

# -------------------------------------------------------------------------
# Description:
#  This is the core MD5 algorithm. It is a lot like the MD4 algorithm but
#  includes an extra round and a set of constant modifiers throughout.
# 
# Note:
#  This function body is substituted later on to inline some of the 
#  procedures and to make is a bit more comprehensible.
#
set ::md5::MD5Hash_body {
    variable $token
    upvar 0 $token state

    #puts TR__=[Hex $msg]([string length $msg])

    # RFC1321:3.4 - Process Message in 16-Word Blocks
    binary scan $msg i* blocks
    foreach {X0 X1 X2 X3 X4 X5 X6 X7 X8 X9 X10 X11 X12 X13 X14 X15} $blocks {
        #puts BL

        set A $state(A)
        set B $state(B)
        set C $state(C)
        set D $state(D)

        # Round 1
        # Let [abcd k s i] denote the operation
        #   a = b + ((a + F(b,c,d) + X[k] + T[i]) <<< s).
        # Do the following 16 operations.
        # [ABCD  0  7  1]  [DABC  1 12  2]  [CDAB  2 17  3]  [BCDA  3 22  4]
        set A [expr {$B + (($A + [F $B $C $D] + $X0 + $T01) <<< 7)}]
        set D [expr {$A + (($D + [F $A $B $C] + $X1 + $T02) <<< 12)}]
        set C [expr {$D + (($C + [F $D $A $B] + $X2 + $T03) <<< 17)}]
        set B [expr {$C + (($B + [F $C $D $A] + $X3 + $T04) <<< 22)}]
        # [ABCD  4  7  5]  [DABC  5 12  6]  [CDAB  6 17  7]  [BCDA  7 22  8]
        set A [expr {$B + (($A + [F $B $C $D] + $X4 + $T05) <<< 7)}]
        set D [expr {$A + (($D + [F $A $B $C] + $X5 + $T06) <<< 12)}]
        set C [expr {$D + (($C + [F $D $A $B] + $X6 + $T07) <<< 17)}]
        set B [expr {$C + (($B + [F $C $D $A] + $X7 + $T08) <<< 22)}]
        # [ABCD  8  7  9]  [DABC  9 12 10]  [CDAB 10 17 11]  [BCDA 11 22 12]
        set A [expr {$B + (($A + [F $B $C $D] + $X8 + $T09) <<< 7)}]
        set D [expr {$A + (($D + [F $A $B $C] + $X9 + $T10) <<< 12)}]
        set C [expr {$D + (($C + [F $D $A $B] + $X10 + $T11) <<< 17)}]
        set B [expr {$C + (($B + [F $C $D $A] + $X11 + $T12) <<< 22)}]
        # [ABCD 12  7 13]  [DABC 13 12 14]  [CDAB 14 17 15]  [BCDA 15 22 16]
        set A [expr {$B + (($A + [F $B $C $D] + $X12 + $T13) <<< 7)}]
        set D [expr {$A + (($D + [F $A $B $C] + $X13 + $T14) <<< 12)}]
        set C [expr {$D + (($C + [F $D $A $B] + $X14 + $T15) <<< 17)}]
        set B [expr {$C + (($B + [F $C $D $A] + $X15 + $T16) <<< 22)}]

        # Round 2.
        # Let [abcd k s i] denote the operation
        #   a = b + ((a + G(b,c,d) + X[k] + Ti) <<< s)
        # Do the following 16 operations.
        # [ABCD  1  5 17]  [DABC  6  9 18]  [CDAB 11 14 19]  [BCDA  0 20 20]
        set A [expr {$B + (($A + [G $B $C $D] + $X1  + $T17) <<<  5)}]
        set D [expr {$A + (($D + [G $A $B $C] + $X6  + $T18) <<<  9)}]
        set C [expr {$D + (($C + [G $D $A $B] + $X11 + $T19) <<< 14)}]
        set B [expr {$C + (($B + [G $C $D $A] + $X0  + $T20) <<< 20)}]
        # [ABCD  5  5 21]  [DABC 10  9 22]  [CDAB 15 14 23]  [BCDA  4 20 24]
        set A [expr {$B + (($A + [G $B $C $D] + $X5  + $T21) <<<  5)}]
        set D [expr {$A + (($D + [G $A $B $C] + $X10 + $T22) <<<  9)}]
        set C [expr {$D + (($C + [G $D $A $B] + $X15 + $T23) <<< 14)}]
        set B [expr {$C + (($B + [G $C $D $A] + $X4  + $T24) <<< 20)}]
        # [ABCD  9  5 25]  [DABC 14  9 26]  [CDAB  3 14 27]  [BCDA  8 20 28]
        set A [expr {$B + (($A + [G $B $C $D] + $X9  + $T25) <<<  5)}]
        set D [expr {$A + (($D + [G $A $B $C] + $X14 + $T26) <<<  9)}]
        set C [expr {$D + (($C + [G $D $A $B] + $X3  + $T27) <<< 14)}]
        set B [expr {$C + (($B + [G $C $D $A] + $X8  + $T28) <<< 20)}]
        # [ABCD 13  5 29]  [DABC  2  9 30]  [CDAB  7 14 31]  [BCDA 12 20 32]
        set A [expr {$B + (($A + [G $B $C $D] + $X13 + $T29) <<<  5)}]
        set D [expr {$A + (($D + [G $A $B $C] + $X2  + $T30) <<<  9)}]
        set C [expr {$D + (($C + [G $D $A $B] + $X7  + $T31) <<< 14)}]
        set B [expr {$C + (($B + [G $C $D $A] + $X12 + $T32) <<< 20)}]
        
        # Round 3.
        # Let [abcd k s i] denote the operation
        #   a = b + ((a + H(b,c,d) + X[k] + T[i]) <<< s)
        # Do the following 16 operations.
        # [ABCD  5  4 33]  [DABC  8 11 34]  [CDAB 11 16 35]  [BCDA 14 23 36]
        set A [expr {$B + (($A + [H $B $C $D] + $X5  + $T33) <<<  4)}]
        set D [expr {$A + (($D + [H $A $B $C] + $X8  + $T34) <<< 11)}]
        set C [expr {$D + (($C + [H $D $A $B] + $X11 + $T35) <<< 16)}]
        set B [expr {$C + (($B + [H $C $D $A] + $X14 + $T36) <<< 23)}]
        # [ABCD  1  4 37]  [DABC  4 11 38]  [CDAB  7 16 39]  [BCDA 10 23 40]
        set A [expr {$B + (($A + [H $B $C $D] + $X1  + $T37) <<<  4)}]
        set D [expr {$A + (($D + [H $A $B $C] + $X4  + $T38) <<< 11)}]
        set C [expr {$D + (($C + [H $D $A $B] + $X7  + $T39) <<< 16)}]
        set B [expr {$C + (($B + [H $C $D $A] + $X10 + $T40) <<< 23)}]
        # [ABCD 13  4 41]  [DABC  0 11 42]  [CDAB  3 16 43]  [BCDA  6 23 44]
        set A [expr {$B + (($A + [H $B $C $D] + $X13 + $T41) <<<  4)}]
        set D [expr {$A + (($D + [H $A $B $C] + $X0  + $T42) <<< 11)}]
        set C [expr {$D + (($C + [H $D $A $B] + $X3  + $T43) <<< 16)}]
        set B [expr {$C + (($B + [H $C $D $A] + $X6  + $T44) <<< 23)}]
        # [ABCD  9  4 45]  [DABC 12 11 46]  [CDAB 15 16 47]  [BCDA  2 23 48]
        set A [expr {$B + (($A + [H $B $C $D] + $X9  + $T45) <<<  4)}]
        set D [expr {$A + (($D + [H $A $B $C] + $X12 + $T46) <<< 11)}]
        set C [expr {$D + (($C + [H $D $A $B] + $X15 + $T47) <<< 16)}]
        set B [expr {$C + (($B + [H $C $D $A] + $X2  + $T48) <<< 23)}]

        # Round 4.
        # Let [abcd k s i] denote the operation
        #   a = b + ((a + I(b,c,d) + X[k] + T[i]) <<< s)
        # Do the following 16 operations.
        # [ABCD  0  6 49]  [DABC  7 10 50]  [CDAB 14 15 51]  [BCDA  5 21 52]
        set A [expr {$B + (($A + [I $B $C $D] + $X0  + $T49) <<<  6)}]
        set D [expr {$A + (($D + [I $A $B $C] + $X7  + $T50) <<< 10)}]
        set C [expr {$D + (($C + [I $D $A $B] + $X14 + $T51) <<< 15)}]
        set B [expr {$C + (($B + [I $C $D $A] + $X5  + $T52) <<< 21)}]
        # [ABCD 12  6 53]  [DABC  3 10 54]  [CDAB 10 15 55]  [BCDA  1 21 56]
        set A [expr {$B + (($A + [I $B $C $D] + $X12 + $T53) <<<  6)}]
        set D [expr {$A + (($D + [I $A $B $C] + $X3  + $T54) <<< 10)}]
        set C [expr {$D + (($C + [I $D $A $B] + $X10 + $T55) <<< 15)}]
        set B [expr {$C + (($B + [I $C $D $A] + $X1  + $T56) <<< 21)}]
        # [ABCD  8  6 57]  [DABC 15 10 58]  [CDAB  6 15 59]  [BCDA 13 21 60]
        set A [expr {$B + (($A + [I $B $C $D] + $X8  + $T57) <<<  6)}]
        set D [expr {$A + (($D + [I $A $B $C] + $X15 + $T58) <<< 10)}]
        set C [expr {$D + (($C + [I $D $A $B] + $X6  + $T59) <<< 15)}]
        set B [expr {$C + (($B + [I $C $D $A] + $X13 + $T60) <<< 21)}]
        # [ABCD  4  6 61]  [DABC 11 10 62]  [CDAB  2 15 63]  [BCDA  9 21 64]
        set A [expr {$B + (($A + [I $B $C $D] + $X4  + $T61) <<<  6)}]
        set D [expr {$A + (($D + [I $A $B $C] + $X11 + $T62) <<< 10)}]
        set C [expr {$D + (($C + [I $D $A $B] + $X2  + $T63) <<< 15)}]
        set B [expr {$C + (($B + [I $C $D $A] + $X9  + $T64) <<< 21)}]

        # Then perform the following additions. (That is, increment each
        # of the four registers by the value it had before this block
        # was started.)
        incr state(A) $A
        incr state(B) $B
        incr state(C) $C
        incr state(D) $D
    }

    return
}

proc ::md5::byte {n v} {expr {((0xFF << (8 * $n)) & $v) >> (8 * $n)}}
proc ::md5::bytes {v} { 
    #format %c%c%c%c [byte 0 $v] [byte 1 $v] [byte 2 $v] [byte 3 $v]
    format %c%c%c%c \
        [expr {0xFF & $v}] \
        [expr {(0xFF00 & $v) >> 8}] \
        [expr {(0xFF0000 & $v) >> 16}] \
        [expr {((0xFF000000 & $v) >> 24) & 0xFF}]
}

# 32bit rotate-left
proc ::md5::<<< {v n} {
    return [expr {((($v << $n) \
                        | (($v >> (32 - $n)) \
                               & (0x7FFFFFFF >> (31 - $n))))) \
                      & 0xFFFFFFFF}]
}

# Convert our <<< pseudo-operator into a procedure call.
regsub -all -line \
    {\[expr {(\$[ABCD]) \+ \(\((.*)\)\s+<<<\s+(\d+)\)}\]} \
    $::md5::MD5Hash_body \
    {[expr {int(\1 + [<<< [expr {\2}] \3])}]} \
    ::md5::MD5Hash_body

# RFC1321:3.4 - function F
proc ::md5::F {X Y Z} {
    return [expr {($X & $Y) | ((~$X) & $Z)}]
}

# Inline the F function
regsub -all -line \
    {\[F (\$[ABCD]) (\$[ABCD]) (\$[ABCD])\]} \
    $::md5::MD5Hash_body \
    {( (\1 \& \2) | ((~\1) \& \3) )} \
    ::md5::MD5Hash_body
    
# RFC1321:3.4 - function G
proc ::md5::G {X Y Z} {
    return [expr {(($X & $Z) | ($Y & (~$Z)))}]
}

# Inline the G function
regsub -all -line \
    {\[G (\$[ABCD]) (\$[ABCD]) (\$[ABCD])\]} \
    $::md5::MD5Hash_body \
    {(((\1 \& \3) | (\2 \& (~\3))))} \
    ::md5::MD5Hash_body

# RFC1321:3.4 - function H
proc ::md5::H {X Y Z} {
    return [expr {$X ^ $Y ^ $Z}]
}

# Inline the H function
regsub -all -line \
    {\[H (\$[ABCD]) (\$[ABCD]) (\$[ABCD])\]} \
    $::md5::MD5Hash_body \
    {(\1 ^ \2 ^ \3)} \
    ::md5::MD5Hash_body

# RFC1321:3.4 - function I
proc ::md5::I {X Y Z} {
    return [expr {$Y ^ ($X | (~$Z))}]
}

# Inline the I function
regsub -all -line \
    {\[I (\$[ABCD]) (\$[ABCD]) (\$[ABCD])\]} \
    $::md5::MD5Hash_body \
    {(\2 ^ (\1 | (~\3)))} \
    ::md5::MD5Hash_body


# RFC 1321:3.4 step 4: inline the set of constant modifiers.
namespace eval md5 {
    variable tName
    variable tVal
    variable map
    foreach tName {
        T01 T02 T03 T04 T05 T06 T07 T08 T09 T10 
        T11 T12 T13 T14 T15 T16 T17 T18 T19 T20 
        T21 T22 T23 T24 T25 T26 T27 T28 T29 T30 
        T31 T32 T33 T34 T35 T36 T37 T38 T39 T40 
        T41 T42 T43 T44 T45 T46 T47 T48 T49 T50 
        T51 T52 T53 T54 T55 T56 T57 T58 T59 T60 
        T61 T62 T63 T64 
    }  tVal {
        0xd76aa478 0xe8c7b756 0x242070db 0xc1bdceee
        0xf57c0faf 0x4787c62a 0xa8304613 0xfd469501
        0x698098d8 0x8b44f7af 0xffff5bb1 0x895cd7be
        0x6b901122 0xfd987193 0xa679438e 0x49b40821
        
        0xf61e2562 0xc040b340 0x265e5a51 0xe9b6c7aa
        0xd62f105d 0x2441453  0xd8a1e681 0xe7d3fbc8
        0x21e1cde6 0xc33707d6 0xf4d50d87 0x455a14ed
        0xa9e3e905 0xfcefa3f8 0x676f02d9 0x8d2a4c8a
        
        0xfffa3942 0x8771f681 0x6d9d6122 0xfde5380c
        0xa4beea44 0x4bdecfa9 0xf6bb4b60 0xbebfbc70
        0x289b7ec6 0xeaa127fa 0xd4ef3085 0x4881d05
        0xd9d4d039 0xe6db99e5 0x1fa27cf8 0xc4ac5665
        
        0xf4292244 0x432aff97 0xab9423a7 0xfc93a039
        0x655b59c3 0x8f0ccc92 0xffeff47d 0x85845dd1
        0x6fa87e4f 0xfe2ce6e0 0xa3014314 0x4e0811a1
        0xf7537e82 0xbd3af235 0x2ad7d2bb 0xeb86d391
    } {
        lappend map \$$tName $tVal
    }
    set ::md5::MD5Hash_body [string map $map $::md5::MD5Hash_body]
    unset map tName tVal
}

# Define the MD5 hashing procedure with inline functions.
proc ::md5::MD5Hash {token msg} $::md5::MD5Hash_body
unset ::md5::MD5Hash_body

# -------------------------------------------------------------------------

if {[package provide Trf] != {}} {
    interp alias {} ::md5::Hex {} ::hex -mode encode --
} else {
    proc ::md5::Hex {data} {
        binary scan $data H* result
        return [string toupper $result]
    }
}

# -------------------------------------------------------------------------

# LoadAccelerator --
#
#	This package can make use of a number of compiled extensions to
#	accelerate the digest computation. This procedure manages the
#	use of these extensions within the package. During normal usage
#	this should not be called, but the test package manipulates the
#	list of enabled accelerators.
#
proc ::md5::LoadAccelerator {name} {
    variable accel
    set r 0
    switch -exact -- $name {
        critcl {
            if {![catch {package require tcllibc}]
                || ![catch {package require md5c}]} {
                set r [expr {[info commands ::md5::md5c] != {}}]
            }
        }
        cryptkit {
            if {![catch {package require cryptkit}]} {
                set r [expr {![catch {cryptkit::cryptInit}]}]
            }
        }
        trf {
            if {![catch {package require Trf}]} {
                set r [expr {![catch {::md5 aa} msg]}]
            }
        }
        default {
            return -code error "invalid accelerator package:\
                must be one of [join [array names accel] {, }]"
        }
    }
    set accel($name) $r
}

# -------------------------------------------------------------------------

# Description:
#  Pop the nth element off a list. Used in options processing.
#
proc ::md5::Pop {varname {nth 0}} {
    upvar $varname args
    set r [lindex $args $nth]
    set args [lreplace $args $nth $nth]
    return $r
}

# -------------------------------------------------------------------------

# fileevent handler for chunked file hashing.
#
proc ::md5::Chunk {token channel {chunksize 4096}} {
    upvar #0 $token state
    
    if {[eof $channel]} {
        fileevent $channel readable {}
        set state(reading) 0
    }
        
    MD5Update $token [read $channel $chunksize]
}

# -------------------------------------------------------------------------

proc ::md5::md5 {args} {
    array set opts {-hex 0 -filename {} -channel {} -chunksize 4096}
    while {[string match -* [set option [lindex $args 0]]]} {
        switch -glob -- $option {
            -hex       { set opts(-hex) 1 }
            -file*     { set opts(-filename) [Pop args 1] }
            -channel   { set opts(-channel) [Pop args 1] }
            -chunksize { set opts(-chunksize) [Pop args 1] }
            default {
                if {[llength $args] == 1} { break }
                if {[string compare $option "--"] == 0} { Pop args; break }
                set err [join [lsort [array names opts]] ", "]
                return -code error "bad option $option:\
                    must be one of $err\nlen: [llength $args]"
            }
        }
        Pop args
    }

    if {$opts(-filename) != {}} {
        set opts(-channel) [open $opts(-filename) r]
        fconfigure $opts(-channel) -translation binary
    }

    if {$opts(-channel) == {}} {

        if {[llength $args] != 1} {
            return -code error "wrong # args:\
                should be \"md5 ?-hex? -filename file | string\""
        }
        set tok [MD5Init]

        #puts md5_______________________________________________
        #parray ::${tok}

        #puts IN=(([lindex $args 0]))
        MD5Update $tok [lindex $args 0]

        #puts md5-final_________________________________________
        #parray ::${tok}

        set r [MD5Final $tok]

    } else {

        set tok [MD5Init]
        # FRINK: nocheck
        set [subst $tok](reading) 1
        fileevent $opts(-channel) readable \
            [list [namespace origin Chunk] \
                 $tok $opts(-channel) $opts(-chunksize)]
        vwait [subst $tok](reading)
        set r [MD5Final $tok]

        # If we opened the channel - we should close it too.
        if {$opts(-filename) != {}} {
            close $opts(-channel)
        }
    }
    
    if {$opts(-hex)} {
        set r [Hex $r]
    }
    return $r
}

# -------------------------------------------------------------------------

proc ::md5::hmac {args} {
    array set opts {-hex 0 -filename {} -channel {} -chunksize 4096}
    while {[string match -* [set option [lindex $args 0]]]} {
        switch -glob -- $option {
            -key       { set opts(-key) [Pop args 1] }
            -hex       { set opts(-hex) 1 }
            -file*     { set opts(-filename) [Pop args 1] }
            -channel   { set opts(-channel) [Pop args 1] }
            -chunksize { set opts(-chunksize) [Pop args 1] }
            default {
                if {[llength $args] == 1} { break }
                if {[string compare $option "--"] == 0} { Pop args; break }
                set err [join [lsort [array names opts]] ", "]
                return -code error "bad option $option:\
                    must be one of $err"
            }
        }
        Pop args
    }

    if {![info exists opts(-key)]} {
        return -code error "wrong # args:\
            should be \"hmac ?-hex? -key key -filename file | string\""
    }

    if {$opts(-filename) != {}} {
        set opts(-channel) [open $opts(-filename) r]
        fconfigure $opts(-channel) -translation binary
    }

    if {$opts(-channel) == {}} {

        if {[llength $args] != 1} {
            return -code error "wrong # args:\
                should be \"hmac ?-hex? -key key -filename file | string\""
        }
        set tok [HMACInit $opts(-key)]
        HMACUpdate $tok [lindex $args 0]
        set r [HMACFinal $tok]

    } else {

        set tok [HMACInit $opts(-key)]
        # FRINK: nocheck
        set [subst $tok](reading) 1
        fileevent $opts(-channel) readable \
            [list [namespace origin Chunk] \
                 $tok $opts(-channel) $opts(-chunksize)]
        vwait [subst $tok](reading)
        set r [HMACFinal $tok]

        # If we opened the channel - we should close it too.
        if {$opts(-filename) != {}} {
            close $opts(-channel)
        }
    }
    
    if {$opts(-hex)} {
        set r [Hex $r]
    }
    return $r
}

# -------------------------------------------------------------------------

# Try and load a compiled extension to help.
namespace eval ::md5 {
    variable e
    foreach  e {critcl cryptkit trf} { if {[LoadAccelerator $e]} { break } }
    unset    e
}

package provide md5 2.0.8

# -------------------------------------------------------------------------
# Local Variables:
#   mode: tcl
#   indent-tabs-mode: nil
# End:



# THATCH /pkgs/tcllib/sha1/pkgIndex.tcl size 816 mtime 1659183357
# Tcl package index file, version 1.1
# This file is generated by the "pkg_mkIndex" command
# and sourced either when an application starts up or
# by a "package unknown" script.  It invokes the
# "package ifneeded" command to set up package-related
# information so that packages will be loaded automatically
# in response to "package require" commands.  When this
# script is sourced, the variable $dir must contain the
# full path name of this file's directory.

if {![package vsatisfies [package provide Tcl] 8.2]} {return}
package ifneeded sha256 1.0.4 [list source [file join $dir sha256.tcl]]
package ifneeded sha256c 1.0.4 [list source [file join $dir sha256c.tcl]]
package ifneeded sha1   2.0.4 [list source [file join $dir sha1.tcl]]
package ifneeded sha1   1.1.1 [list source [file join $dir sha1v1.tcl]]

# THATCH /pkgs/tcllib/sha1/sha256.tcl size 24593 mtime 1659183357
# sha256.tcl - Copyright (C) 2005 Pat Thoyts <patthoyts@users.sourceforge.net>
#
# SHA1 defined by FIPS 180-2, "The Secure Hash Standard"
# HMAC defined by RFC 2104, "Keyed-Hashing for Message Authentication"
#
# This is an implementation of the secure hash algorithms specified in the
# FIPS 180-2 document.
#
# This implementation permits incremental updating of the hash and 
# provides support for external compiled implementations using critcl.
#
# This implementation permits incremental updating of the hash and 
# provides support for external compiled implementations either using
# critcl (sha256c).
#
# Ref: http://csrc.nist.gov/publications/fips/fips180-2/fips180-2.pdf
#      http://csrc.nist.gov/publications/fips/fips180-2/fips180-2withchangenotice.pdf
#
# -------------------------------------------------------------------------
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# -------------------------------------------------------------------------
# @mdgen EXCLUDE: sha256c.tcl

package require Tcl 8.2;                # tcl minimum version

namespace eval ::sha2 {
    variable  accel
    array set accel {tcl 0 critcl 0}
    variable  loaded {}

    namespace export sha256 hmac \
            SHA256Init SHA256Update SHA256Final


    variable uid
    if {![info exists uid]} {
        set uid 0
    }

    variable K
    if {![info exists K]} {
        # FIPS 180-2: 4.2.2 SHA-256 constants
        set K [list \
                   0x428a2f98 0x71374491 0xb5c0fbcf 0xe9b5dba5 \
                   0x3956c25b 0x59f111f1 0x923f82a4 0xab1c5ed5 \
                   0xd807aa98 0x12835b01 0x243185be 0x550c7dc3 \
                   0x72be5d74 0x80deb1fe 0x9bdc06a7 0xc19bf174 \
                   0xe49b69c1 0xefbe4786 0x0fc19dc6 0x240ca1cc \
                   0x2de92c6f 0x4a7484aa 0x5cb0a9dc 0x76f988da \
                   0x983e5152 0xa831c66d 0xb00327c8 0xbf597fc7 \
                   0xc6e00bf3 0xd5a79147 0x06ca6351 0x14292967 \
                   0x27b70a85 0x2e1b2138 0x4d2c6dfc 0x53380d13 \
                   0x650a7354 0x766a0abb 0x81c2c92e 0x92722c85 \
                   0xa2bfe8a1 0xa81a664b 0xc24b8b70 0xc76c51a3 \
                   0xd192e819 0xd6990624 0xf40e3585 0x106aa070 \
                   0x19a4c116 0x1e376c08 0x2748774c 0x34b0bcb5 \
                   0x391c0cb3 0x4ed8aa4a 0x5b9cca4f 0x682e6ff3 \
                   0x748f82ee 0x78a5636f 0x84c87814 0x8cc70208 \
                   0x90befffa 0xa4506ceb 0xbef9a3f7 0xc67178f2 \
                  ]
    }
    
}

# -------------------------------------------------------------------------
# Management of sha256 implementations.

# LoadAccelerator --
#
#	This package can make use of a number of compiled extensions to
#	accelerate the digest computation. This procedure manages the
#	use of these extensions within the package. During normal usage
#	this should not be called, but the test package manipulates the
#	list of enabled accelerators.
#
proc ::sha2::LoadAccelerator {name} {
    variable accel
    set r 0
    switch -exact -- $name {
        tcl {
            # Already present (this file)
            set r 1
        }
        critcl {
            if {![catch {package require tcllibc}]
                || ![catch {package require sha256c}]} {
                set r [expr {[info commands ::sha2::sha256c_update] != {}}]
            }
        }
        default {
            return -code error "invalid accelerator $key:\
                must be one of [join [KnownImplementations] {, }]"
        }
    }
    set accel($name) $r
    return $r
}

# ::sha2::Implementations --
#
#	Determines which implementations are
#	present, i.e. loaded.
#
# Arguments:
#	None.
#
# Results:
#	A list of implementation keys.

proc ::sha2::Implementations {} {
    variable accel
    set res {}
    foreach n [array names accel] {
	if {!$accel($n)} continue
	lappend res $n
    }
    return $res
}

# ::sha2::KnownImplementations --
#
#	Determines which implementations are known
#	as possible implementations.
#
# Arguments:
#	None.
#
# Results:
#	A list of implementation keys. In the order
#	of preference, most prefered first.

proc ::sha2::KnownImplementations {} {
    return {critcl tcl}
}

proc ::sha2::Names {} {
    return {
	critcl   {tcllibc based}
	tcl      {pure Tcl}
    }
}

# ::sha2::SwitchTo --
#
#	Activates a loaded named implementation.
#
# Arguments:
#	key	Name of the implementation to activate.
#
# Results:
#	None.

proc ::sha2::SwitchTo {key} {
    variable accel
    variable loaded

    if {[string equal $key $loaded]} {
	# No change, nothing to do.
	return
    } elseif {![string equal $key ""]} {
	# Validate the target implementation of the switch.

	if {![info exists accel($key)]} {
	    return -code error "Unable to activate unknown implementation \"$key\""
	} elseif {![info exists accel($key)] || !$accel($key)} {
	    return -code error "Unable to activate missing implementation \"$key\""
	}
    }

    # Deactivate the previous implementation, if there was any.

    if {![string equal $loaded ""]} {
        foreach c {
            SHA256Init   SHA224Init
            SHA256Final  SHA224Final
            SHA256Update
        } {
            interp alias {} ::sha2::$c {}
        }
    }

    # Activate the new implementation, if there is any.

    if {![string equal $key ""]} {
        foreach c {
            SHA256Init   SHA224Init
            SHA256Final  SHA224Final
            SHA256Update
        } {
	    interp alias {} ::sha2::$c {} ::sha2::${c}-${key}
        }
    }

    # Remember the active implementation, for deactivation by future
    # switches.

    set loaded $key
    return
}

# -------------------------------------------------------------------------

# SHA256Init --
#
#   Create and initialize an SHA256 state variable. This will be
#   cleaned up when we call SHA256Final
#

proc ::sha2::SHA256Init-tcl {} {
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token tok

    # FIPS 180-2: 5.3.2 Setting the initial hash value
    array set tok \
            [list \
            A [expr {int(0x6a09e667)}] \
            B [expr {int(0xbb67ae85)}] \
            C [expr {int(0x3c6ef372)}] \
            D [expr {int(0xa54ff53a)}] \
            E [expr {int(0x510e527f)}] \
            F [expr {int(0x9b05688c)}] \
            G [expr {int(0x1f83d9ab)}] \
            H [expr {int(0x5be0cd19)}] \
            n 0 i "" v 256]
    return $token
}

proc ::sha2::SHA256Init-critcl {} {
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token tok

    # FIPS 180-2: 5.3.2 Setting the initial hash value
    set tok(sha256c) [sha256c_init256]
    return $token
}

# SHA256Update --
#
#   This is called to add more data into the hash. You may call this
#   as many times as you require. Note that passing in "ABC" is equivalent
#   to passing these letters in as separate calls -- hence this proc 
#   permits hashing of chunked data
#
#   If we have a C-based implementation available, then we will use
#   it here in preference to the pure-Tcl implementation.
#

proc ::sha2::SHA256Update-tcl {token data} {
    upvar #0 $token state

    # Update the state values
    incr   state(n) [string length $data]
    append state(i) $data

    # Calculate the hash for any complete blocks
    set len [string length $state(i)]
    for {set n 0} {($n + 64) <= $len} {} {
        SHA256Transform $token [string range $state(i) $n [incr n 64]]
    }

    # Adjust the state for the blocks completed.
    set state(i) [string range $state(i) $n end]
    return
}

proc ::sha2::SHA256Update-critcl {token data} {
    upvar #0 $token state

    set state(sha256c) [sha256c_update $data $state(sha256c)]
    return
}

# SHA256Final --
#
#    This procedure is used to close the current hash and returns the
#    hash data. Once this procedure has been called the hash context
#    is freed and cannot be used again.
#
#    Note that the output is 256 bits represented as binary data.
#

proc ::sha2::SHA256Final-tcl {token} {
    upvar #0 $token state
    SHA256Penultimate $token
    
    # Output
    set r [bytes $state(A)][bytes $state(B)][bytes $state(C)][bytes $state(D)][bytes $state(E)][bytes $state(F)][bytes $state(G)][bytes $state(H)]
    unset state
    return $r
}

proc ::sha2::SHA256Final-critcl {token} {
    upvar #0 $token state
    set r $state(sha256c)
    unset  state
    return $r
}

# SHA256Penultimate --
#
#
proc ::sha2::SHA256Penultimate {token} {
    upvar #0 $token state

    # FIPS 180-2: 5.1.1: Padding the message
    #
    set len [string length $state(i)]
    set pad [expr {56 - ($len % 64)}]
    if {$len % 64 > 56} {
        incr pad 64
    }
    if {$pad == 0} {
        incr pad 64
    }
    append state(i) [binary format a$pad \x80]

    # Append length in bits as big-endian wide int.
    set dlen [expr {8 * $state(n)}]
    append state(i) [binary format II 0 $dlen]

    # Calculate the hash for the remaining block.
    set len [string length $state(i)]
    for {set n 0} {($n + 64) <= $len} {} {
        SHA256Transform $token [string range $state(i) $n [incr n 64]]
    }
}

# -------------------------------------------------------------------------

proc ::sha2::SHA224Init-tcl {} {
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token tok

    # FIPS 180-2 (change notice 1) (1): SHA-224 initialization values
    array set tok \
            [list \
            A [expr {int(0xc1059ed8)}] \
            B [expr {int(0x367cd507)}] \
            C [expr {int(0x3070dd17)}] \
            D [expr {int(0xf70e5939)}] \
            E [expr {int(0xffc00b31)}] \
            F [expr {int(0x68581511)}] \
            G [expr {int(0x64f98fa7)}] \
            H [expr {int(0xbefa4fa4)}] \
            n 0 i "" v 224]
    return $token
}

proc ::sha2::SHA224Init-critcl {} {
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token tok

    # FIPS 180-2 (change notice 1) (1): SHA-224 initialization values
    set tok(sha256c) [sha256c_init224]
    return $token
}

interp alias {} ::sha2::SHA224Update {} ::sha2::SHA256Update

proc ::sha2::SHA224Final-tcl {token} {
    upvar #0 $token state
    SHA256Penultimate $token
    
    # Output
    set r [bytes $state(A)][bytes $state(B)][bytes $state(C)][bytes $state(D)][bytes $state(E)][bytes $state(F)][bytes $state(G)]
    unset state
    return $r
}

proc ::sha2::SHA224Final-critcl {token} {
    upvar #0 $token state
    # Trim result down to 224 bits (by 4 bytes).
    # See output below, A..G, not A..H
    set r [string range $state(sha256c) 0 end-4]
    unset state
    return $r
}

# -------------------------------------------------------------------------
# HMAC Hashed Message Authentication (RFC 2104)
#
# hmac = H(K xor opad, H(K xor ipad, text))
#

# HMACInit --
#
#    This is equivalent to the SHA1Init procedure except that a key is
#    added into the algorithm
#
proc ::sha2::HMACInit {K} {

    # Key K is adjusted to be 64 bytes long. If K is larger, then use
    # the SHA1 digest of K and pad this instead.
    set len [string length $K]
    if {$len > 64} {
        set tok [SHA256Init]
        SHA256Update $tok $K
        set K [SHA256Final $tok]
        set len [string length $K]
    }
    set pad [expr {64 - $len}]
    append K [string repeat \0 $pad]

    # Cacluate the padding buffers.
    set Ki {}
    set Ko {}
    binary scan $K i16 Ks
    foreach k $Ks {
        append Ki [binary format i [expr {$k ^ 0x36363636}]]
        append Ko [binary format i [expr {$k ^ 0x5c5c5c5c}]]
    }

    set tok [SHA256Init]
    SHA256Update $tok $Ki;                 # initialize with the inner pad
    
    # preserve the Ko value for the final stage.
    # FRINK: nocheck
    set [subst $tok](Ko) $Ko

    return $tok
}

# HMACUpdate --
#
#    Identical to calling SHA256Update
#
proc ::sha2::HMACUpdate {token data} {
    SHA256Update $token $data
    return
}

# HMACFinal --
#
#    This is equivalent to the SHA256Final procedure. The hash context is
#    closed and the binary representation of the hash result is returned.
#
proc ::sha2::HMACFinal {token} {
    upvar #0 $token state

    set tok [SHA256Init];                 # init the outer hashing function
    SHA256Update $tok $state(Ko);         # prepare with the outer pad.
    SHA256Update $tok [SHA256Final $token]; # hash the inner result
    return [SHA256Final $tok]
}

# -------------------------------------------------------------------------
# Description:
#  This is the core SHA1 algorithm. It is a lot like the MD4 algorithm but
#  includes an extra round and a set of constant modifiers throughout.
#
set ::sha2::SHA256Transform_body {
    variable K
    upvar #0 $token state

    # FIPS 180-2: 6.2.2 SHA-256 Hash computation.
    binary scan $msg I* blocks
    set blockLen [llength $blocks]
    for {set i 0} {$i < $blockLen} {incr i 16} {
        set W [lrange $blocks $i [expr {$i+15}]]

        # FIPS 180-2: 6.2.2 (1) Prepare the message schedule
        # For t = 16 to 64 
        #   let Wt = (sigma1(Wt-2) + Wt-7 + sigma0(Wt-15) + Wt-16)
        set t2  13
        set t7   8
        set t15  0
        set t16 -1
        for {set t 16} {$t < 64} {incr t} {
            lappend W [expr {([sigma1 [lindex $W [incr t2]]] \
                                 + [lindex $W [incr t7]] \
                                 + [sigma0 [lindex $W [incr t15]]] \
                                 + [lindex $W [incr t16]]) & 0xffffffff}]
        }
        
        # FIPS 180-2: 6.2.2 (2) Initialise the working variables
        set A $state(A)
        set B $state(B)
        set C $state(C)
        set D $state(D)
        set E $state(E)
        set F $state(F)
        set G $state(G)
        set H $state(H)

        # FIPS 180-2: 6.2.2 (3) Do permutation rounds
        # For t = 0 to 63 do
        #   T1 = h + SIGMA1(e) + Ch(e,f,g) + Kt + Wt
        #   T2 = SIGMA0(a) + Maj(a,b,c)
        #   h = g; g = f;  f = e;  e = d + T1;  d = c;  c = b; b = a;
        #   a = T1 + T2
        #
        for {set t 0} {$t < 64} {incr t} {
            set T1 [expr {($H + [SIGMA1 $E] + [Ch $E $F $G] 
                          + [lindex $K $t] + [lindex $W $t]) & 0xffffffff}]
            set T2 [expr {([SIGMA0 $A] + [Maj $A $B $C]) & 0xffffffff}]
            set H $G
            set G $F
            set F $E
            set E [expr {($D + $T1) & 0xffffffff}]
            set D $C
            set C $B
            set B $A
            set A [expr {($T1 + $T2) & 0xffffffff}]
        }

        # FIPS 180-2: 6.2.2 (4) Compute the intermediate hash
        incr state(A) $A
        incr state(B) $B
        incr state(C) $C
        incr state(D) $D
        incr state(E) $E
        incr state(F) $F
        incr state(G) $G
        incr state(H) $H
    }

    return
}

# -------------------------------------------------------------------------

# FIPS 180-2: 4.1.2 equation 4.2
proc ::sha2::Ch {x y z} {
    return [expr {($x & $y) ^ (~$x & $z)}]
}

# FIPS 180-2: 4.1.2 equation 4.3
proc ::sha2::Maj {x y z} {
    return [expr {($x & $y) ^ ($x & $z) ^ ($y & $z)}]
}

# FIPS 180-2: 4.1.2 equation 4.4
#  (x >>> 2) ^ (x >>> 13) ^ (x >>> 22)
proc ::sha2::SIGMA0 {x} {
    return [expr {[>>> $x 2] ^ [>>> $x 13] ^ [>>> $x 22]}]
}

# FIPS 180-2: 4.1.2 equation 4.5
#  (x >>> 6) ^ (x >>> 11) ^ (x >>> 25)
proc ::sha2::SIGMA1 {x} {
    return [expr {[>>> $x 6] ^ [>>> $x 11] ^ [>>> $x 25]}]
}

# FIPS 180-2: 4.1.2 equation 4.6
#  s0 = (x >>> 7)  ^ (x >>> 18) ^ (x >> 3)
proc ::sha2::sigma0 {x} {
    #return [expr {[>>> $x 7] ^ [>>> $x 18] ^ (($x >> 3) & 0x1fffffff)}]
    return [expr {((($x<<25) | (($x>>7) & (0x7FFFFFFF>>6))) \
                 ^ (($x<<14) | (($x>>18) & (0x7FFFFFFF>>17))) & 0xFFFFFFFF) \
                 ^ (($x>>3) & 0x1fffffff)}]
}

# FIPS 180-2: 4.1.2 equation 4.7
#  s1 = (x >>> 17) ^ (x >>> 19) ^ (x >> 10)
proc ::sha2::sigma1 {x} {
    #return [expr {[>>> $x 17] ^ [>>> $x 19] ^ (($x >> 10) & 0x003fffff)}]
    return [expr {((($x<<15) | (($x>>17) & (0x7FFFFFFF>>16))) \
                 ^ (($x<<13) | (($x>>19) & (0x7FFFFFFF>>18))) & 0xFFFFFFFF) \
                 ^ (($x >> 10) & 0x003fffff)}]
}

# 32bit rotate-right
proc ::sha2::>>> {v n} {
    return [expr {(($v << (32 - $n)) \
                       | (($v >> $n) & (0x7FFFFFFF >> ($n - 1)))) \
                      & 0xFFFFFFFF}]
}

# 32bit rotate-left
proc ::sha2::<<< {v n} {
    return [expr {((($v << $n) \
                        | (($v >> (32 - $n)) \
                               & (0x7FFFFFFF >> (31 - $n))))) \
                      & 0xFFFFFFFF}]
}

# -------------------------------------------------------------------------
# We speed up the SHA256Transform code while maintaining readability in the
# source code by substituting inline for a number of functions.
# The idea is to reduce the number of [expr] calls.

# Inline the Ch function
regsub -all -line \
    {\[Ch (\$[ABCDEFGH]) (\$[ABCDEFGH]) (\$[ABCDEFGH])\]} \
    $::sha2::SHA256Transform_body \
    {((\1 \& \2) ^ ((~\1) \& \3))} \
    ::sha2::SHA256Transform_body

# Inline the Maj function
regsub -all -line \
    {\[Maj (\$[ABCDEFGH]) (\$[ABCDEFGH]) (\$[ABCDEFGH])\]} \
    $::sha2::SHA256Transform_body \
    {((\1 \& \2) ^ (\1 \& \3) ^ (\2 \& \3))} \
    ::sha2::SHA256Transform_body


# Inline the SIGMA0 function
regsub -all -line \
    {\[SIGMA0 (\$[ABCDEFGH])\]} \
    $::sha2::SHA256Transform_body \
    {((((\1<<30) | ((\1>>2) \& (0x7FFFFFFF>>1))) \& 0xFFFFFFFF) \
          ^ (((\1<<19) | ((\1>>13) \& (0x7FFFFFFF>>12))) \& 0xFFFFFFFF) \
          ^ (((\1<<10) | ((\1>>22) \& (0x7FFFFFFF>>21))) \& 0xFFFFFFFF) \
          )} \
    ::sha2::SHA256Transform_body

# Inline the SIGMA1 function
regsub -all -line \
    {\[SIGMA1 (\$[ABCDEFGH])\]} \
    $::sha2::SHA256Transform_body \
    {((((\1<<26) | ((\1>>6) \& (0x7FFFFFFF>>5))) \& 0xFFFFFFFF) \
          ^ (((\1<<21) | ((\1>>11) \& (0x7FFFFFFF>>10))) \& 0xFFFFFFFF) \
          ^ (((\1<<7) | ((\1>>25) \& (0x7FFFFFFF>>24))) \& 0xFFFFFFFF) \
          )} \
    ::sha2::SHA256Transform_body

proc ::sha2::SHA256Transform {token msg} $::sha2::SHA256Transform_body

# -------------------------------------------------------------------------

# Convert a integer value into a binary string in big-endian order.
proc ::sha2::byte {n v} {expr {((0xFF << (8 * $n)) & $v) >> (8 * $n)}}
proc ::sha2::bytes {v} { 
    #format %c%c%c%c [byte 3 $v] [byte 2 $v] [byte 1 $v] [byte 0 $v]
    format %c%c%c%c \
        [expr {((0xFF000000 & $v) >> 24) & 0xFF}] \
        [expr {(0xFF0000 & $v) >> 16}] \
        [expr {(0xFF00 & $v) >> 8}] \
        [expr {0xFF & $v}]
}

# -------------------------------------------------------------------------

proc ::sha2::Hex {data} {
    binary scan $data H* result
    return $result
}

# -------------------------------------------------------------------------

# Description:
#  Pop the nth element off a list. Used in options processing.
#
proc ::sha2::Pop {varname {nth 0}} {
    upvar $varname args
    set r [lindex $args $nth]
    set args [lreplace $args $nth $nth]
    return $r
}

# -------------------------------------------------------------------------

# fileevent handler for chunked file hashing.
#
proc ::sha2::Chunk {token channel {chunksize 4096}} {
    upvar #0 $token state
    
    SHA256Update $token [read $channel $chunksize]

    if {[eof $channel]} {
        fileevent $channel readable {}
        set state(reading) 0
    }
    return
}

# -------------------------------------------------------------------------

proc ::sha2::_sha256 {ver args} {
    array set opts {-hex 0 -filename {} -channel {} -chunksize 4096}
    if {[llength $args] == 1} {
        set opts(-hex) 1
    } else {
        while {[string match -* [set option [lindex $args 0]]]} {
            switch -glob -- $option {
                -hex       { set opts(-hex) 1 }
                -bin       { set opts(-hex) 0 }
                -file*     { set opts(-filename) [Pop args 1] }
                -channel   { set opts(-channel) [Pop args 1] }
                -chunksize { set opts(-chunksize) [Pop args 1] }
                default {
                    if {[llength $args] == 1} { break }
                    if {[string compare $option "--"] == 0} { Pop args; break }
                    set err [join [lsort [concat -bin [array names opts]]] ", "]
                    return -code error "bad option $option:\
                    must be one of $err"
                }
            }
            Pop args
        }
    }

    if {$opts(-filename) != {}} {
        set opts(-channel) [open $opts(-filename) r]
        fconfigure $opts(-channel) -translation binary
    }

    if {$opts(-channel) == {}} {
        if {[llength $args] != 1} {
            return -code error "wrong # args: should be\
                \"[namespace current]::sha$ver ?-hex|-bin? -filename file\
                | -channel channel | string\""
        }
        set tok [SHA${ver}Init]
        SHA${ver}Update $tok [lindex $args 0]
        set r [SHA${ver}Final $tok]

    } else {

        set tok [SHA${ver}Init]
        # FRINK: nocheck
        set [subst $tok](reading) 1
        fileevent $opts(-channel) readable \
            [list [namespace origin Chunk] \
                 $tok $opts(-channel) $opts(-chunksize)]
        # FRINK: nocheck
        vwait [subst $tok](reading)
        set r [SHA${ver}Final $tok]

        # If we opened the channel - we should close it too.
        if {$opts(-filename) != {}} {
            close $opts(-channel)
        }
    }
    
    if {$opts(-hex)} {
        set r [Hex $r]
    }
    return $r
}

interp alias {} ::sha2::sha256 {} ::sha2::_sha256 256
interp alias {} ::sha2::sha224 {} ::sha2::_sha256 224

# -------------------------------------------------------------------------

proc ::sha2::hmac {args} {
    array set opts {-hex 1 -filename {} -channel {} -chunksize 4096}
    if {[llength $args] != 2} {
        while {[string match -* [set option [lindex $args 0]]]} {
            switch -glob -- $option {
                -key       { set opts(-key) [Pop args 1] }
                -hex       { set opts(-hex) 1 }
                -bin       { set opts(-hex) 0 }
                -file*     { set opts(-filename) [Pop args 1] }
                -channel   { set opts(-channel) [Pop args 1] }
                -chunksize { set opts(-chunksize) [Pop args 1] }
                default {
                    if {[llength $args] == 1} { break }
                    if {[string compare $option "--"] == 0} { Pop args; break }
                    set err [join [lsort [array names opts]] ", "]
                    return -code error "bad option $option:\
                    must be one of $err"
                }
            }
            Pop args
        }
    }

    if {[llength $args] == 2} {
        set opts(-key) [Pop args]
    }

    if {![info exists opts(-key)]} {
        return -code error "wrong # args:\
            should be \"hmac ?-hex? -key key -filename file | string\""
    }

    if {$opts(-filename) != {}} {
        set opts(-channel) [open $opts(-filename) r]
        fconfigure $opts(-channel) -translation binary
    }

    if {$opts(-channel) == {}} {

        if {[llength $args] != 1} {
            return -code error "wrong # args:\
                should be \"hmac ?-hex? -key key -filename file | string\""
        }
        set tok [HMACInit $opts(-key)]
        HMACUpdate $tok [lindex $args 0]
        set r [HMACFinal $tok]

    } else {

        set tok [HMACInit $opts(-key)]
        # FRINK: nocheck
        set [subst $tok](reading) 1
        fileevent $opts(-channel) readable \
            [list [namespace origin Chunk] \
                 $tok $opts(-channel) $opts(-chunksize)]
        # FRINK: nocheck
        vwait [subst $tok](reading)
        set r [HMACFinal $tok]

        # If we opened the channel - we should close it too.
        if {$opts(-filename) != {}} {
            close $opts(-channel)
        }
    }
    
    if {$opts(-hex)} {
        set r [Hex $r]
    }
    return $r
}

# -------------------------------------------------------------------------

# Try and load a compiled extension to help.
namespace eval ::sha2 {
    variable e {}
    foreach e [KnownImplementations] {
	if {[LoadAccelerator $e]} {
	    SwitchTo $e
	    break
	}
    }
    unset e
}

package provide sha256 1.0.4

# -------------------------------------------------------------------------
# Local Variables:
#   mode: tcl
#   indent-tabs-mode: nil
# End:

# THATCH /pkgs/tcllib/base64/pkgIndex.tcl size 360 mtime 1659183355
if {![package vsatisfies [package provide Tcl] 8.2]} {return}
package ifneeded base64   2.5 [list source [file join $dir base64.tcl]]
package ifneeded uuencode 1.1.5 [list source [file join $dir uuencode.tcl]]
package ifneeded yencode  1.1.3 [list source [file join $dir yencode.tcl]]
package ifneeded ascii85  1.0   [list source [file join $dir ascii85.tcl]]

# THATCH /pkgs/tcllib/base64/base64.tcl size 12080 mtime 1659183355
# base64.tcl --
#
# Encode/Decode base64 for a string
# Stephen Uhler / Brent Welch (c) 1997 Sun Microsystems
# The decoder was done for exmh by Chris Garrigues
#
# Copyright (c) 1998-2000 by Ajuba Solutions.
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Version 1.0   implemented Base64_Encode, Base64_Decode
# Version 2.0   uses the base64 namespace
# Version 2.1   fixes various decode bugs and adds options to encode
# Version 2.2   is much faster, Tcl8.0 compatible
# Version 2.2.1 bugfixes
# Version 2.2.2 bugfixes
# Version 2.3   bugfixes and extended to support Trf
# Version 2.4.x bugfixes

# @mdgen EXCLUDE: base64c.tcl

package require Tcl 8.2
namespace eval ::base64 {
    namespace export encode decode
}

package provide base64 2.5

if {[package vsatisfies [package require Tcl] 8.6]} {
    proc ::base64::encode {args} {
	binary encode base64 -maxlen 76 {*}$args
    }

    proc ::base64::decode {string} {
	# Tcllib is strict with respect to end of input, yet lax for
	# invalid characters outside of that.
	regsub -all -- {[^ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/]} $string {} string
	binary decode base64 -strict $string
    }

    return
}

if {![catch {package require Trf 2.0}]} {
    # Trf is available, so implement the functionality provided here
    # in terms of calls to Trf for speed.

    # ::base64::encode --
    #
    #	Base64 encode a given string.
    #
    # Arguments:
    #	args	?-maxlen maxlen? ?-wrapchar wrapchar? string
    #
    #		If maxlen is 0, the output is not wrapped.
    #
    # Results:
    #	A Base64 encoded version of $string, wrapped at $maxlen characters
    #	by $wrapchar.

    proc ::base64::encode {args} {
	# Set the default wrapchar and maximum line length to match
	# the settings for MIME encoding (RFC 3548, RFC 2045). These
	# are the settings used by Trf as well. Various RFCs allow for
	# different wrapping characters and wraplengths, so these may
	# be overridden by command line options.
	set wrapchar "\n"
	set maxlen 76

	if { [llength $args] == 0 } {
	    error "wrong # args: should be \"[lindex [info level 0] 0]\
		    ?-maxlen maxlen? ?-wrapchar wrapchar? string\""
	}

	set optionStrings [list "-maxlen" "-wrapchar"]
	for {set i 0} {$i < [llength $args] - 1} {incr i} {
	    set arg [lindex $args $i]
	    set index [lsearch -glob $optionStrings "${arg}*"]
	    if { $index == -1 } {
		error "unknown option \"$arg\": must be -maxlen or -wrapchar"
	    }
	    incr i
	    if { $i >= [llength $args] - 1 } {
		error "value for \"$arg\" missing"
	    }
	    set val [lindex $args $i]

	    # The name of the variable to assign the value to is extracted
	    # from the list of known options, all of which have an
	    # associated variable of the same name as the option without
	    # a leading "-". The [string range] command is used to strip
	    # of the leading "-" from the name of the option.
	    #
	    # FRINK: nocheck
	    set [string range [lindex $optionStrings $index] 1 end] $val
	}

	# [string is] requires Tcl8.2; this works with 8.0 too
	if {[catch {expr {$maxlen % 2}}]} {
	    return -code error "expected integer but got \"$maxlen\""
	} elseif {$maxlen < 0} {
	    return -code error "expected positive integer but got \"$maxlen\""
	}

	set string [lindex $args end]
	set result [::base64 -mode encode -- $string]

	# Trf's encoder implicitly uses the settings -maxlen 76,
	# -wrapchar \n for its output. We may have to reflow this for
	# the settings chosen by the user. A second difference is that
	# Trf closes the output with the wrap char sequence,
	# always. The code here doesn't. Therefore 'trimright' is
	# needed in the fast cases.

	if {($maxlen == 76) && [string equal $wrapchar \n]} {
	    # Both maxlen and wrapchar are identical to Trf's
	    # settings. This is the super-fast case, because nearly
	    # nothing has to be done. Only thing to do is strip a
	    # terminating wrapchar.
	    set result [string trimright $result]
	} elseif {$maxlen == 76} {
	    # wrapchar has to be different here, length is the
	    # same. We can use 'string map' to transform the wrap
	    # information.
	    set result [string map [list \n $wrapchar] \
			    [string trimright $result]]
	} elseif {$maxlen == 0} {
	    # Have to reflow the output to no wrapping. Another fast
	    # case using only 'string map'. 'trimright' is not needed
	    # here.

	    set result [string map [list \n ""] $result]
	} else {
	    # Have to reflow the output from 76 to the chosen maxlen,
	    # and possibly change the wrap sequence as well.

	    # Note: After getting rid of the old wrap sequence we
	    # extract the relevant segments from the string without
	    # modifying the string. Modification, i.e. removal of the
	    # processed part, means 'shifting down characters in
	    # memory', making the algorithm O(n^2). By avoiding the
	    # modification we stay in O(n).

	    set result [string map [list \n ""] $result]
	    set l [expr {[string length $result]-$maxlen}]
	    for {set off 0} {$off < $l} {incr off $maxlen} {
		append res [string range $result $off [expr {$off+$maxlen-1}]] $wrapchar
	    }
	    append res [string range $result $off end]
	    set result $res
	}

	return $result
    }

    # ::base64::decode --
    #
    #	Base64 decode a given string.
    #
    # Arguments:
    #	string	The string to decode.  Characters not in the base64
    #		alphabet are ignored (e.g., newlines)
    #
    # Results:
    #	The decoded value.

    proc ::base64::decode {string} {
	regsub -all {\s} $string {} string
	::base64 -mode decode -- $string
    }

} else {
    # Without Trf use a pure tcl implementation

    namespace eval base64 {
	variable base64 {}
	variable base64_en {}

	# We create the auxiliary array base64_tmp, it will be unset later.
	variable base64_tmp
	variable i

	set i 0
	foreach char {A B C D E F G H I J K L M N O P Q R S T U V W X Y Z \
		a b c d e f g h i j k l m n o p q r s t u v w x y z \
		0 1 2 3 4 5 6 7 8 9 + /} {
	    set base64_tmp($char) $i
	    lappend base64_en $char
	    incr i
	}

	#
	# Create base64 as list: to code for instance C<->3, specify
	# that [lindex $base64 67] be 3 (C is 67 in ascii); non-coded
	# ascii chars get a {}. we later use the fact that lindex on a
	# non-existing index returns {}, and that [expr {} < 0] is true
	#

	# the last ascii char is 'z'
	variable char
	variable len
	variable val

	scan z %c len
	for {set i 0} {$i <= $len} {incr i} {
	    set char [format %c $i]
	    set val {}
	    if {[info exists base64_tmp($char)]} {
		set val $base64_tmp($char)
	    } else {
		set val {}
	    }
	    lappend base64 $val
	}

	# code the character "=" as -1; used to signal end of message
	scan = %c i
	set base64 [lreplace $base64 $i $i -1]

	# remove unneeded variables
	unset base64_tmp i char len val

	namespace export encode decode
    }

    # ::base64::encode --
    #
    #	Base64 encode a given string.
    #
    # Arguments:
    #	args	?-maxlen maxlen? ?-wrapchar wrapchar? string
    #
    #		If maxlen is 0, the output is not wrapped.
    #
    # Results:
    #	A Base64 encoded version of $string, wrapped at $maxlen characters
    #	by $wrapchar.

    proc ::base64::encode {args} {
	set base64_en $::base64::base64_en

	# Set the default wrapchar and maximum line length to match
	# the settings for MIME encoding (RFC 3548, RFC 2045). These
	# are the settings used by Trf as well. Various RFCs allow for
	# different wrapping characters and wraplengths, so these may
	# be overridden by command line options.
	set wrapchar "\n"
	set maxlen 76

	if { [llength $args] == 0 } {
	    error "wrong # args: should be \"[lindex [info level 0] 0]\
		    ?-maxlen maxlen? ?-wrapchar wrapchar? string\""
	}

	set optionStrings [list "-maxlen" "-wrapchar"]
	for {set i 0} {$i < [llength $args] - 1} {incr i} {
	    set arg [lindex $args $i]
	    set index [lsearch -glob $optionStrings "${arg}*"]
	    if { $index == -1 } {
		error "unknown option \"$arg\": must be -maxlen or -wrapchar"
	    }
	    incr i
	    if { $i >= [llength $args] - 1 } {
		error "value for \"$arg\" missing"
	    }
	    set val [lindex $args $i]

	    # The name of the variable to assign the value to is extracted
	    # from the list of known options, all of which have an
	    # associated variable of the same name as the option without
	    # a leading "-". The [string range] command is used to strip
	    # of the leading "-" from the name of the option.
	    #
	    # FRINK: nocheck
	    set [string range [lindex $optionStrings $index] 1 end] $val
	}

	# [string is] requires Tcl8.2; this works with 8.0 too
	if {[catch {expr {$maxlen % 2}}]} {
	    return -code error "expected integer but got \"$maxlen\""
	} elseif {$maxlen < 0} {
	    return -code error "expected positive integer but got \"$maxlen\""
	}

	set string [lindex $args end]

	set result {}
	set state 0
	set length 0


	# Process the input bytes 3-by-3

	binary scan $string c* X

	foreach {x y z} $X {
	    ADD [lindex $base64_en [expr {($x >>2) & 0x3F}]]
	    if {$y != {}} {
		ADD [lindex $base64_en [expr {(($x << 4) & 0x30) | (($y >> 4) & 0xF)}]]
		if {$z != {}} {
		    ADD [lindex $base64_en [expr {(($y << 2) & 0x3C) | (($z >> 6) & 0x3)}]]
		    ADD [lindex $base64_en [expr {($z & 0x3F)}]]
		} else {
		    set state 2
		    break
		}
	    } else {
		set state 1
		break
	    }
	}
	if {$state == 1} {
	    ADD [lindex $base64_en [expr {(($x << 4) & 0x30)}]]
	    ADD =
	    ADD =
	} elseif {$state == 2} {
	    ADD [lindex $base64_en [expr {(($y << 2) & 0x3C)}]]
	    ADD =
	}
	return $result
    }

    proc ::base64::ADD {x} {
	# The line length check is always done before appending so
	# that we don't get an extra newline if the output is a
	# multiple of $maxlen chars long.

	upvar 1 maxlen maxlen length length result result wrapchar wrapchar
	if {$maxlen && $length >= $maxlen} {
	    append result $wrapchar
	    set length 0
	}
	append result $x
	incr length
	return
    }

    # ::base64::decode --
    #
    #	Base64 decode a given string.
    #
    # Arguments:
    #	string	The string to decode.  Characters not in the base64
    #		alphabet are ignored (e.g., newlines)
    #
    # Results:
    #	The decoded value.

    proc ::base64::decode {string} {
	if {[string length $string] == 0} {return ""}

	set base64 $::base64::base64
	set output "" ; # Fix for [Bug 821126]
	set nums {}

	binary scan $string c* X
	lappend X 61 ;# force a terminator
	foreach x $X {
	    set bits [lindex $base64 $x]
	    if {$bits >= 0} {
		if {[llength [lappend nums $bits]] == 4} {
		    foreach {v w z y} $nums break
		    set a [expr {($v << 2) | ($w >> 4)}]
		    set b [expr {(($w & 0xF) << 4) | ($z >> 2)}]
		    set c [expr {(($z & 0x3) << 6) | $y}]
		    append output [binary format ccc $a $b $c]
		    set nums {}
		}
	    } elseif {$bits == -1} {
		# = indicates end of data.  Output whatever chars are
		# left, if any.
		if {![llength $nums]} break
		# The encoding algorithm dictates that we can only
		# have 1 or 2 padding characters.  If x=={}, we must
		# (*) have 12 bits of input (enough for 1 8-bit
		# output).  If x!={}, we have 18 bits of input (enough
		# for 2 8-bit outputs).
		#
		# (*) If we don't then the input is broken (bug 2976290).

		foreach {v w z} $nums break

		# Bug 2976290
		if {$w == {}} {
		    return -code error "Not enough data to process padding"
		}

		set a [expr {($v << 2) | (($w & 0x30) >> 4)}]
		if {$z == {}} {
		    append output [binary format c $a ]
		} else {
		    set b [expr {(($w & 0xF) << 4) | (($z & 0x3C) >> 2)}]
		    append output [binary format cc $a $b]
		}
		break
	    } else {
		# RFC 2045 says that line breaks and other characters not part
		# of the Base64 alphabet must be ignored, and that the decoder
		# can optionally emit a warning or reject the message.  We opt
		# not to do so, but to just ignore the character.
		continue
	    }
	}
	return $output
    }
}

# # ## ### ##### ######## ############# #####################
return


# THATCH /pkgs/tcllib/license.terms size 2155 mtime 1659183355
This software is copyrighted by Ajuba Solutions and other parties.
The following terms apply to all files associated with the software
unless explicitly disclaimed in individual files.

The authors hereby grant permission to use, copy, modify, distribute,
and license this software and its documentation for any purpose,
provided that existing copyright notices are retained in all copies
and that this notice is included verbatim in any distributions. No
written agreement, license, or royalty fee is required for any of the
authorized uses.  Modifications to this software may be copyrighted by
their authors and need not follow the licensing terms described here,
provided that the new terms are clearly indicated on the first page of
each file where they apply.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
NON-INFRINGEMENT.  THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, AND
THE AUTHORS AND DISTRIBUTORS HAVE NO OBLIGATION TO PROVIDE
MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

GOVERNMENT USE: If you are acquiring this software on behalf of the
U.S. government, the Government shall have only "Restricted Rights" in
the software and related documentation as defined in the Federal
Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
are acquiring the software on behalf of the Department of Defense, the
software shall be classified as "Commercial Computer Software" and the
Government shall have only "Restricted Rights" as defined in Clause
252.227-7013 (c) (1) of DFARs.  Notwithstanding the foregoing, the
authors grant the U.S. Government and others acting in its behalf
permission to use and distribute the software in accordance with the
terms specified in this license.
